//========================================================================================
//  MemoryMappedDevice.java
//    en:Memory mapped device
//    ja:メモリマップトデバイス
//  Copyright (C) 2003-2022 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

//----------------------------------------------------------------------------------------
//  デバイスにアクセスするためのメソッドをenum bodyに記述する
//    mmdPbs,mmdPbz,mmdPws,mmdPwz,mmdPls  ピーク
//    mmdRbs,mmdRbz,mmdRws,mmdRwz,mmdRls  リード
//    mmdWb,mmdWw,mmdWl                   ライト
//  ピーク、リード、ライトの命名規則
//    4文字目  P=ピーク,R=リード,W=ライト
//    5文字目  b=バイト,w=ワード,l=ロング
//    6文字目  s=符号拡張,z=ゼロ拡張
//  ピークとリードの返却値の型はmmdPbsとmmdRbsだけbyte、他はint
//  ピークはSRAMスイッチの読み取りやデバッガなどで使用する
//  ピークはMPUやデバイスの状態を変化させず、例外もスローしない
//  リードとライトはMPUやDMAによる通常のアクセスで使用する
//  リードとライトはバスエラーをスローする場合がある
//  アドレスの未使用ビットはデバイスに渡る前にすべてクリアされていなければならない
//    バスエラーは未使用ビットがクリアされたアドレスで通知されることになる
//  異なるデバイスに跨るアクセスはデバイスに渡る前に分割されていなければならない
//  奇数アドレスに対するワードアクセスはデバイスに渡る前に分割または排除されていなければならない
//  4の倍数でないアドレスに対するロングアクセスはデバイスに渡る前に分割または排除されていなければならない
//  デバイスのメソッドを直接呼び出すときはアドレスのマスクや分割を忘れないこと
//----------------------------------------------------------------------------------------

package xeij;

import java.lang.*;  //Boolean,Character,Class,Comparable,Double,Exception,Float,IllegalArgumentException,Integer,Long,Math,Number,Object,Runnable,SecurityException,String,StringBuilder,System
import java.util.*;  //ArrayList,Arrays,Calendar,GregorianCalendar,HashMap,Map,Map.Entry,Timer,TimerTask,TreeMap

public enum MemoryMappedDevice {

  //--------------------------------------------------------------------------------
  //MMD_MMR メインメモリ
  MMD_MMR {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "メインメモリ" : "Main Memory";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getShort (a);
       } else {
         return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getChar (a);
       } else {
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getInt (a);
       } else {
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a    ] = (byte)  d;
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         MainMemory.mmrBuffer.putShort (a, (short) d);
       } else {
         MainMemory.mmrM8[a    ] = (byte) (d >> 8);
         MainMemory.mmrM8[a + 1] = (byte)  d;
       }
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         MainMemory.mmrBuffer.putInt (a, d);
       } else {
         MainMemory.mmrM8[a    ] = (byte) (d >> 24);
         MainMemory.mmrM8[a + 1] = (byte) (d >> 16);
         MainMemory.mmrM8[a + 2] = (byte) (d >> 8);
         MainMemory.mmrM8[a + 3] = (byte)  d;
       }
     }
  },  //MMD_MMR

  //--------------------------------------------------------------------------------
  //MMD_XMM 拡張メモリ
  MMD_XMM {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張メモリ" : "Expansion Memory";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a -= XEiJ.busExMemoryStart;
      return XEiJ.busExMemoryArray[a];
    }
    @Override protected int mmdPbz (int a) {
      a -= XEiJ.busExMemoryStart;
      return XEiJ.busExMemoryArray[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a -= XEiJ.busExMemoryStart;
      return XEiJ.busExMemoryArray[a] << 8 | (XEiJ.busExMemoryArray[a + 1] & 0xff);
    }
    @Override protected int mmdPwz (int a) {
      a -= XEiJ.busExMemoryStart;
      return (char) (XEiJ.busExMemoryArray[a] << 8 | (XEiJ.busExMemoryArray[a + 1] & 0xff));
    }
    @Override protected int mmdPls (int a) {
      a -= XEiJ.busExMemoryStart;
      return XEiJ.busExMemoryArray[a] << 24 | (XEiJ.busExMemoryArray[a + 1] & 0xff) << 16 | (char) (XEiJ.busExMemoryArray[a + 2] << 8 | (XEiJ.busExMemoryArray[a + 3] & 0xff));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       return XEiJ.busExMemoryArray[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       return XEiJ.busExMemoryArray[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       return XEiJ.busExMemoryArray[a] << 8 | (XEiJ.busExMemoryArray[a + 1] & 0xff);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       return (char) (XEiJ.busExMemoryArray[a] << 8 | (XEiJ.busExMemoryArray[a + 1] & 0xff));
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       return XEiJ.busExMemoryArray[a] << 24 | (XEiJ.busExMemoryArray[a + 1] & 0xff) << 16 | (char) (XEiJ.busExMemoryArray[a + 2] << 8 | (XEiJ.busExMemoryArray[a + 3] & 0xff));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       XEiJ.busExMemoryArray[a    ] = (byte)  d;
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       XEiJ.busExMemoryArray[a    ] = (byte) (d >> 8);
       XEiJ.busExMemoryArray[a + 1] = (byte)  d;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a -= XEiJ.busExMemoryStart;
       XEiJ.busExMemoryArray[a    ] = (byte) (d >> 24);
       XEiJ.busExMemoryArray[a + 1] = (byte) (d >> 16);
       XEiJ.busExMemoryArray[a + 2] = (byte) (d >> 8);
       XEiJ.busExMemoryArray[a + 3] = (byte)  d;
     }
  },  //MMD_XMM

  //--------------------------------------------------------------------------------
  //MMD_GE0 グラフィックス画面(512ドット16色ページ0)
  //
  //  512ドット16色
  //                アドレス             アクセス                  格納
  //    GE0  0x00c00000～0x00c7ffff  ............3210  ──  ............3210
  //    GE1  0x00c80000～0x00cfffff  ............7654  ──  ............7654
  //    GE2  0x00d00000～0x00d7ffff  ............ba98  ──  ............ba98
  //    GE3  0x00d80000～0x00dfffff  ............fedc  ──  ............fedc
  //
  MMD_GE0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 16 色 ページ 0)" : "Graphics Screen (512 dots 16 colors page 0)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a] = (byte) (d          & 15);
         int y = (a >> 10) - CRTC.crtR13GrYCurr[0] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 3] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GE0

  //--------------------------------------------------------------------------------
  //MMD_GE1 グラフィックス画面(512ドット16色ページ1)
  //
  //  512ドット16色
  //                アドレス             アクセス                  格納
  //    GE0  0x00c00000～0x00c7ffff  ............3210  ──  ............3210
  //    GE1  0x00c80000～0x00cfffff  ............7654  ──  ............7654
  //    GE2  0x00d00000～0x00d7ffff  ............ba98  ──  ............ba98
  //    GE3  0x00d80000～0x00dfffff  ............fedc  ──  ............fedc
  //
  MMD_GE1 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 16 色 ページ 1)" : "Graphics Screen (512 dots 16 colors page 1)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
      a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
      a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
      a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
      a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a] = (byte) (d          & 15);
         int y = (a >> 10) - CRTC.crtR13GrYCurr[1] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 3] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GE1

  //--------------------------------------------------------------------------------
  //MMD_GE2 グラフィックス画面(512ドット16色ページ2)
  //
  //  512ドット16色
  //                アドレス             アクセス                  格納
  //    GE0  0x00c00000～0x00c7ffff  ............3210  ──  ............3210
  //    GE1  0x00c80000～0x00cfffff  ............7654  ──  ............7654
  //    GE2  0x00d00000～0x00d7ffff  ............ba98  ──  ............ba98
  //    GE3  0x00d80000～0x00dfffff  ............fedc  ──  ............fedc
  //
  MMD_GE2 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 16 色 ページ 2)" : "Graphics Screen (512 dots 16 colors page 2)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a] = (byte) (d          & 15);
         int y = (a >> 10) - CRTC.crtR13GrYCurr[2] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 3] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GE2

  //--------------------------------------------------------------------------------
  //MMD_GE3 グラフィックス画面(512ドット16色ページ3)
  //
  //  512ドット16色
  //                アドレス             アクセス                  格納
  //    GE0  0x00c00000～0x00c7ffff  ............3210  ──  ............3210
  //    GE1  0x00c80000～0x00cfffff  ............7654  ──  ............7654
  //    GE2  0x00d00000～0x00d7ffff  ............ba98  ──  ............ba98
  //    GE3  0x00d80000～0x00dfffff  ............fedc  ──  ............fedc
  //
  MMD_GE3 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 16 色 ページ 3)" : "Graphics Screen (512 dots 16 colors page 3)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 3];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a] = (byte) (d          & 15);
         int y = (a >> 10) - CRTC.crtR13GrYCurr[3] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 1] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 3] = (byte) (d          & 15);
       int y = (a >> 10) - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GE3

  //--------------------------------------------------------------------------------
  //MMD_GF0 グラフィックス画面(512ドット256色ページ0)
  //
  //  512ドット256色
  //                アドレス             アクセス                  格納
  //    GF0  0x00c00000～0x00c7ffff  ........76543210  ─┬  ............3210
  //    GF1  0x00c80000～0x00cfffff  ........fedcba98  ┐└  ............7654
  //         0x00d00000～0x00d7ffff                    ├─  ............ba98
  //         0x00d80000～0x00dfffff                    └─  ............fedc
  //
  MMD_GF0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 256 色 ページ 0)" : "Graphics Screen (512 dots 256 colors page 0)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return (byte) (MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a]);
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00080001] << 20 | MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 0x00080003] << 4 | MainMemory.mmrM8[a + 3];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (byte) (MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a]);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00080001] << 20 | MainMemory.mmrM8[a + 1] << 16 | MainMemory.mmrM8[a + 0x00080003] << 4 | MainMemory.mmrM8[a + 3];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a + 0x00080000] = (byte) (d << 24  >>> 28);
         MainMemory.mmrM8[a             ] = (byte) (d          & 15);
         a >>= 10;
         int y = a - CRTC.crtR13GrYCurr[0] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[1] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00000001] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d <<  8  >>> 28);
       MainMemory.mmrM8[a + 0x00000001] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 0x00080003] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00000003] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GF0

  //--------------------------------------------------------------------------------
  //MMD_GF1 グラフィックス画面(512ドット256色ページ1)
  //
  //  512ドット256色
  //                アドレス             アクセス                  格納
  //    GF0  0x00c00000～0x00c7ffff  ........76543210  ─┬  ............3210
  //    GF1  0x00c80000～0x00cfffff  ........fedcba98  ┐└  ............7654
  //         0x00d00000～0x00d7ffff                    ├─  ............ba98
  //         0x00d80000～0x00dfffff                    └─  ............fedc
  //
  MMD_GF1 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 256 色 ページ 1)" : "Graphics Screen (512 dots 256 colors page 1)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return (byte) (MainMemory.mmrM8[a + 0x00100000] << 4 | MainMemory.mmrM8[a + 0x00080000]);
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00100000] << 4 | MainMemory.mmrM8[a + 0x00080000];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00100001] << 4 | MainMemory.mmrM8[a + 0x00080001];
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00100001] << 4 | MainMemory.mmrM8[a + 0x00080001];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00100001] << 20 | MainMemory.mmrM8[a + 0x00080001] << 16 | MainMemory.mmrM8[a + 0x00100003] << 4 | MainMemory.mmrM8[a + 0x00080003];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (byte) (MainMemory.mmrM8[a + 0x00100000] << 4 | MainMemory.mmrM8[a + 0x00080000]);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00100000] << 4 | MainMemory.mmrM8[a + 0x00080000];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00100001] << 4 | MainMemory.mmrM8[a + 0x00080001];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00100001] << 4 | MainMemory.mmrM8[a + 0x00080001];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00100001] << 20 | MainMemory.mmrM8[a + 0x00080001] << 16 | MainMemory.mmrM8[a + 0x00100003] << 4 | MainMemory.mmrM8[a + 0x00080003];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 < 0 : (a & 1) != 0) {
         MainMemory.mmrM8[a + 0x00100000] = (byte) (d << 24  >>> 28);
         MainMemory.mmrM8[a + 0x00080000] = (byte) (d          & 15);
         a >>= 10;
         int y = a - CRTC.crtR13GrYCurr[2] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[3] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00100001] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00100001] = (byte) (d <<  8  >>> 28);
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 0x00100003] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00080003] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GF1

  //--------------------------------------------------------------------------------
  //MMD_GM2 グラフィックス画面(メモリモード2)
  //
  //  バイトリード
  //    0x00c00000  0x00が読み出される
  //    0x00c00001  4bitページ1-0が読み出される
  //    0x00c80000  0x00が読み出される
  //    0x00c80001  0x00が読み出される
  //    0x00d00000  0x00が読み出される
  //    0x00d00001  4bitページ1-0が読み出される
  //    0x00d80000  0x00が読み出される
  //    0x00d80001  0x00が読み出される
  //
  //  ワードリード
  //    0x00c00000  上位バイトは0x00で下位バイトに4bitページ1-0が読み出される
  //    0x00c80000  上位バイトは0x00で下位バイトに4bitページ3-2が読み出される
  //    0x00d00000  上位バイトは0x00で下位バイトに4bitページ1-0が読み出される
  //    0x00d80000  上位バイトは0x00で下位バイトに4bitページ3-2が読み出される
  //
  //  バイトライト
  //    0x00c00000  060turboの060モードのときデータは無視され0x00が4bitページ3-2に書き込まれる。それ以外はデータが4bitページ3-2に書き込まれる
  //    0x00c00001  データが4bitページ1-0に書き込まれる
  //    0x00c80000  060turboの060モードのときデータは無視され0x00が4bitページ3-2に書き込まれる。それ以外はデータが4bitページ3-2に書き込まれる
  //    0x00c80001  データが4bitページ1-0に書き込まれる
  //    0x00d00000  060turboの060モードのときデータは無視され0x00が4bitページ3-2に書き込まれる。それ以外はデータが4bitページ3-2に書き込まれる
  //    0x00d00001  データが4bitページ1-0に書き込まれる
  //    0x00d80000  060turboの060モードのときデータは無視され0x00が4bitページ3-2に書き込まれる。それ以外はデータが4bitページ3-2に書き込まれる
  //    0x00d80001  データが4bitページ1-0に書き込まれる
  //
  //  ワードライト
  //    0x00c00000  上位バイトは無視され下位バイトが4bitページ3-2と4bitページ1-0の両方に書き込まれる
  //    0x00c80000  上位バイトは無視され下位バイトが4bitページ3-2と4bitページ1-0の両方に書き込まれる
  //    0x00d00000  上位バイトは無視され下位バイトが4bitページ3-2と4bitページ1-0の両方に書き込まれる
  //    0x00d80000  上位バイトは無視され下位バイトが4bitページ3-2と4bitページ1-0の両方に書き込まれる
  //
  MMD_GM2 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (メモリモード 2)" : "Graphics Screen (memory mode 2)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      if ((a & 0x00080001) == 0x00000001) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffffかつ下位バイト
        a &= 0x00c7fffe;
        return (byte) (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                       MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
      } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffffまたは上位バイト
        return 0x00;
      }
    }
    @Override protected int mmdPbz (int a) {
      if ((a & 0x00080001) == 0x00000001) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffffかつ下位バイト
        a &= 0x00c7fffe;
        return (byte) (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                       MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
      } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffffまたは上位バイト
        return 0x00;
      }
    }
    @Override protected int mmdPws (int a) {
      if ((a & 0x00080000) == 0x00000000) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffff
        a &= 0x00c7fffe;
        return (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
      } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffff
        a &= 0x00c7fffe;
        return (MainMemory.mmrM8[0x00180001 + a] << 4 |  //4bitページ3
                MainMemory.mmrM8[0x00100001 + a]);  //4bitページ2
      }
    }
    @Override protected int mmdPwz (int a) {
      if ((a & 0x00080000) == 0x00000000) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffff
        a &= 0x00c7fffe;
        return (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
      } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffff
        a &= 0x00c7fffe;
        return (MainMemory.mmrM8[0x00180001 + a] << 4 |  //4bitページ3
                MainMemory.mmrM8[0x00100001 + a]);  //4bitページ2
      }
    }
    @Override protected int mmdPls (int a) {
      a &= 0x00d7fffe;
      return (MainMemory.mmrM8[0x00080001 + a] << 20 |  //4bitページ1,3
              MainMemory.mmrM8[0x00000001 + a] << 16 |  //4bitページ0,2
              MainMemory.mmrM8[0x00080003 + a] <<  4 |  //4bitページ1,3
              MainMemory.mmrM8[0x00000003 + a]);  //4bitページ0,2
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       if ((a & 0x00080001) == 0x00000001) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffffかつ下位バイト
         a &= 0x00c7fffe;
         return (byte) (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                        MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
       } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffffまたは上位バイト
         return 0x00;
       }
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       if ((a & 0x00080001) == 0x00000001) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffffかつ下位バイト
         a &= 0x00c7fffe;
         return (byte) (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                        MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
       } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffffまたは上位バイト
         return 0x00;
       }
     }
    @Override protected int mmdRws (int a) throws M68kException {
       if ((a & 0x00080000) == 0x00000000) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffff
         a &= 0x00c7fffe;
         return (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                 MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
       } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffff
         a &= 0x00c7fffe;
         return (MainMemory.mmrM8[0x00180001 + a] << 4 |  //4bitページ3
                 MainMemory.mmrM8[0x00100001 + a]);  //4bitページ2
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       if ((a & 0x00080000) == 0x00000000) {  //0x00c00000..0x00c7ffff,0x00d00000..0x00d7ffff
         a &= 0x00c7fffe;
         return (MainMemory.mmrM8[0x00080001 + a] << 4 |  //4bitページ1
                 MainMemory.mmrM8[0x00000001 + a]);  //4bitページ0
       } else {  //0x00c80000..0x00cfffff,0x00d80000..0x00dfffff
         a &= 0x00c7fffe;
         return (MainMemory.mmrM8[0x00180001 + a] << 4 |  //4bitページ3
                 MainMemory.mmrM8[0x00100001 + a]);  //4bitページ2
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= 0x00d7fffe;
       return (MainMemory.mmrM8[0x00080001 + a] << 20 |  //4bitページ1,3
               MainMemory.mmrM8[0x00000001 + a] << 16 |  //4bitページ0,2
               MainMemory.mmrM8[0x00080003 + a] <<  4 |  //4bitページ1,3
               MainMemory.mmrM8[0x00000003 + a]);  //4bitページ0,2
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & 0x00000001) == 0x00000000) {  //上位バイト
         a &= 0x00c7fffe;
         if (XEiJ.mpuCoreType == 6) {
           MainMemory.mmrM8[0x00180001 + a] = 0;  //4bitページ3
           MainMemory.mmrM8[0x00100001 + a] = 0;  //4bitページ2
         } else {
           MainMemory.mmrM8[0x00180001 + a] = (byte) (d >>> 4 & 15);  //4bitページ3
           MainMemory.mmrM8[0x00100001 + a] = (byte) (d       & 15);  //4bitページ2
         }
         a >>>= 10;
         int y = a - CRTC.crtR13GrYCurr[3] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[2] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       } else {  //下位バイト
         a &= 0x00c7fffe;
         MainMemory.mmrM8[0x00080001 + a] = (byte) (d >>> 4 & 15);  //4bitページ1
         MainMemory.mmrM8[0x00000001 + a] = (byte) (d       & 15);  //4bitページ0
         a >>>= 10;
         int y = a - CRTC.crtR13GrYCurr[1] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[0] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= 0x00c7fffe;
       MainMemory.mmrM8[0x00180001 + a] = MainMemory.mmrM8[0x00080001 + a] = (byte) (d >>> 4 & 15);  //4bitページ1,3
       MainMemory.mmrM8[0x00100001 + a] = MainMemory.mmrM8[0x00000001 + a] = (byte) (d       & 15);  //4bitページ0,2
       a >>>= 10;
       int y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= 0x00c7fffe;
       MainMemory.mmrM8[0x00180001 + a] = MainMemory.mmrM8[0x00080001 + a] = (byte) (d >>> 20 & 15);  //ページ1,3
       MainMemory.mmrM8[0x00100001 + a] = MainMemory.mmrM8[0x00000001 + a] = (byte) (d >>> 16 & 15);  //ページ0,2
       MainMemory.mmrM8[0x00180003 + a] = MainMemory.mmrM8[0x00080003 + a] = (byte) (d >>>  4 & 15);  //ページ1,3
       MainMemory.mmrM8[0x00100003 + a] = MainMemory.mmrM8[0x00000003 + a] = (byte) (d        & 15);  //ページ0,2
       a >>>= 10;
       int y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GM2

  //--------------------------------------------------------------------------------
  //MMD_GG0 グラフィックス画面(512ドット65536色)
  //
  //  512ドット65536色
  //                アドレス             アクセス                  格納
  //    GG0  0x00c00000～0x00c7ffff  fedcba9876543210  ─┬  ............3210
  //         0x00c80000～0x00cfffff                      ├  ............7654
  //         0x00d00000～0x00d7ffff                      ├  ............ba98
  //         0x00d80000～0x00dfffff                      └  ............fedc
  //
  MMD_GG0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (512 ドット 65536 色)" : "Graphics Screen (512 dots 65536 colors)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      a += -(~a & 1) & 0x00100001;
      return (byte) (MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a]);
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      a += -(~a & 1) & 0x00100001;
      return MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a];
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return (short) (MainMemory.mmrM8[a + 0x00180001] << 12 | MainMemory.mmrM8[a + 0x00100001] << 8 | MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1]);
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a + 0x00180001] << 12 | MainMemory.mmrM8[a + 0x00100001] << 8 | MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return (MainMemory.mmrM8[a + 0x00180001] << 28 | MainMemory.mmrM8[a + 0x00100001] << 24 | MainMemory.mmrM8[a + 0x00080001] << 20 | MainMemory.mmrM8[a + 1] << 16 |
              MainMemory.mmrM8[a + 0x00180003] << 12 | MainMemory.mmrM8[a + 0x00100003] << 8 | MainMemory.mmrM8[a + 0x00080003] << 4 | MainMemory.mmrM8[a + 3]);
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       a += -(~a & 1) & 0x00100001;
       return (byte) (MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a]);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       a += -(~a & 1) & 0x00100001;
       return MainMemory.mmrM8[a + 0x00080000] << 4 | MainMemory.mmrM8[a];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (short) (MainMemory.mmrM8[a + 0x00180001] << 12 | MainMemory.mmrM8[a + 0x00100001] << 8 | MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1]);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a + 0x00180001] << 12 | MainMemory.mmrM8[a + 0x00100001] << 8 | MainMemory.mmrM8[a + 0x00080001] << 4 | MainMemory.mmrM8[a + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (MainMemory.mmrM8[a + 0x00180001] << 28 | MainMemory.mmrM8[a + 0x00100001] << 24 | MainMemory.mmrM8[a + 0x00080001] << 20 | MainMemory.mmrM8[a + 1] << 16 |
               MainMemory.mmrM8[a + 0x00180003] << 12 | MainMemory.mmrM8[a + 0x00100003] << 8 | MainMemory.mmrM8[a + 0x00080003] << 4 | MainMemory.mmrM8[a + 3]);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.TEST_BIT_0_SHIFT ? a << 31 - 0 >= 0 : (a & 1) == 0) {
         MainMemory.mmrM8[a + 0x00180001] = (byte) (d << 24  >>> 28);
         MainMemory.mmrM8[a + 0x00100001] = (byte) (d          & 15);
         a >>= 10;
         int y = a - CRTC.crtR13GrYCurr[2] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[3] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       } else {
         MainMemory.mmrM8[a + 0x00080000] = (byte) (d << 24  >>> 28);
         MainMemory.mmrM8[a             ] = (byte) (d          & 15);
         a >>= 10;
         int y = a - CRTC.crtR13GrYCurr[0] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
         y = a - CRTC.crtR13GrYCurr[1] & 511;
         CRTC.crtRasterStamp[y      ] = 0;
         CRTC.crtRasterStamp[y + 512] = 0;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00180001] = (byte) ((char) d >>> 12);
       MainMemory.mmrM8[a + 0x00100001] = (byte) (d << 20  >>> 28);
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00000001] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       MainMemory.mmrM8[a + 0x00180001] = (byte) (d        >>> 28);
       MainMemory.mmrM8[a + 0x00100001] = (byte) (d <<  4  >>> 28);
       MainMemory.mmrM8[a + 0x00080001] = (byte) (d <<  8  >>> 28);
       MainMemory.mmrM8[a + 0x00000001] = (byte) (d << 12  >>> 28);
       MainMemory.mmrM8[a + 0x00180003] = (byte) ((char) d >>> 12);
       MainMemory.mmrM8[a + 0x00100003] = (byte) (d << 20  >>> 28);
       MainMemory.mmrM8[a + 0x00080003] = (byte) (d << 24  >>> 28);
       MainMemory.mmrM8[a + 0x00000003] = (byte) (d          & 15);
       a >>= 10;
       int y = a - CRTC.crtR13GrYCurr[0] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[1] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[2] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
       y = a - CRTC.crtR13GrYCurr[3] & 511;
       CRTC.crtRasterStamp[y      ] = 0;
       CRTC.crtRasterStamp[y + 512] = 0;
     }
  },  //MMD_GG0

  //--------------------------------------------------------------------------------
  //MMD_GH0 グラフィックス画面(1024ドット16色)
  //
  //  参照アドレス
  //    0b00000000_110Yyyyy_yyyyyXxx_xxxxxxx?
  //    +--------------------------+--------------------------+
  //    | 0x00c00000 .. 0x00c003ff | 0x00c00400 .. 0x00c007ff |
  //    | 0x00c00800    0x00c00bff | 0x00c00c00    0x00c00fff |
  //    |      :             :     |      :             :     |
  //    | 0x00cff000    0x00cff3ff | 0x00cff400    0x00cff7ff |
  //    | 0x00cff800 .. 0x00cffbff | 0x00cffc00 .. 0x00cfffff |
  //    +--------------------------+--------------------------+
  //    | 0x00d00000 .. 0x00d003ff | 0x00d00400 .. 0x00d007ff |
  //    | 0x00d00800    0x00d00bff | 0x00d00c00    0x00d00fff |
  //    |      :             :     |      :             :     |
  //    | 0x00dff000    0x00dff3ff | 0x00dff400    0x00dff7ff |
  //    | 0x00dff800 .. 0x00dffbff | 0x00dffc00 .. 0x00dfffff |
  //    +--------------------------+--------------------------+
  //
  //  格納アドレス
  //    0b00000000_110YXyyy_yyyyyyxx_xxxxxxx?
  //    +--------------------------+--------------------------+
  //    | 0x00c00000 .. 0x00c003ff | 0x00c80000 .. 0x00c803ff |
  //    | 0x00c00400    0x00c007ff | 0x00c80400    0x00c807ff |
  //    |      :     G0      :     |      :     G1      :     |
  //    | 0x00c7f800    0x00c7fbff | 0x00cff800    0x00cffbff |
  //    | 0x00c7fc00 .. 0x00c7ffff | 0x00cffc00 .. 0x00cfffff |
  //    +--------------------------+--------------------------+
  //    | 0x00d00000 .. 0x00d003ff | 0x00d80000 .. 0x00d803ff |
  //    | 0x00d00400    0x00d007ff | 0x00d80400    0x00d807ff |
  //    |      :     G2      :     |      :     G3      :     |
  //    | 0x00d7f800    0x00d7fbff | 0x00dff800    0x00dffbff |
  //    | 0x00d7fc00 .. 0x00d7ffff | 0x00dffc00 .. 0x00dfffff |
  //    +--------------------------+--------------------------+
  //
  //    perl -e "@m=(0,1,2,3);sub f{my($a)=@_;my$x=$a>>1&1023;my$y=$a>>11&1023;($a&0x00c00001)+($m[($y>>8&2)+($x>>9&1)]<<19)+(($y&511)<<10)+(($x&511)<<1);}my@a=(0xc00000,0xc003ff,0xc00400,0xc007ff,0xc00800,0xc00bff,0xc00c00,0xc00fff,0xcff000,0xcff3ff,0xcff400,0xcff7ff,0xcff800,0xcffbff,0xcffc00,0xcfffff,0xd00000,0xd003ff,0xd00400,0xd007ff,0xd00800,0xd00bff,0xd00c00,0xd00fff,0xdff000,0xdff3ff,0xdff400,0xdff7ff,0xdff800,0xdffbff,0xdffc00,0xdfffff);for$v(0..7){print'  //    ';for$u(0..3){$a=$a[4*$v+$u];$b=f($a);printf'  0x%08x',$b;}print qq@\n@;}"
  //    perl -e "sub f{my($a)=@_;($a&0x00c00000+(1<<20)+1023)+($a<<9&1<<19)+($a>>1&511<<10);}my@a=(0xc00000,0xc003ff,0xc00400,0xc007ff,0xc00800,0xc00bff,0xc00c00,0xc00fff,0xcff000,0xcff3ff,0xcff400,0xcff7ff,0xcff800,0xcffbff,0xcffc00,0xcfffff,0xd00000,0xd003ff,0xd00400,0xd007ff,0xd00800,0xd00bff,0xd00c00,0xd00fff,0xdff000,0xdff3ff,0xdff400,0xdff7ff,0xdff800,0xdffbff,0xdffc00,0xdfffff);for$v(0..7){print'  //    ';for$u(0..3){$a=$a[4*$v+$u];$b=f($a);printf'  0x%08x',$b;}print qq@\n@;}"
  //      0x00c00000  0x00c003ff  0x00c80000  0x00c803ff
  //      0x00c00400  0x00c007ff  0x00c80400  0x00c807ff
  //      0x00c7f800  0x00c7fbff  0x00cff800  0x00cffbff
  //      0x00c7fc00  0x00c7ffff  0x00cffc00  0x00cfffff
  //      0x00d00000  0x00d003ff  0x00d80000  0x00d803ff
  //      0x00d00400  0x00d007ff  0x00d80400  0x00d807ff
  //      0x00d7f800  0x00d7fbff  0x00dff800  0x00dffbff
  //      0x00d7fc00  0x00d7ffff  0x00dffc00  0x00dfffff
  //
  MMD_GH0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (1024 ドット 16 色)" : "Graphics Screen (1024 dots 16 colors)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0];
    }
    @Override protected int mmdPbz (int a) {
      int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0];
    }
    @Override protected int mmdPws (int a) {
      int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0 + 1];
    }
    @Override protected int mmdPwz (int a) {
      int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0 + 1];
    }
    @Override protected int mmdPls (int a) {
      //4bitページを跨ぐ場合があることに注意
      int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      a += 2;
      int b2 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0 + 1] << 16 | MainMemory.mmrM8[b2 + 1];
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0];
     }
    @Override protected int mmdRws (int a) throws M68kException {
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0 + 1];
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0 + 1];
     }
    @Override protected int mmdRls (int a) throws M68kException {
       //4bitページを跨ぐ場合があることに注意
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0 + 1] << 16 | MainMemory.mmrM8[b2 + 1];
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & 1) != 0) {
         CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
         int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
         MainMemory.mmrM8[b0] = (byte) (d & 15);
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0 + 1] = (byte) (d & 15);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       //4bitページを跨ぐ場合があることに注意
       int b0 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = (a & 0x00c00000 + (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0 + 1] = (byte) (d >>> 16 & 15);
       MainMemory.mmrM8[b2 + 1] = (byte) (d & 15);
     }
  },  //MMD_GH0

  //--------------------------------------------------------------------------------
  //MMD_GI0 グラフィックス画面(1024ドット256色)
  //
  //  参照アドレスと格納アドレスはGH0と同じ
  //
  MMD_GI0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (1024 ドット 256 色)" : "Graphics Screen (1024 dots 256 colors)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0];
    }
    @Override protected int mmdPbz (int a) {
      int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0 + 1] & 0xff;
    }
    @Override protected int mmdPwz (int a) {
      int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0 + 1] & 0xff;
    }
    @Override protected int mmdPls (int a) {
      //4bitページを跨ぐ場合があることに注意
      int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      a += 2;
      int b2 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return (MainMemory.mmrM8[b0 + 1] & 0xff) << 16 | MainMemory.mmrM8[b2 + 1] & 0xff;
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0 + 1] & 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0 + 1] & 0xff;
     }
    @Override protected int mmdRls (int a) throws M68kException {
       //4bitページを跨ぐ場合があることに注意
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return (MainMemory.mmrM8[b0 + 1] & 0xff) << 16 | MainMemory.mmrM8[b2 + 1] & 0xff;
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & 1) != 0) {
         CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
         int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
         MainMemory.mmrM8[b0] = (byte) d;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0 + 1] = (byte) d;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       //4bitページを跨ぐ場合があることに注意
       int b0 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = XEiJ.BUS_MODIFIED_256_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0 + 1] = (byte) (d >>> 16);
       MainMemory.mmrM8[b2 + 1] = (byte) d;
     }
  },  //MMD_GI0

  //--------------------------------------------------------------------------------
  //MMD_GJ0 グラフィックス画面(1024ドット65536色)
  //
  //  参照アドレスと格納アドレスはGH0と同じ
  //
  MMD_GJ0 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "グラフィックス画面 (1024 ドット 65536 色)" : "Graphics Screen (1024 dots 65536 colors)";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0];
    }
    @Override protected int mmdPbz (int a) {
      int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return MainMemory.mmrM8[b0] << 8 | (MainMemory.mmrM8[b0 + 1] & 0xff);
    }
    @Override protected int mmdPwz (int a) {
      int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return (char) (MainMemory.mmrM8[b0] << 8 | (MainMemory.mmrM8[b0 + 1] & 0xff));
    }
    @Override protected int mmdPls (int a) {
       //4bitページを跨ぐ場合があることに注意
      int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      a += 2;
      int b2 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
      return (MainMemory.mmrM8[b0] << 24 | (MainMemory.mmrM8[b0 + 1] & 0xff) << 16 |
              (char) (MainMemory.mmrM8[b2] << 8 | (MainMemory.mmrM8[b2 + 1] & 0xff)));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       return MainMemory.mmrM8[XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10)];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       return MainMemory.mmrM8[XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10)] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return MainMemory.mmrM8[b0] << 8 | (MainMemory.mmrM8[b0 + 1] & 0xff);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return (char) (MainMemory.mmrM8[b0] << 8 | (MainMemory.mmrM8[b0 + 1] & 0xff));
     }
    @Override protected int mmdRls (int a) throws M68kException {
       //4bitページを跨ぐ場合があることに注意
       int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       return (MainMemory.mmrM8[b0] << 24 | (MainMemory.mmrM8[b0 + 1] & 0xff) << 16 |
               (char) (MainMemory.mmrM8[b2] << 8 | (MainMemory.mmrM8[b2 + 1] & 0xff)));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & 1) != 0) {
         CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
         MainMemory.mmrM8[XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10)] = (byte) d;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0] = (byte) (d >> 8);
       MainMemory.mmrM8[b0 + 1] = (byte) d;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       CRTC.crtRasterStamp[(a >> 11) - CRTC.crtR13GrYCurr[0] & 1023] = 0;
       //4bitページを跨ぐ場合があることに注意
       int b0 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       a += 2;
       int b2 = XEiJ.BUS_MODIFIED_65536_PAGE + (a & (1 << 20) + 1023) + (a << 9 & 1 << 19) + (a >>> 1 & 511 << 10);
       MainMemory.mmrM8[b0] = (byte) (d >>> 24);
       MainMemory.mmrM8[b0 + 1] = (byte) (d >>> 16);
       MainMemory.mmrM8[b2] = (byte) (d >>> 8);
       MainMemory.mmrM8[b2 + 1] = (byte) d;
     }
  },  //MMD_GJ0

  //--------------------------------------------------------------------------------
  //MMD_TXT テキスト画面
  MMD_TXT {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "テキスト画面" : "Text Screen";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getShort (a);
       } else {
         return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getChar (a);
       } else {
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getInt (a);
       } else {
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       int x;  //マスク
       if (CRTC.crtSimAccess) {  //同時アクセスあり
         a &= 0x00e1ffff;
         if (CRTC.crtBitMask) {  //同時アクセスあり,ビットマスクあり
           d &= ~(x = CRTC.crtR23Mask >> ((~a & 1) << 3));
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) (MainMemory.mmrM8[a             ] & x | d);
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) (MainMemory.mmrM8[a + 0x00020000] & x | d);
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) (MainMemory.mmrM8[a + 0x00040000] & x | d);
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) (MainMemory.mmrM8[a + 0x00060000] & x | d);
           }
         } else {  //同時アクセスあり,ビットマスクなし
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) d;
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) d;
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) d;
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) d;
           }
         }
       } else if (CRTC.crtBitMask) {  //同時アクセスなし,ビットマスクあり
         x = CRTC.crtR23Mask >> ((~a & 1) << 3);
         MainMemory.mmrM8[a] = (byte) (MainMemory.mmrM8[a] & x | d & ~x);
       } else {  //同時アクセスなし,ビットマスクなし
         MainMemory.mmrM8[a] = (byte) d;
       }
       CRTC.crtRasterStamp[((a & 0x0001ffff) >> 7) - CRTC.crtR11TxYCurr & 1023] = 0;  //同時アクセスやビットマスクで1ピクセルも書き換えなくても更新することになる
     }  //mmdWb
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       int e;  //上位バイトのデータ
       int x;  //下位バイトのマスク
       int y;  //上位バイトのマスク
       if (CRTC.crtSimAccess) {  //同時アクセスあり
         a &= 0x00e1ffff;
         if (CRTC.crtBitMask) {  //同時アクセスあり,ビットマスクあり
           e = d >> 8 & ~(y = (x = CRTC.crtR23Mask) >> 8);
           d &= ~x;
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) (MainMemory.mmrM8[a             ] & y | e);
             MainMemory.mmrM8[a + 0x00000001] = (byte) (MainMemory.mmrM8[a + 0x00000001] & x | d);
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) (MainMemory.mmrM8[a + 0x00020000] & y | e);
             MainMemory.mmrM8[a + 0x00020001] = (byte) (MainMemory.mmrM8[a + 0x00020001] & x | d);
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) (MainMemory.mmrM8[a + 0x00040000] & y | e);
             MainMemory.mmrM8[a + 0x00040001] = (byte) (MainMemory.mmrM8[a + 0x00040001] & x | d);
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) (MainMemory.mmrM8[a + 0x00060000] & y | e);
             MainMemory.mmrM8[a + 0x00060001] = (byte) (MainMemory.mmrM8[a + 0x00060001] & x | d);
           }
         } else {  //同時アクセスあり,ビットマスクなし
           e = d >> 8;
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) e;
             MainMemory.mmrM8[a + 0x00000001] = (byte) d;
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) e;
             MainMemory.mmrM8[a + 0x00020001] = (byte) d;
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) e;
             MainMemory.mmrM8[a + 0x00040001] = (byte) d;
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) e;
             MainMemory.mmrM8[a + 0x00060001] = (byte) d;
           }
         }
       } else if (CRTC.crtBitMask) {  //同時アクセスなし,ビットマスクあり
         y = (x = CRTC.crtR23Mask) >> 8;
         MainMemory.mmrM8[a    ] = (byte) (MainMemory.mmrM8[a    ] & y | (d >> 8) & ~y);
         MainMemory.mmrM8[a + 1] = (byte) (MainMemory.mmrM8[a + 1] & x |  d       & ~x);
       } else {  //同時アクセスなし,ビットマスクなし
         if (MainMemory.MMR_USE_BYTE_BUFFER) {
           MainMemory.mmrBuffer.putShort (a, (short) d);
         } else {
           MainMemory.mmrM8[a    ] = (byte) (d >> 8);
           MainMemory.mmrM8[a + 1] = (byte)  d;
         }
       }
       CRTC.crtRasterStamp[((a & 0x0001ffff) >> 7) - CRTC.crtR11TxYCurr & 1023] = 0;  //同時アクセスやビットマスクで1ピクセルも書き換えなくても更新することになる
     }  //mmdWw
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       int e;  //下位ワードの上位バイトのデータ
       int f;  //上位ワードの下位バイトのデータ
       int g;  //上位ワードの上位バイトのデータ
       int x;  //下位バイトのマスク
       int y;  //上位バイトのマスク
       if (CRTC.crtSimAccess) {  //同時アクセスあり
         a &= 0x00e1ffff;
         if (CRTC.crtBitMask) {  //同時アクセスあり,ビットマスクあり
           g = d >> 24 & ~(y = (x = CRTC.crtR23Mask) >> 8);
           f = d >> 16 & ~x;
           e = d >>  8 & ~y;
           d &= ~x;
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) (MainMemory.mmrM8[a             ] & y | g);
             MainMemory.mmrM8[a + 0x00000001] = (byte) (MainMemory.mmrM8[a + 0x00000001] & x | f);
             MainMemory.mmrM8[a + 0x00000002] = (byte) (MainMemory.mmrM8[a + 0x00000002] & y | e);
             MainMemory.mmrM8[a + 0x00000003] = (byte) (MainMemory.mmrM8[a + 0x00000003] & x | d);
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) (MainMemory.mmrM8[a + 0x00020000] & y | g);
             MainMemory.mmrM8[a + 0x00020001] = (byte) (MainMemory.mmrM8[a + 0x00020001] & x | f);
             MainMemory.mmrM8[a + 0x00020002] = (byte) (MainMemory.mmrM8[a + 0x00020002] & y | e);
             MainMemory.mmrM8[a + 0x00020003] = (byte) (MainMemory.mmrM8[a + 0x00020003] & x | d);
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) (MainMemory.mmrM8[a + 0x00040000] & y | g);
             MainMemory.mmrM8[a + 0x00040001] = (byte) (MainMemory.mmrM8[a + 0x00040001] & x | f);
             MainMemory.mmrM8[a + 0x00040002] = (byte) (MainMemory.mmrM8[a + 0x00040002] & y | e);
             MainMemory.mmrM8[a + 0x00040003] = (byte) (MainMemory.mmrM8[a + 0x00040003] & x | d);
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) (MainMemory.mmrM8[a + 0x00060000] & y | g);
             MainMemory.mmrM8[a + 0x00060001] = (byte) (MainMemory.mmrM8[a + 0x00060001] & x | f);
             MainMemory.mmrM8[a + 0x00060002] = (byte) (MainMemory.mmrM8[a + 0x00060002] & y | e);
             MainMemory.mmrM8[a + 0x00060003] = (byte) (MainMemory.mmrM8[a + 0x00060003] & x | d);
           }
         } else {  //同時アクセスあり,ビットマスクなし
           g = d >> 24;
           f = d >> 16;
           e = d >>  8;
           if (CRTC.crtSimPlane0) {
             MainMemory.mmrM8[a             ] = (byte) g;
             MainMemory.mmrM8[a + 0x00000001] = (byte) f;
             MainMemory.mmrM8[a + 0x00000002] = (byte) e;
             MainMemory.mmrM8[a + 0x00000003] = (byte) d;
           }
           if (CRTC.crtSimPlane1) {
             MainMemory.mmrM8[a + 0x00020000] = (byte) g;
             MainMemory.mmrM8[a + 0x00020001] = (byte) f;
             MainMemory.mmrM8[a + 0x00020002] = (byte) e;
             MainMemory.mmrM8[a + 0x00020003] = (byte) d;
           }
           if (CRTC.crtSimPlane2) {
             MainMemory.mmrM8[a + 0x00040000] = (byte) g;
             MainMemory.mmrM8[a + 0x00040001] = (byte) f;
             MainMemory.mmrM8[a + 0x00040002] = (byte) e;
             MainMemory.mmrM8[a + 0x00040003] = (byte) d;
           }
           if (CRTC.crtSimPlane3) {
             MainMemory.mmrM8[a + 0x00060000] = (byte) g;
             MainMemory.mmrM8[a + 0x00060001] = (byte) f;
             MainMemory.mmrM8[a + 0x00060002] = (byte) e;
             MainMemory.mmrM8[a + 0x00060003] = (byte) d;
           }
         }
       } else if (CRTC.crtBitMask) {  //同時アクセスなし,ビットマスクあり
         y = (x = CRTC.crtR23Mask) >> 8;
         MainMemory.mmrM8[a    ] = (byte) (MainMemory.mmrM8[a    ] & y | (d >> 24) & ~y);
         MainMemory.mmrM8[a + 1] = (byte) (MainMemory.mmrM8[a + 1] & x | (d >> 16) & ~x);
         MainMemory.mmrM8[a + 2] = (byte) (MainMemory.mmrM8[a + 1] & y | (d >>  8) & ~y);
         MainMemory.mmrM8[a + 3] = (byte) (MainMemory.mmrM8[a + 1] & x |  d        & ~x);
       } else {  //同時アクセスなし,ビットマスクなし
         if (MainMemory.MMR_USE_BYTE_BUFFER) {
           MainMemory.mmrBuffer.putInt (a, d);
         } else {
           MainMemory.mmrM8[a    ] = (byte) (d >> 24);
           MainMemory.mmrM8[a + 1] = (byte) (d >> 16);
           MainMemory.mmrM8[a + 2] = (byte) (d >>  8);
           MainMemory.mmrM8[a + 3] = (byte)  d;
         }
       }
       CRTC.crtRasterStamp[((a     & 0x0001ffff) >> 7) - CRTC.crtR11TxYCurr & 1023] = 0;  //同時アクセスやビットマスクで1ピクセルも書き換えなくても更新することになる
       CRTC.crtRasterStamp[((a + 2 & 0x0001ffff) >> 7) - CRTC.crtR11TxYCurr & 1023] = 0;  //同時アクセスやビットマスクで1ピクセルも書き換えなくても更新することになる
     }  //mmdWl
  },  //MMD_TXT

  //--------------------------------------------------------------------------------
  //MMD_CRT CRTコントローラ
  //
  //  $00E80000～$00E8002F  ワードレジスタ
  //  $00E80030～$00E8003F  $0000
  //  $00E80040～$00E803FF  $00E80000～$00E8003Fの繰り返し
  //  $00E80400～$00E8047F  バスエラー
  //  $00E80480～$00E80481  ワードレジスタ
  //  $00E80482～$00E804FF  $00E80480～$00E80481の繰り返し
  //  $00E80500～$00E807FF  $00E80400～$00E804FFの繰り返し
  //  $00E80800～$00E81FFF  $00E80000～$00E807FFの繰り返し
  //
  MMD_CRT {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "CRT コントローラ" : "CRT Controller";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       int aa = a & 0x07ff;  //$00E80800～$00E81FFF→$00E80000～$00E807FF
       if (aa < 0x0400) {  //$00E80000～$00E803FF
         aa &= 0x003f;
         switch (aa) {
           //case 0x0001:  //R00の下位
           //  return CRTC.crtR00HFrontEndPort;
           //case 0x0003:  //R01の下位
           //  return CRTC.crtR01HSyncEndPort;
           //case 0x0005:  //R02の下位
           //  return CRTC.crtR02HBackEndPort;
           //case 0x0007:  //R03の下位
           //  return CRTC.crtR03HDispEndPort;
           //case 0x0008:  //R04の上位
           //  return CRTC.crtR04VFrontEndPort >> 8;
           //case 0x0009:  //R04の下位
           //  return CRTC.crtR04VFrontEndPort & 0xff;
           //case 0x000a:  //R05の上位
           //  return CRTC.crtR05VSyncEndPort >> 8;
           //case 0x000b:  //R05の下位
           //  return CRTC.crtR05VSyncEndPort & 0xff;
           //case 0x000c:  //R06の上位
           //  return CRTC.crtR06VBackEndPort >> 8;
           //case 0x000d:  //R06の下位
           //  return CRTC.crtR06VBackEndPort & 0xff;
           //case 0x000e:  //R07の上位
           //  return CRTC.crtR07VDispEndPort >> 8;
           //case 0x000f:  //R07の下位
           //  return CRTC.crtR07VDispEndPort & 0xff;
           //case 0x0011:  //R08の下位
           //  return CRTC.crtR08Adjust;
           //case 0x0012:  //R09の上位
           //  return CRTC.crtR09IRQRasterPort >> 8;
           //case 0x0013:  //R09の下位
           //  return CRTC.crtR09IRQRasterPort & 0xff;
           //case 0x0014:  //R10の上位
           //  return CRTC.crtR10TxXPort >> 8;
           //case 0x0015:  //R10の下位
           //  return CRTC.crtR10TxXPort & 0xff;
           //case 0x0016:  //R11の上位
           //  return CRTC.crtR11TxYCurr >> 8;
           //case 0x0017:  //R11の下位
           //  return CRTC.crtR11TxYCurr & 0xff;
           //case 0x0018:  //R12の上位
           //case 0x001c:  //R14の上位
           //case 0x0020:  //R16の上位
           //case 0x0024:  //R18の上位
           //  return CRTC.crtR12GrXPort[(aa - 0x0018) >> 2] >> 8;
           //case 0x0019:  //R12の下位
           //case 0x001d:  //R14の下位
           //case 0x0021:  //R16の下位
           //case 0x0025:  //R18の下位
           //  return CRTC.crtR12GrXPort[(aa - 0x0018) >> 2] & 0xff;
           //case 0x001a:  //R13の上位
           //case 0x001e:  //R15の上位
           //case 0x0022:  //R17の上位
           //case 0x0026:  //R19の上位
           //  return CRTC.crtR13GrYPort[(aa - 0x0018) >> 2] >> 8;
           //case 0x001b:  //R13の下位
           //case 0x001f:  //R15の下位
           //case 0x0023:  //R17の下位
           //case 0x0027:  //R19の下位
           //  return CRTC.crtR13GrYPort[(aa - 0x0018) >> 2] & 0xff;
         case 0x0028:  //R20の上位
           return (CRTC.crtTextStorage    << 4 |
                   CRTC.crtGraphicStorage << 3 |
                   CRTC.crtMemoryModePort);
         case 0x0029:  //R20の下位
           return (CRTC.crtHighResoPort << 4 |
                   CRTC.crtVResoPort    << 2 |
                   CRTC.crtHResoPort);
         case 0x002a:  //R21の上位
           return ((CRTC.crtBitMask   ? 0b00000010 : 0) |
                   (CRTC.crtSimAccess ? 0b00000001 : 0));
         case 0x002b:  //R21の下位
           return ((CRTC.crtSimPlane3 ? 0b10000000 : 0) |
                   (CRTC.crtSimPlane2 ? 0b01000000 : 0) |
                   (CRTC.crtSimPlane1 ? 0b00100000 : 0) |
                   (CRTC.crtSimPlane0 ? 0b00010000 : 0) |
                   (CRTC.crtCCPlane3  ? 0b00001000 : 0) |
                   (CRTC.crtCCPlane2  ? 0b00000100 : 0) |
                   (CRTC.crtCCPlane1  ? 0b00000010 : 0) |
                   (CRTC.crtCCPlane0  ? 0b00000001 : 0));
           //case 0x002c:  //R22の上位
           //  return CRTC.crtR22SrcBlock;
           //case 0x002d:  //R22の下位
           //  return CRTC.crtR22DstBlock;
           //case 0x002e:  //R23の上位
           //  return CRTC.crtR23Mask >> 8;
           //case 0x002f:  //R23の下位
           //  return CRTC.crtR23Mask & 0xff;
           //case 0x0000:  //R00の上位
           //case 0x0002:  //R01の上位
           //case 0x0004:  //R02の上位
           //case 0x0006:  //R03の上位
           //case 0x0010:  //R08の上位
           //case 0x0030:  //R24の上位
           //case 0x0031:  //R24の下位
           //case 0x0032:  //R25の上位
           //case 0x0033:  //R25の下位
           //case 0x0034:  //R26の上位
           //case 0x0035:  //R26の下位
           //case 0x0036:  //R27の上位
           //case 0x0037:  //R27の下位
           //case 0x0038:  //R28の上位
           //case 0x0039:  //R28の下位
           //case 0x003a:  //R29の上位
           //case 0x003b:  //R29の下位
           //case 0x003c:  //R30の上位
           //case 0x003d:  //R30の下位
           //case 0x003e:  //R31の上位
           //case 0x003f:  //R31の下位
         default:
           return 0x00;
           //return VideoController.vcnMode.ordinal () >> 8;
           //return VideoController.vcnMode.ordinal () & 0xff;
         }
       } else {  //$00E80400～$00E807FF
         aa &= 0xff;  //$00E80500～$00E807FF→$00E80400～$00E804FF
         if (aa < 0x80) {  //$00E80400～$00E8047F
           return super.mmdRbz (a);  //バスエラー
         } else {  //$00E80480～$00E804FF
           aa &= 0x01;  //$00E80482～$00E804FF→$00E80480～$00E80481
           if (aa == 0) {  //動作ポートの上位
             return 0;
           } else {  //動作ポートの下位
             return ((CRTC.crtRasterCopyOn ? 8 : 0) |  //ラスタコピー
                     (CRTC.crtClearStandby || CRTC.crtClearFrames != 0 ? 2 : 0));  //高速クリア
           }
         }
       }
     }  //mmdRbz
    @Override protected int mmdRwz (int a) throws M68kException {
       int aa = a & 0x07ff;  //$00E80800～$00E81FFF→$00E80000～$00E807FF
       if (aa < 0x0400) {  //$00E80000～$00E803FF
         aa &= 0x003f;
         switch (aa) {
           //case 0x0000:  //R00
           //  return CRTC.crtR00HFrontEndPort;
           //case 0x0002:  //R01
           //  return CRTC.crtR01HSyncEndPort;
           //case 0x0004:  //R02
           //  return CRTC.crtR02HBackEndPort;
           //case 0x0006:  //R03
           //  return CRTC.crtR03HDispEndPort;
           //case 0x0008:  //R04
           //  return CRTC.crtR04VFrontEndPort;
           //case 0x000a:  //R05
           //  return CRTC.crtR05VSyncEndPort;
           //case 0x000c:  //R06
           //  return CRTC.crtR06VBackEndPort;
           //case 0x000e:  //R07
           //  return CRTC.crtR07VDispEndPort;
           //case 0x0010:  //R08
           //  return CRTC.crtR08Adjust;
           //case 0x0012:  //R09
           //  return CRTC.crtR09IRQRasterPort;
           //case 0x0014:  //R10
           //  return CRTC.crtR10TxXPort;
           //case 0x0016:  //R11
           //  return CRTC.crtR11TxYPort;
           //case 0x0018:  //R12
           //case 0x001c:  //R14
           //case 0x0020:  //R16
           //case 0x0024:  //R18
           //  return CRTC.crtR12GrXPort[(aa - 0x0018) >> 2];
           //case 0x001a:  //R13
           //case 0x001e:  //R15
           //case 0x0022:  //R17
           //case 0x0026:  //R19
           //  return CRTC.crtR13GrYPort[(aa - 0x0018) >> 2];
         case 0x0028:  //R20
           return (CRTC.crtTextStorage    << 12 |
                   CRTC.crtGraphicStorage << 11 |
                   CRTC.crtMemoryModePort <<  8 |
                   CRTC.crtHighResoPort   <<  4 |
                   CRTC.crtVResoPort      <<  2 |
                   CRTC.crtHResoPort);
         case 0x002a:  //R21
           return ((CRTC.crtBitMask   ? 0b00000010_00000000 : 0) |
                   (CRTC.crtSimAccess ? 0b00000001_00000000 : 0) |
                   (CRTC.crtSimPlane3 ? 0b00000000_10000000 : 0) |
                   (CRTC.crtSimPlane2 ? 0b00000000_01000000 : 0) |
                   (CRTC.crtSimPlane1 ? 0b00000000_00100000 : 0) |
                   (CRTC.crtSimPlane0 ? 0b00000000_00010000 : 0) |
                   (CRTC.crtCCPlane3  ? 0b00000000_00001000 : 0) |
                   (CRTC.crtCCPlane2  ? 0b00000000_00000100 : 0) |
                   (CRTC.crtCCPlane1  ? 0b00000000_00000010 : 0) |
                   (CRTC.crtCCPlane0  ? 0b00000000_00000001 : 0));
           //case 0x002c:  //R22
           //  return CRTC.crtR22SrcBlock << 8 | CRTC.crtR22DstBlock;
           //case 0x002e:  //R23
           //  return CRTC.crtR23Mask;
           //case 0x0030:  //R24
           //case 0x0032:  //R25
           //case 0x0034:  //R26
           //case 0x0036:  //R27
           //case 0x0038:  //R28
           //case 0x003a:  //R29
           //case 0x003c:  //R30
           //case 0x003e:  //R31
         default:
           return 0x0000;
         }
       } else {  //$00E80400～$00E807FF
         aa &= 0xff;  //$00E80500～$00E807FF→$00E80400～$00E804FF
         if (aa < 0x80) {  //$00E80400～$00E8047F
           return super.mmdRbz (a);  //バスエラー
         } else {  //$00E80480～$00E804FF  動作ポート
           return ((CRTC.crtRasterCopyOn ? 8 : 0) |  //ラスタコピー
                   (CRTC.crtClearStandby || CRTC.crtClearFrames != 0 ? 2 : 0));  //高速クリア
         }
       }
     }  //mmdRwz
    @Override protected int mmdRls (int a) throws M68kException {
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       int aa = a & 0x07ff;  //$00E80800～$00E81FFF→$00E80000～$00E807FF
       if (aa < 0x0400) {  //$00E80000～$00E803FF
         aa &= 0x003f;
         switch (aa) {
         case 0x0001:  //R00の下位
           CRTC.crtR00HFrontEndPort = d & 0xff;
           {
             int curr = CRTC.crtR00HFrontEndMask == 0 ? CRTC.crtR00HFrontEndPort : CRTC.crtR00HFrontEndTest;
             if (CRTC.crtR00HFrontEndCurr != curr) {
               CRTC.crtR00HFrontEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
         return;
         case 0x0003:  //R01の下位
           CRTC.crtR01HSyncEndPort = d & 0xff;
           {
             int curr = CRTC.crtR01HSyncEndMask == 0 ? CRTC.crtR01HSyncEndPort : CRTC.crtR01HSyncEndTest;
             if (CRTC.crtR01HSyncEndCurr != curr) {
               CRTC.crtR01HSyncEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0005:  //R02の下位
           CRTC.crtR02HBackEndPort = d & 0xff;
           {
             int curr = CRTC.crtR02HBackEndMask == 0 ? CRTC.crtR02HBackEndPort : CRTC.crtR02HBackEndTest;
             if (CRTC.crtR02HBackEndCurr != curr) {
               CRTC.crtR02HBackEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0007:  //R03の下位
           CRTC.crtR03HDispEndPort = d & 0xff;
           {
             int curr = CRTC.crtR03HDispEndMask == 0 ? CRTC.crtR03HDispEndPort : CRTC.crtR03HDispEndTest;
             if (CRTC.crtR03HDispEndCurr != curr) {
               CRTC.crtR03HDispEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0008:  //R04の上位
           CRTC.crtR04VFrontEndPort = (d & 0x03) << 8 | (CRTC.crtR04VFrontEndPort & 0xff);
           {
             int curr = CRTC.crtR04VFrontEndMask == 0 ? CRTC.crtR04VFrontEndPort : CRTC.crtR04VFrontEndTest;
             if (CRTC.crtR04VFrontEndCurr != curr) {
               CRTC.crtR04VFrontEndCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0009:  //R04の下位
           CRTC.crtR04VFrontEndPort = (CRTC.crtR04VFrontEndPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR04VFrontEndMask == 0 ? CRTC.crtR04VFrontEndPort : CRTC.crtR04VFrontEndTest;
             if (CRTC.crtR04VFrontEndCurr != curr) {
               CRTC.crtR04VFrontEndCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000a:  //R05の上位
           CRTC.crtR05VSyncEndPort = (d & 0x03) << 8 | (CRTC.crtR05VSyncEndPort & 0xff);
           {
             int curr = CRTC.crtR05VSyncEndMask == 0 ? CRTC.crtR05VSyncEndPort : CRTC.crtR05VSyncEndTest;
             if (CRTC.crtR05VSyncEndCurr != curr) {
               CRTC.crtR05VSyncEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000b:  //R05の下位
           CRTC.crtR05VSyncEndPort = (CRTC.crtR05VSyncEndPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR05VSyncEndMask == 0 ? CRTC.crtR05VSyncEndPort : CRTC.crtR05VSyncEndTest;
             if (CRTC.crtR05VSyncEndCurr != curr) {
               CRTC.crtR05VSyncEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000c:  //R06の上位
           CRTC.crtR06VBackEndPort = (d & 0x03) << 8 | (CRTC.crtR06VBackEndPort & 0xff);
           {
             int curr = CRTC.crtR06VBackEndMask == 0 ? CRTC.crtR06VBackEndPort : CRTC.crtR06VBackEndTest;
             if (CRTC.crtR06VBackEndCurr != curr) {
               CRTC.crtR06VBackEndCurr = curr;
               CRTC.crtVDispStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000d:  //R06の下位
           CRTC.crtR06VBackEndPort = (CRTC.crtR06VBackEndPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR06VBackEndMask == 0 ? CRTC.crtR06VBackEndPort : CRTC.crtR06VBackEndTest;
             if (CRTC.crtR06VBackEndCurr != curr) {
               CRTC.crtR06VBackEndCurr = curr;
               CRTC.crtVDispStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000e:  //R07の上位
           CRTC.crtR07VDispEndPort = (d & 0x03) << 8 | (CRTC.crtR07VDispEndPort & 0xff);
           {
             int curr = CRTC.crtR07VDispEndMask == 0 ? CRTC.crtR07VDispEndPort : CRTC.crtR07VDispEndTest;
             if (CRTC.crtR07VDispEndCurr != curr) {
               CRTC.crtR07VDispEndCurr = curr;
               CRTC.crtVIdleStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000f:  //R07の下位
           CRTC.crtR07VDispEndPort = (CRTC.crtR07VDispEndPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR07VDispEndMask == 0 ? CRTC.crtR07VDispEndPort : CRTC.crtR07VDispEndTest;
             if (CRTC.crtR07VDispEndCurr != curr) {
               CRTC.crtR07VDispEndCurr = curr;
               CRTC.crtVIdleStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0011:  //R08の下位
           d &= 0xff;
           if (CRTC.crtR08Adjust != d) {
             CRTC.crtR08Adjust = d;
             CRTC.crtRestart ();
           }
           return;
         case 0x0012:  //R09の上位
           CRTC.crtR09IRQRasterPort = (d & 0x03) << 8 | (CRTC.crtR09IRQRasterPort & 0xff);
           {
             int curr = CRTC.crtR09IRQRasterMask == 0 ? CRTC.crtR09IRQRasterPort : CRTC.crtR09IRQRasterTest;
             if (CRTC.crtR09IRQRasterCurr != curr) {
               CRTC.crtR09IRQRasterCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               if (RasterBreakPoint.RBP_ON) {
                 RasterBreakPoint.rbpCheckIRQ ();
               }
             }
           }
           return;
         case 0x0013:  //R09の下位
           CRTC.crtR09IRQRasterPort = (CRTC.crtR09IRQRasterPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR09IRQRasterMask == 0 ? CRTC.crtR09IRQRasterPort : CRTC.crtR09IRQRasterTest;
             if (CRTC.crtR09IRQRasterCurr != curr) {
               CRTC.crtR09IRQRasterCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               if (RasterBreakPoint.RBP_ON) {
                 RasterBreakPoint.rbpCheckIRQ ();
               }
             }
           }
           return;
         case 0x0014:  //R10の上位
           CRTC.crtR10TxXPort = (d & 0x03) << 8 | (CRTC.crtR10TxXPort & 0xff);
           {
             int curr = CRTC.crtR10TxXMask == 0 ? CRTC.crtR10TxXPort : CRTC.crtR10TxXTest;
             if (CRTC.crtR10TxXCurr != curr) {
               CRTC.crtR10TxXCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0015:  //R10の下位
           CRTC.crtR10TxXPort = (CRTC.crtR10TxXPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR10TxXMask == 0 ? CRTC.crtR10TxXPort : CRTC.crtR10TxXTest;
             if (CRTC.crtR10TxXCurr != curr) {
               CRTC.crtR10TxXCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0016:  //R11の上位
           CRTC.crtR11TxYPort = (d & 0x03) << 8 | (CRTC.crtR11TxYPort & 0xff);
           {
             int curr = CRTC.crtR11TxYMask == 0 ? CRTC.crtR11TxYPort : CRTC.crtR11TxYTest;
             if (CRTC.crtR11TxYCurr != curr) {
               CRTC.crtR11TxYCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0017:  //R11の下位
           CRTC.crtR11TxYPort = (CRTC.crtR11TxYPort & ~0xff) | (d & 0xff);
           {
             int curr = CRTC.crtR11TxYMask == 0 ? CRTC.crtR11TxYPort : CRTC.crtR11TxYTest;
             if (CRTC.crtR11TxYCurr != curr) {
               CRTC.crtR11TxYCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0018:  //R12の上位
         case 0x001c:  //R14の上位
         case 0x0020:  //R16の上位
         case 0x0024:  //R18の上位
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR12GrXPort[p] = (d & (p == 0 ? 0x03 : 0x01)) << 8 | (CRTC.crtR12GrXPort[p] & 0xff);
             int curr = CRTC.crtR12GrXMask[p] == 0 ? CRTC.crtR12GrXPort[p] : CRTC.crtR12GrXTest[p];
             if (CRTC.crtR12GrXCurr[p] != curr) {
               CRTC.crtR12GrXCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0019:  //R12の下位
         case 0x001d:  //R14の下位
         case 0x0021:  //R16の下位
         case 0x0025:  //R18の下位
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR12GrXPort[p] = (CRTC.crtR12GrXPort[p] & ~0xff) | (d & 0xff);
             int curr = CRTC.crtR12GrXMask[p] == 0 ? CRTC.crtR12GrXPort[p] : CRTC.crtR12GrXTest[p];
             if (CRTC.crtR12GrXCurr[p] != curr) {
               CRTC.crtR12GrXCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x001a:  //R13の上位
         case 0x001e:  //R15の上位
         case 0x0022:  //R17の上位
         case 0x0026:  //R19の上位
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR13GrYPort[p] = (d & (p == 0 ? 0x03 : 0x01)) << 8 | (CRTC.crtR13GrYPort[p] & 0xff);
             int curr = CRTC.crtR13GrYMask[p] == 0 ? CRTC.crtR13GrYPort[p] : CRTC.crtR13GrYTest[p];
             if (CRTC.crtR13GrYCurr[p] != curr) {
               CRTC.crtR13GrYCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x001b:  //R13の下位
         case 0x001f:  //R15の下位
         case 0x0023:  //R17の下位
         case 0x0027:  //R19の下位
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR13GrYPort[p] = (CRTC.crtR13GrYPort[p] & ~0xff) | (d & 0xff);
             int curr = CRTC.crtR13GrYMask[p] == 0 ? CRTC.crtR13GrYPort[p] : CRTC.crtR13GrYTest[p];
             if (CRTC.crtR13GrYCurr[p] != curr) {
               CRTC.crtR13GrYCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0028:  //R20の上位
           CRTC.crtSetTextStorage (d >> 4);
           CRTC.crtSetGraphicStorage (d >> 3);
           CRTC.crtSetMemoryMode (d);
           return;
         case 0x0029:  //R20の下位
           CRTC.crtHighResoPort = d >>> 4 & 1;
           CRTC.crtVResoPort    = d >>> 2 & 3;
           CRTC.crtHResoPort    = d       & 3;
           SpriteScreen.sprAccessible = (d & 0b10010) != 0b10010;
           int highResoCurr = CRTC.crtHighResoMask == 0 ? CRTC.crtHighResoPort : CRTC.crtHighResoTest;
           int vResoCurr = CRTC.crtVResoMask == 0 ? CRTC.crtVResoPort : CRTC.crtVResoTest;
           int hResoCurr = CRTC.crtHResoMask == 0 ? CRTC.crtHResoPort : CRTC.crtHResoTest;
           if (CRTC.crtHighResoCurr != highResoCurr ||
               CRTC.crtVResoCurr != vResoCurr ||
               CRTC.crtHResoCurr != hResoCurr) {
             CRTC.crtHighResoCurr = highResoCurr;
             CRTC.crtVResoCurr = vResoCurr;
             CRTC.crtHResoCurr = hResoCurr;
             CRTC.crtRestart ();
           }
           return;
         case 0x002a:  //R21の上位
           CRTC.crtBitMask   = XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 < 0 : (d & 2) != 0;
           CRTC.crtSimAccess = XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 < 0 : (d & 1) != 0;
           return;
         case 0x002b:  //R21の下位
           CRTC.crtSimPlane3 = (byte) d < 0;  //(d & 128) != 0。d << 24 < 0
           CRTC.crtSimPlane2 = d << 25 < 0;  //(d & 64) != 0
           CRTC.crtSimPlane1 = d << 26 < 0;  //(d & 32) != 0
           CRTC.crtSimPlane0 = d << 27 < 0;  //(d & 16) != 0
           CRTC.crtCCPlane3  = XEiJ.TEST_BIT_3_SHIFT ? d << 31 - 3 < 0 : (d & 8) != 0;
           CRTC.crtCCPlane2  = XEiJ.TEST_BIT_2_SHIFT ? d << 31 - 2 < 0 : (d & 4) != 0;
           CRTC.crtCCPlane1  = XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 < 0 : (d & 2) != 0;
           CRTC.crtCCPlane0  = XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 < 0 : (d & 1) != 0;
           return;
         case 0x002c:  //R22の上位
           CRTC.crtR22SrcBlock = d & 0xff;
           return;
         case 0x002d:  //R22の下位
           CRTC.crtR22DstBlock = d & 0xff;
           return;
         case 0x002e:  //R23の上位
           CRTC.crtR23Mask = (d & 0xff) << 8 | (CRTC.crtR23Mask & 0xff);
           return;
         case 0x002f:  //R23の下位
           CRTC.crtR23Mask = (CRTC.crtR23Mask & ~0xff) | (d & 0xff);
           return;
           //case 0x0000:  //R00の上位
           //case 0x0002:  //R01の上位
           //case 0x0004:  //R02の上位
           //case 0x0006:  //R03の上位
           //case 0x0010:  //R08の上位
           //case 0x0030:  //R24の上位
           //case 0x0031:  //R24の下位
           //case 0x0032:  //R25の上位
           //case 0x0033:  //R25の下位
           //case 0x0034:  //R26の上位
           //case 0x0035:  //R26の下位
           //case 0x0036:  //R27の上位
           //case 0x0037:  //R27の下位
           //case 0x0038:  //R28の上位
           //case 0x0039:  //R28の下位
           //case 0x003a:  //R29の上位
           //case 0x003b:  //R29の下位
           //case 0x003c:  //R30の上位
           //case 0x003d:  //R30の下位
           //case 0x003e:  //R31の上位
           //case 0x003f:  //R31の下位
         default:
           return;
         }
       } else {  //$00E80400～$00E807FF
         aa &= 0xff;  //$00E80500～$00E807FF→$00E80400～$00E804FF
         if (aa < 0x80) {  //$00E80400～$00E8047F
           super.mmdWb (a, d);  //バスエラー
         } else {  //$00E80480～$00E804FF
           aa &= 0x01;  //$00E80482～$00E804FF→$00E80480～$00E80481
           if (aa == 0) {  //動作ポートの上位
             return;
           } else {  //動作ポートの下位
             boolean rasterCopyOn = (d & 8) != 0;
             if (CRTC.crtRasterCopyOn != rasterCopyOn) {
               CRTC.crtRasterCopyOn = rasterCopyOn;  //ラスタコピー
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
             }
             CRTC.crtClearStandby = (d & 2) != 0;  //高速クリア
             return;
           }
         }
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       int aa = a & 0x07ff;  //$00E80800～$00E81FFF→$00E80000～$00E807FF
       if (aa < 0x0400) {  //$00E80000～$00E803FF
         aa &= 0x003f;
         switch (aa) {
         case 0x0000:  //R00
           CRTC.crtR00HFrontEndPort = d & 0xff;
           {
             int curr = CRTC.crtR00HFrontEndMask == 0 ? CRTC.crtR00HFrontEndPort : CRTC.crtR00HFrontEndTest;
             if (CRTC.crtR00HFrontEndCurr != curr) {
               CRTC.crtR00HFrontEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0002:  //R01
           CRTC.crtR01HSyncEndPort = d & 0xff;
           {
             int curr = CRTC.crtR01HSyncEndMask == 0 ? CRTC.crtR01HSyncEndPort : CRTC.crtR01HSyncEndTest;
             if (CRTC.crtR01HSyncEndCurr != curr) {
               CRTC.crtR01HSyncEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0004:  //R02
           CRTC.crtR02HBackEndPort = d & 0xff;
           {
             int curr = CRTC.crtR02HBackEndMask == 0 ? CRTC.crtR02HBackEndPort : CRTC.crtR02HBackEndTest;
             if (CRTC.crtR02HBackEndCurr != curr) {
               CRTC.crtR02HBackEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0006:  //R03
           CRTC.crtR03HDispEndPort = d & 0xff;
           {
             int curr = CRTC.crtR03HDispEndMask == 0 ? CRTC.crtR03HDispEndPort : CRTC.crtR03HDispEndTest;
             if (CRTC.crtR03HDispEndCurr != curr) {
               CRTC.crtR03HDispEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0008:  //R04
           CRTC.crtR04VFrontEndPort = d & 0x03ff;
           {
             int curr = CRTC.crtR04VFrontEndMask == 0 ? CRTC.crtR04VFrontEndPort : CRTC.crtR04VFrontEndTest;
             if (CRTC.crtR04VFrontEndCurr != curr) {
               CRTC.crtR04VFrontEndCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000a:  //R05
           CRTC.crtR05VSyncEndPort = d & 0x03ff;
           {
             int curr = CRTC.crtR05VSyncEndMask == 0 ? CRTC.crtR05VSyncEndPort : CRTC.crtR05VSyncEndTest;
             if (CRTC.crtR05VSyncEndCurr != curr) {
               CRTC.crtR05VSyncEndCurr = curr;
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000c:  //R06
           CRTC.crtR06VBackEndPort = d & 0x03ff;
           {
             int curr = CRTC.crtR06VBackEndMask == 0 ? CRTC.crtR06VBackEndPort : CRTC.crtR06VBackEndTest;
             if (CRTC.crtR06VBackEndCurr != curr) {
               CRTC.crtR06VBackEndCurr = curr;
               CRTC.crtVDispStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x000e:  //R07
           CRTC.crtR07VDispEndPort = d & 0x03ff;
           {
             int curr = CRTC.crtR07VDispEndMask == 0 ? CRTC.crtR07VDispEndPort : CRTC.crtR07VDispEndTest;
             if (CRTC.crtR07VDispEndCurr != curr) {
               CRTC.crtR07VDispEndCurr = curr;
               CRTC.crtVIdleStart = curr + 1;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               CRTC.crtRestart ();
             }
           }
           return;
         case 0x0010:  //R08
           d &= 0xff;
           if (CRTC.crtR08Adjust != d) {
             CRTC.crtR08Adjust = d;
             CRTC.crtRestart ();
           }
           return;
         case 0x0012:  //R09
           CRTC.crtR09IRQRasterPort = d & 0x03ff;
           {
             int curr = CRTC.crtR09IRQRasterMask == 0 ? CRTC.crtR09IRQRasterPort : CRTC.crtR09IRQRasterTest;
             if (CRTC.crtR09IRQRasterCurr != curr) {
               CRTC.crtR09IRQRasterCurr = curr;
               if (CRTC.CRT_RASTER_HASH_ON) {
                 CRTC.crtUpdateRasterHash ();
               }
               if (RasterBreakPoint.RBP_ON) {
                 RasterBreakPoint.rbpCheckIRQ ();
               }
             }
           }
           return;
         case 0x0014:  //R10
           CRTC.crtR10TxXPort = d & 0x03ff;
           {
             int curr = CRTC.crtR10TxXMask == 0 ? CRTC.crtR10TxXPort : CRTC.crtR10TxXTest;
             if (CRTC.crtR10TxXCurr != curr) {
               CRTC.crtR10TxXCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0016:  //R11
           CRTC.crtR11TxYPort = d & 0x03ff;
           {
             int curr = CRTC.crtR11TxYMask == 0 ? CRTC.crtR11TxYPort : CRTC.crtR11TxYTest;
             if (CRTC.crtR11TxYCurr != curr) {
               CRTC.crtR11TxYCurr = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0018:  //R12
         case 0x001c:  //R14
         case 0x0020:  //R16
         case 0x0024:  //R18
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR12GrXPort[p] = d & (p == 0 ? 0x03ff : 0x01ff);
             int curr = CRTC.crtR12GrXMask[p] == 0 ? CRTC.crtR12GrXPort[p] : CRTC.crtR12GrXTest[p];
             if (CRTC.crtR12GrXCurr[p] != curr) {
               CRTC.crtR12GrXCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x001a:  //R13
         case 0x001e:  //R15
         case 0x0022:  //R17
         case 0x0026:  //R19
           {
             int p = (aa - 0x0018) >> 2;  //0,1,2,3
             CRTC.crtR13GrYPort[p] = d & (p == 0 ? 0x03ff : 0x01ff);
             int curr = CRTC.crtR13GrYMask[p] == 0 ? CRTC.crtR13GrYPort[p] : CRTC.crtR13GrYTest[p];
             if (CRTC.crtR13GrYCurr[p] != curr) {
               CRTC.crtR13GrYCurr[p] = curr;
               CRTC.crtAllStamp += 2;
             }
           }
           return;
         case 0x0028:  //R20
           CRTC.crtSetTextStorage (d >> 12);
           CRTC.crtSetGraphicStorage (d >> 11);
           CRTC.crtSetMemoryMode (d >> 8);
           CRTC.crtHighResoPort = d >>> 4 & 1;
           CRTC.crtVResoPort    = d >>> 2 & 3;
           CRTC.crtHResoPort    = d       & 3;
           SpriteScreen.sprAccessible = (d & 0b10010) != 0b10010;
           int highResoCurr = CRTC.crtHighResoMask == 0 ? CRTC.crtHighResoPort : CRTC.crtHighResoTest;
           int vResoCurr = CRTC.crtVResoMask == 0 ? CRTC.crtVResoPort : CRTC.crtVResoTest;
           int hResoCurr = CRTC.crtHResoMask == 0 ? CRTC.crtHResoPort : CRTC.crtHResoTest;
           if (CRTC.crtHighResoCurr != highResoCurr ||
               CRTC.crtVResoCurr != vResoCurr ||
               CRTC.crtHResoCurr != hResoCurr) {
             CRTC.crtHighResoCurr = highResoCurr;
             CRTC.crtVResoCurr = vResoCurr;
             CRTC.crtHResoCurr = hResoCurr;
             CRTC.crtRestart ();
           }
           return;
         case 0x002a:  //R21
           CRTC.crtBitMask   = d << 22 < 0;  //(d & 512) != 0
           CRTC.crtSimAccess = d << 23 < 0;  //(d & 256) != 0
           CRTC.crtSimPlane3 = (byte) d < 0;  //(d & 128) != 0。d << 24 < 0
           CRTC.crtSimPlane2 = d << 25 < 0;  //(d & 64) != 0
           CRTC.crtSimPlane1 = d << 26 < 0;  //(d & 32) != 0
           CRTC.crtSimPlane0 = d << 27 < 0;  //(d & 16) != 0
           CRTC.crtCCPlane3  = XEiJ.TEST_BIT_3_SHIFT ? d << 31 - 3 < 0 : (d & 8) != 0;
           CRTC.crtCCPlane2  = XEiJ.TEST_BIT_2_SHIFT ? d << 31 - 2 < 0 : (d & 4) != 0;
           CRTC.crtCCPlane1  = XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 < 0 : (d & 2) != 0;
           CRTC.crtCCPlane0  = XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 < 0 : (d & 1) != 0;
           return;
         case 0x002c:  //R22
           CRTC.crtR22SrcBlock = d >> 8 & 0xff;
           CRTC.crtR22DstBlock = d      & 0xff;
           return;
         case 0x002e:  //R23
           CRTC.crtR23Mask = (char) d;
           return;
           //case 0x0030:  //R24
           //case 0x0032:  //R25
           //case 0x0034:  //R26
           //case 0x0036:  //R27
           //case 0x0038:  //R28
           //case 0x003a:  //R29
           //case 0x003c:  //R30
           //case 0x003e:  //R31
         default:
           return;
         }
       } else {  //$00E80400～$00E807FF
         aa &= 0xff;  //$00E80500～$00E807FF→$00E80400～$00E804FF
         if (aa < 0x80) {  //$00E80400～$00E8047F
           super.mmdWw (a, d);  //バスエラー
         } else {  //$00E80480～$00E804FF  動作ポート
           boolean rasterCopyOn = (d & 8) != 0;
           if (CRTC.crtRasterCopyOn != rasterCopyOn) {
             CRTC.crtRasterCopyOn = rasterCopyOn;  //ラスタコピー
             if (CRTC.CRT_RASTER_HASH_ON) {
               CRTC.crtUpdateRasterHash ();
             }
           }
           CRTC.crtClearStandby = (d & 2) != 0;  //高速クリア
           return;
         }
       }
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_CRT

  //--------------------------------------------------------------------------------
  //MMD_VCN ビデオコントローラ
  //
  //  $00E82000～$00E821FF  ワードレジスタ
  //  $00E82200～$00E823FF  ワードレジスタ
  //  $00E82400～$00E82401  ワードレジスタ
  //  $00E82402～$00E824FF  $00E82400～$00E82401の繰り返し
  //  $00E82500～$00E82501  ワードレジスタ
  //  $00E82502～$00E825FF  $00E82500～$00E82501の繰り返し
  //  $00E82600～$00E82601  ワードレジスタ
  //  $00E82602～$00E826FF  $00E82600～$00E82601の繰り返し
  //  $00E82700～$00E82FFF  $0000
  //  $00E83000～$00E83FFF  $00E82000～$00E82FFFの繰り返し
  //
  MMD_VCN {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "ビデオコントローラ" : "Video Controller";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       int aa = a & 0x1fff;  //$00E83000～$00E83FFF→$00E82000～$00E82FFF
       int d = (aa < 0x0200 ? VideoController.vcnPal16G8[aa >> 1] :  //$00E82000～$00E821FF  グラフィックスパレット
                aa < 0x0400 ? VideoController.vcnPal16TS[(aa - 0x0200) >> 1] :  //$00E82200～$00E823FF  テキストスプライトパレット
                aa < 0x0500 ? VideoController.vcnReg1Port :  //$00E82400～$00E824FF
                aa < 0x0600 ? VideoController.vcnReg2Port :  //$00E82500～$00E825FF
                aa < 0x0700 ? VideoController.vcnReg3Port :  //$00E82600～$00E826FF
                0);  //$00E82700～$00E82FFF
       return (aa & 1) == 0 ? d >> 8 : d & 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int aa = a & 0x1fff;  //$00E83000～$00E83FFF→$00E82000～$00E82FFF
       return (aa < 0x0200 ? VideoController.vcnPal16G8[aa >> 1] :  //$00E82000～$00E821FF  グラフィックスパレット
               aa < 0x0400 ? VideoController.vcnPal16TS[(aa - 0x0200) >> 1] :  //$00E82200～$00E823FF  テキストスプライトパレット
               aa < 0x0500 ? VideoController.vcnReg1Port :  //$00E82400～$00E824FF
               aa < 0x0600 ? VideoController.vcnReg2Port :  //$00E82500～$00E825FF
               aa < 0x0700 ? VideoController.vcnReg3Port :  //$00E82600～$00E826FF
               0);  //$00E82700～$00E82FFF
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       int aa = a & 0x1fff;  //$00E83000～$00E83FFF→$00E82000～$00E82FFF
       if (aa < 0x0200) {  //$00E82000～$00E821FF  グラフィックスパレット
         d &= 0xff;
         int n = aa >> 1;
         if ((aa & 1) == 0) {  //a=0,2,4,6,8,10,12,14
           VideoController.vcnPal32G8[n] = VideoController.vcnPalTbl[
             VideoController.vcnPal16G8[n] = d << 8 | (VideoController.vcnPal16G8[n] & 0xff)
             ];
           if ((n & 1) == 0) {  //a=0,4,8,12 n=0,2,4,6
             VideoController.vcnPal8G16L[n] = d;
           } else {  //a=2,6,10,14 n-1=0,2,4,6
             VideoController.vcnPal8G16H[n - 1] = d << 8;
           }
         } else {  //a=1,3,5,7,9,11,13,15
           VideoController.vcnPal32G8[n] = VideoController.vcnPalTbl[
             VideoController.vcnPal16G8[n] = (VideoController.vcnPal16G8[n] & ~0xff) | d
             ];
           if ((n & 1) == 0) {  //a=1,5,9,13 n+1=1,3,5,7
             VideoController.vcnPal8G16L[n + 1] = d;
           } else {  //a=3,7,11,15 n=1,3,5,7
             VideoController.vcnPal8G16H[n] = d << 8;
           }
         }
         if ((VideoController.vcnReg3Curr & 0x001f) != 0) {  //グラフィックス画面が表示されている
           CRTC.crtAllStamp += 2;
         }
       } else if (aa < 0x0400) {  //$00E82200～$00E823FF  テキストスプライトパレット
         d &= 0xff;
         int n = (aa - 0x0200) >> 1;
         VideoController.vcnPal32TS[n] = VideoController.vcnPalTbl[
           VideoController.vcnPal16TS[n] = ((aa & 1) == 0 ?
                                            d << 8 | (VideoController.vcnPal16TS[n] & 0xff) :
                                            (VideoController.vcnPal16TS[n] & ~0xff) | d)
           ];
         if ((VideoController.vcnReg3Curr << 31 - 6 < 0 && SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) ||  //スプライト画面が表示されている
             (VideoController.vcnReg3Curr << 31 - 5 < 0 && n < 16)) {  //テキスト画面が表示されていてテキストパレットが操作された
           CRTC.crtAllStamp += 2;
         }
       } else if (aa < 0x0500) {  //$00E82400～$00E824FF
         d = (aa & 1) == 0 ? 0x00 : d & 0x07;
         if (VideoController.vcnReg1Port != d) {
           VideoController.vcnReg1Port = d;
           VideoController.vcnReg1Curr = ((VideoController.vcnReg1Port & ~VideoController.vcnReg1Mask) |
                                          (VideoController.vcnReg1Test & VideoController.vcnReg1Mask));
           VideoController.vcnUpdateMode ();
         }
       } else if (aa < 0x0600) {  //$00E82500～$00E825FF
         d = ((aa & 1) == 0 ?
              (d & 0x3f) << 8 | (VideoController.vcnReg2Port & 0xff) :
              (VideoController.vcnReg2Port & ~0xff) | (d & 0xff));
         if (VideoController.vcnReg2Port != d) {
           VideoController.vcnReg2Port = d;
           VideoController.vcnReg2Curr = ((VideoController.vcnReg2Port & ~VideoController.vcnReg2Mask) |
                                          (VideoController.vcnReg2Test & VideoController.vcnReg2Mask));
           VideoController.vcnUpdateMode ();
         }
       } else if (aa < 0x0700) {  //$00E82600～$00E826FF
         d = ((aa & 1) == 0 ?
              (d & 0xff) << 8 | (VideoController.vcnReg3Port & 0xff) :
              (VideoController.vcnReg3Port & ~0xff) | (d & 0xff));
         if (VideoController.vcnReg3Port != d) {
           VideoController.vcnReg3Port = d;
           VideoController.vcnReg3Curr = ((VideoController.vcnReg3Port & ~VideoController.vcnReg3Mask) |
                                          (VideoController.vcnReg3Test & VideoController.vcnReg3Mask));
           VideoController.vcnUpdateMode ();
         }
       } else {  //$00E82700～$00E82FFF
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       int aa = a & 0x1fff;  //$00E83000～$00E83FFF→$00E82000～$00E82FFF
       if (aa < 0x0200) {  //$00E82000～$00E821FF  グラフィックスパレット
         int n = aa >> 1;  //a=0,2,4,6,8,10,12,14 n=0,1,2,3,4,5,6,7
         VideoController.vcnPal32G8[n] = VideoController.vcnPalTbl[
           VideoController.vcnPal16G8[n] = (char) d
           ];
         if ((n & 1) == 0) {  //a=0,4,8,12 n=0,2,4,6 n+1=1,3,5,7
           VideoController.vcnPal8G16L[n] = d >> 8 & 0xff;
           VideoController.vcnPal8G16L[n + 1] = d & 0xff;
         } else {  //a=2,4,6,8 n-1=0,2,4,6 n=1,3,5,7
           VideoController.vcnPal8G16H[n - 1] = d & 0xff00;
           VideoController.vcnPal8G16H[n] = (d & 0xff) << 8;
         }
         if ((VideoController.vcnReg3Curr & 0x001f) != 0) {  //グラフィックス画面が表示されている
           CRTC.crtAllStamp += 2;
         }
       } else if (aa < 0x0400) {  //$00E82200～$00E823FF  テキストスプライトパレット
         int n = (aa - 0x0200) >> 1;
         VideoController.vcnPal32TS[n] = VideoController.vcnPalTbl[
           VideoController.vcnPal16TS[n] = (char) d
           ];
         if ((VideoController.vcnReg3Curr << 31 - 6 < 0 && SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) ||  //スプライト画面が表示されている
             (VideoController.vcnReg3Curr << 31 - 5 < 0 && n < 16)) {  //テキスト画面が表示されていてテキストパレットが操作された
           CRTC.crtAllStamp += 2;
         }
       } else if (aa < 0x0500) {  //$00E82400～$00E824FF
         d &= 0x0007;
         if (VideoController.vcnReg1Port != d) {
           VideoController.vcnReg1Port = d;
           VideoController.vcnReg1Curr = ((VideoController.vcnReg1Port & ~VideoController.vcnReg1Mask) |
                                          (VideoController.vcnReg1Test & VideoController.vcnReg1Mask));
           VideoController.vcnUpdateMode ();
         }
       } else if (aa < 0x0600) {  //$00E82500～$00E825FF
         d &= 0x3fff;
         if (VideoController.vcnReg2Port != d) {
           VideoController.vcnReg2Port = d;
           VideoController.vcnReg2Curr = ((VideoController.vcnReg2Port & ~VideoController.vcnReg2Mask) |
                                          (VideoController.vcnReg2Test & VideoController.vcnReg2Mask));
           VideoController.vcnUpdateMode ();
         }
       } else if (aa < 0x0700) {  //$00E82600～$00E826FF
         d &= 0xffff;
         if (VideoController.vcnReg3Port != d) {
           VideoController.vcnReg3Port = d;
           VideoController.vcnReg3Curr = ((VideoController.vcnReg3Port & ~VideoController.vcnReg3Mask) |
                                          (VideoController.vcnReg3Test & VideoController.vcnReg3Mask));
           VideoController.vcnUpdateMode ();
         }
       } else {  //$00E82700～$00E82FFF
       }
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_VCN

  //--------------------------------------------------------------------------------
  //MMD_DMA DMAコントローラ
  //
  //  $00E84000～$00E8403F  ワードまたはロングワードレジスタ(CH0)
  //  $00E84040～$00E8407F  ワードまたはロングワードレジスタ(CH1)
  //  $00E84080～$00E840BF  ワードまたはロングワードレジスタ(CH2)
  //  $00E840C0～$00E840FF  ワードまたはロングワードレジスタ(CH4)
  //  $00E84100～$00E85FFF  $00E84000～$00E840FFの繰り返し
  //
  MMD_DMA {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "DMA コントローラ" : "DMA Controller";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       int d;
       int al = a & 0xff;
       if (al == HD63450.DMA_GCR) {
         d = HD63450.dmaBT | HD63450.dmaBR;
         if (HD63450.DMA_DEBUG_TRACE != 0) {
           System.out.printf ("%d %08x HD63450.dmaRbz(0x%08x)=0x%02x\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
         }
       } else {
         int i = al >> 6;  //チャンネル
         switch (al & 0x3f) {
         case HD63450.DMA_CSR:
           d = HD63450.dmaCOC[i] | HD63450.dmaBLC[i] | HD63450.dmaNDT[i] | HD63450.dmaERR[i] | HD63450.dmaACT[i] | HD63450.dmaDIT[i] | HD63450.dmaPCT[i] | HD63450.dmaPCS[i];
           break;
         case HD63450.DMA_CER:
           d = HD63450.dmaErrorCode[i];
           break;
         case HD63450.DMA_DCR:
           d = HD63450.dmaXRM[i] | HD63450.dmaDTYP[i] | HD63450.dmaDPS[i] | HD63450.dmaPCL[i];
           break;
         case HD63450.DMA_OCR:
           d = HD63450.dmaDIR[i] | HD63450.dmaBTD[i] | HD63450.dmaSIZE[i] | HD63450.dmaCHAIN[i] | HD63450.dmaREQG[i];
           break;
         case HD63450.DMA_SCR:
           d = HD63450.dmaMAC[i] | HD63450.dmaDAC[i];
           break;
         case HD63450.DMA_CCR:
           d = HD63450.dmaSTR[i] | HD63450.dmaCNT[i] | HD63450.dmaHLT[i] | HD63450.dmaSAB[i] | HD63450.dmaITE[i];
           break;
         case HD63450.DMA_MTC:
           d = HD63450.dmaMTC[i] >> 8;
           break;
         case HD63450.DMA_MTC + 1:
           d = HD63450.dmaMTC[i] & 0xff;
           break;
         case HD63450.DMA_MAR:
           d = HD63450.dmaMAR[i] >>> 24;
           break;
         case HD63450.DMA_MAR + 1:
           d = HD63450.dmaMAR[i] >> 16 & 0xff;
           break;
         case HD63450.DMA_MAR + 2:
           d = (char) HD63450.dmaMAR[i] >> 8;
           break;
         case HD63450.DMA_MAR + 3:
           d = HD63450.dmaMAR[i] & 0xff;
           break;
         case HD63450.DMA_DAR:
           d = HD63450.dmaDAR[i] >>> 24;
           break;
         case HD63450.DMA_DAR + 1:
           d = HD63450.dmaDAR[i] >> 16 & 0xff;
           break;
         case HD63450.DMA_DAR + 2:
           d = (char) HD63450.dmaDAR[i] >> 8;
           break;
         case HD63450.DMA_DAR + 3:
           d = HD63450.dmaDAR[i] & 0xff;
           break;
         case HD63450.DMA_BTC:
           d = HD63450.dmaBTC[i] >> 8;
           break;
         case HD63450.DMA_BTC + 1:
           d = HD63450.dmaBTC[i] & 0xff;
           break;
         case HD63450.DMA_BAR:
           d = HD63450.dmaBAR[i] >>> 24;
           break;
         case HD63450.DMA_BAR + 1:
           d = HD63450.dmaBAR[i] >> 16 & 0xff;
           break;
         case HD63450.DMA_BAR + 2:
           d = (char) HD63450.dmaBAR[i] >> 8;
           break;
         case HD63450.DMA_BAR + 3:
           d = HD63450.dmaBAR[i] & 0xff;
           break;
         case HD63450.DMA_NIV:
           d = HD63450.dmaNIV[i];
           break;
         case HD63450.DMA_EIV:
           d = HD63450.dmaEIV[i];
           break;
         case HD63450.DMA_MFC:
           d = HD63450.dmaMFC[i];
           break;
         case HD63450.DMA_CPR:
           d = HD63450.dmaCP[i];
           break;
         case HD63450.DMA_DFC:
           d = HD63450.dmaDFC[i];
           break;
         case HD63450.DMA_BFC:
           d = HD63450.dmaBFC[i];
           break;
         default:
           d = 0;
         }
         if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
           System.out.printf ("%d %08x HD63450.dmaRbz(0x%08x)=0x%02x\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
         }
       }
       return d;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int d;
       int al = a & 0xff;
       int i = al >> 6;  //チャンネル
       switch (al & 0x3f) {
       case HD63450.DMA_MTC:
         d = HD63450.dmaMTC[i];
         break;
       case HD63450.DMA_MAR:
         d = HD63450.dmaMAR[i] >>> 16;
         break;
       case HD63450.DMA_MAR + 2:
         d = (char) HD63450.dmaMAR[i];
         break;
       case HD63450.DMA_DAR:
         d = HD63450.dmaDAR[i] >>> 16;
         break;
       case HD63450.DMA_DAR + 2:
         d = (char) HD63450.dmaDAR[i];
         break;
       case HD63450.DMA_BTC:
         d = HD63450.dmaBTC[i];
         break;
       case HD63450.DMA_BAR:
         d = HD63450.dmaBAR[i] >>> 16;
         break;
       case HD63450.DMA_BAR + 2:
         d = (char) HD63450.dmaBAR[i];
         break;
       default:
         d = mmdRbz (a) << 8 | mmdRbz (a + 1);
       }
       if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
         System.out.printf ("%d %08x HD63450.dmaRwz(0x%08x)=0x%04x\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
       }
       return d;
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       int d;
       int al = a & 0xff;
       int i = al >> 6;  //チャンネル
       switch (al & 0x3f) {
       case HD63450.DMA_MAR:
         d = HD63450.dmaMAR[i];
         break;
       case HD63450.DMA_DAR:
         d = HD63450.dmaDAR[i];
         break;
       case HD63450.DMA_BAR:
         d = HD63450.dmaBAR[i];
         break;
       default:
         d = mmdRwz (a) << 16 | mmdRwz (a + 2);
       }
       if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
         System.out.printf ("%d %08x HD63450.dmaRls(0x%08x)=0x%08x\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
       }
       return d;
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       d &= 0xff;
       int al = a & 0xff;
       if (al == HD63450.DMA_GCR) {
         if (HD63450.DMA_DEBUG_TRACE != 0) {
           System.out.printf ("%d %08x HD63450.dmaWb(0x%08x,0x%02x)\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
         }
         HD63450.dmaBT = d & HD63450.DMA_BT;
         HD63450.dmaBR = d & HD63450.DMA_BR;
         HD63450.dmaBurstInterval = XEiJ.dmaCycleUnit << 4 + (HD63450.dmaBT >> 2);
         HD63450.dmaBurstSpan = HD63450.dmaBurstInterval >> 1 + (HD63450.dmaBR & 3);
         return;
       }
       int i = al >> 6;  //チャンネル
       if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
         System.out.printf ("%d %08x HD63450.dmaWb(0x%08x,0x%02x)\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
       }
       switch (al & 0x3f) {
       case HD63450.DMA_CSR:
         if (d == 0xff) {  //HD63450.DMA_CSRは-1を書き込むとクリアされる。それ以外は無視
           HD63450.dmaCOC[i] = 0;
           HD63450.dmaBLC[i] = 0;
           HD63450.dmaNDT[i] = 0;
           HD63450.dmaERR[i] = 0;
         }
         return;
       case HD63450.DMA_CER:
         //HD63450.DMA_CERはread-only
         return;
       case HD63450.DMA_DCR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaXRM[i] = d & HD63450.DMA_XRM;
         HD63450.dmaDTYP[i] = d & HD63450.DMA_DTYP;
         HD63450.dmaDPS[i] = d & HD63450.DMA_DPS;
         HD63450.dmaPCL[i] = d & HD63450.DMA_PCL;
         return;
       case HD63450.DMA_OCR:
         HD63450.dmaDIR[i] = d & HD63450.DMA_DIR;
         HD63450.dmaBTD[i] = d & HD63450.DMA_BTD;
         HD63450.dmaSIZE[i] = d & HD63450.DMA_SIZE;
         HD63450.dmaCHAIN[i] = d & HD63450.DMA_CHAIN;
         HD63450.dmaREQG[i] = d & HD63450.DMA_REQG;
         return;
       case HD63450.DMA_SCR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMAC[i] = d & HD63450.DMA_MAC;
         HD63450.dmaMACValue[i] = HD63450.dmaMAC[i] == HD63450.DMA_INC_MAR ? 1 : HD63450.dmaMAC[i] == HD63450.DMA_DEC_MAR ? -1 : 0;
         HD63450.dmaDAC[i] = d & HD63450.DMA_DAC;
         HD63450.dmaDACValue[i] = HD63450.dmaDAC[i] == HD63450.DMA_INC_DAR ? 1 : HD63450.dmaDAC[i] == HD63450.DMA_DEC_DAR ? -1 : 0;
         return;
       case HD63450.DMA_CCR:
         HD63450.dmaHLT[i] = d & HD63450.DMA_HLT;
         HD63450.dmaITE[i] = d & HD63450.DMA_ITE;
         //HD63450.DMA_CNT
         if ((d & HD63450.DMA_CNT) != 0) {
           if ((HD63450.dmaACT[i] == 0 && (d & HD63450.DMA_STR) == 0) || HD63450.dmaBLC[i] != 0) {  //動作中でないかブロック転送完了済みの状態でHD63450.DMA_CNTをセットしようとした
             HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
             return;
           }
           if (HD63450.dmaCHAIN[i] != HD63450.DMA_NO_CHAINING) {  //チェインモードのときHD63450.DMA_CNTをセットしようとした
             HD63450.dmaErrorExit (i, HD63450.DMA_CONFIGURATION_ERROR);
             return;
           }
           HD63450.dmaCNT[i] = HD63450.DMA_CNT;  //HD63450.DMA_CNTに0を書き込んでもクリアされない
         }
         //HD63450.DMA_SAB
         if ((d & HD63450.DMA_SAB) != 0) {
           //HD63450.dmaSABには書き込まない。SABは読み出すと常に0
           HD63450.dmaCOC[i] = 0;
           HD63450.dmaBLC[i] = 0;
           HD63450.dmaNDT[i] = 0;
           HD63450.dmaHLT[i] = 0;
           HD63450.dmaCNT[i] = 0;
           if (HD63450.dmaACT[i] != 0 || (d & HD63450.DMA_STR) != 0) {  //動作中
             HD63450.dmaErrorExit (i, HD63450.DMA_SOFTWARE_ABORT);
           }
           return;
         }
         //HD63450.DMA_STR
         if ((d & HD63450.DMA_STR) != 0) {
           HD63450.dmaStart (i);
         }
         return;
       case HD63450.DMA_MTC:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMTC[i] = d << 8 | (HD63450.dmaMTC[i] & 0xff);
         return;
       case HD63450.DMA_MTC + 1:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMTC[i] = (HD63450.dmaMTC[i] & ~0xff) | d;
         return;
       case HD63450.DMA_MAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMAR[i] = d << 24 | (HD63450.dmaMAR[i] & ~(0xff << 24));
         return;
       case HD63450.DMA_MAR + 1:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMAR[i] = d << 16 | (HD63450.dmaMAR[i] & ~(0xff << 16));
         return;
       case HD63450.DMA_MAR + 2:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMAR[i] = (HD63450.dmaMAR[i] & ~(0xff << 8)) | d << 8;
         return;
       case HD63450.DMA_MAR + 3:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMAR[i] = (HD63450.dmaMAR[i] & ~0xff) | d;
         return;
       case HD63450.DMA_DAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaDAR[i] = d << 24 | (HD63450.dmaDAR[i] & ~(0xff << 24));
         return;
       case HD63450.DMA_DAR + 1:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaDAR[i] = d << 16 | (HD63450.dmaDAR[i] & ~(0xff << 16));
         return;
       case HD63450.DMA_DAR + 2:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaDAR[i] = (HD63450.dmaDAR[i] & ~(0xff << 8)) | d << 8;
         return;
       case HD63450.DMA_DAR + 3:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaDAR[i] = (HD63450.dmaDAR[i] & ~0xff) | d;
         return;
       case HD63450.DMA_BTC:
         HD63450.dmaBTC[i] = d << 8 | (HD63450.dmaBTC[i] & 0xff);
         return;
       case HD63450.DMA_BTC + 1:
         HD63450.dmaBTC[i] = (HD63450.dmaBTC[i] & ~0xff) | d;
         return;
       case HD63450.DMA_BAR:
         HD63450.dmaBAR[i] = d << 24 | (HD63450.dmaBAR[i] & ~(0xff << 24));
         return;
       case HD63450.DMA_BAR + 1:
         HD63450.dmaBAR[i] = d << 16 | (HD63450.dmaBAR[i] & ~(0xff << 16));
         return;
       case HD63450.DMA_BAR + 2:
         HD63450.dmaBAR[i] = (HD63450.dmaBAR[i] & ~(0xff << 8)) | d << 8;
         return;
       case HD63450.DMA_BAR + 3:
         HD63450.dmaBAR[i] = (HD63450.dmaBAR[i] & ~0xff) | d;
         return;
       case HD63450.DMA_NIV:
         HD63450.dmaNIV[i] = d;
         return;
       case HD63450.DMA_EIV:
         HD63450.dmaEIV[i] = d;
         return;
       case HD63450.DMA_MFC:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaMFC[i] = d & HD63450.DMA_FC2;
         if (DataBreakPoint.DBP_ON) {
           HD63450.dmaMFCMap[i] = HD63450.dmaMFC[i] == 0 ? DataBreakPoint.dbpUserMap : DataBreakPoint.dbpSuperMap;
         } else {
           HD63450.dmaMFCMap[i] = HD63450.dmaMFC[i] == 0 ? XEiJ.busUserMap : XEiJ.busSuperMap;
         }
         return;
       case HD63450.DMA_CPR:
         HD63450.dmaCP[i] = d & HD63450.DMA_CP;
         return;
       case HD63450.DMA_DFC:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
           return;
         }
         HD63450.dmaDFC[i] = d & HD63450.DMA_FC2;
         if (DataBreakPoint.DBP_ON) {
           HD63450.dmaDFCMap[i] = HD63450.dmaDFC[i] == 0 ? DataBreakPoint.dbpUserMap : DataBreakPoint.dbpSuperMap;
         } else {
           HD63450.dmaDFCMap[i] = HD63450.dmaDFC[i] == 0 ? XEiJ.busUserMap : XEiJ.busSuperMap;
         }
         return;
       case HD63450.DMA_BFC:
         HD63450.dmaBFC[i] = d & HD63450.DMA_FC2;
         if (DataBreakPoint.DBP_ON) {
           HD63450.dmaBFCMap[i] = HD63450.dmaBFC[i] == 0 ? DataBreakPoint.dbpUserMap : DataBreakPoint.dbpSuperMap;
         } else {
           HD63450.dmaBFCMap[i] = HD63450.dmaBFC[i] == 0 ? XEiJ.busUserMap : XEiJ.busSuperMap;
         }
         return;
       default:
         return;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       d = (char) d;
       int al = a & 0xff;
       int i = al >> 6;  //チャンネル
       if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
         System.out.printf ("%d %08x HD63450.dmaWw(0x%08x,0x%04x)\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
       }
       switch (al & 0x3f) {
       case HD63450.DMA_MTC:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaMTC[i] = d;
         }
         return;
       case HD63450.DMA_MAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaMAR[i] = d << 16 | (char) HD63450.dmaMAR[i];
         }
         return;
       case HD63450.DMA_MAR + 2:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaMAR[i] = (HD63450.dmaMAR[i] & ~0xffff) | d;
         }
         return;
       case HD63450.DMA_DAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaDAR[i] = d << 16 | (char) HD63450.dmaDAR[i];
         }
         return;
       case HD63450.DMA_DAR + 2:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaDAR[i] = (HD63450.dmaDAR[i] & ~0xffff) | d;
         }
         return;
       case HD63450.DMA_BTC:
         HD63450.dmaBTC[i] = (char) d;
         return;
       case HD63450.DMA_BAR:
         HD63450.dmaBAR[i] = d << 16 | (char) HD63450.dmaBAR[i];
         return;
       case HD63450.DMA_BAR + 2:
         HD63450.dmaBAR[i] = (HD63450.dmaBAR[i] & ~0xffff) | d;
         return;
       default:
         mmdWb (a, d >> 8);
         mmdWb (a + 1, d);
       }
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       int al = a & 0xff;
       int i = al >> 6;  //チャンネル
       if (HD63450.DMA_DEBUG_TRACE != 0 && (HD63450.DMA_DEBUG_TRACE & 1 << i) != 0) {
         System.out.printf ("%d %08x HD63450.dmaWl(0x%08x,0x%08x)\n", XEiJ.mpuClockTime, XEiJ.regPC0, a, d);
       }
       switch (al & 0x3f) {
       case HD63450.DMA_MAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaMAR[i] = d;
         }
         return;
       case HD63450.DMA_DAR:
         if (HD63450.dmaACT[i] != 0) {
           HD63450.dmaErrorExit (i, HD63450.DMA_TIMING_ERROR);
         } else {
           HD63450.dmaDAR[i] = d;
         }
         return;
       case HD63450.DMA_BAR:
         HD63450.dmaBAR[i] = d;
         return;
       default:
         mmdWw (a, d >> 16);
         mmdWw (a + 2, d);
       }
     }
  },  //MMD_DMA

  //--------------------------------------------------------------------------------
  //MMD_SVS スーパーバイザ領域設定
  //
  //  $00E86000  バイトアクセスはバスエラー。ワードリードはバスエラー。ワードライトは$??xx
  //  $00E86001  バイトリードはバスエラー。バイトライトは$xx
  //  $00E86002～$00E87FFF  $00E86000～$00E86001の繰り返し
  //
  MMD_SVS {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "スーパーバイザ領域設定" : "Supervisor Area Setting";
    }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & 1) == 0) {  //偶数アドレスへのバイトライト
         super.mmdWb (a, d);  //バスエラー
       } else {  //奇数アドレスへのバイトライト
         a = (d & 0xff) + 1 << 13;
         XEiJ.busSuper (MemoryMappedDevice.MMD_MMR, 0x00000000, a);
         XEiJ.busUser (MemoryMappedDevice.MMD_MMR, a, Math.min (0x00200000, MainMemory.mmrMemorySizeCurrent));
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       mmdWb (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWw (a, d >> 16);
       mmdWw (a + 2, d);
     }
  },  //MMD_SVS

  //--------------------------------------------------------------------------------
  //MMD_MFP MFP
  //
  //  $00E88000～$00E8803F(偶数)  バイトアクセスはバスエラー。ワードリードは$FFxx。ワードライトは$??xx
  //  $00E88000～$00E8803F(奇数)  バイトレジスタ
  //  $00E88040～$00E89FFF  $00E88000～$00E8803Fの繰り返し
  //
  MMD_MFP {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "MFP" : "MFP";  //Multi Function Peripheral
    }
    //ピーク
    @Override protected int mmdPbz (int a) {
      try {
        return (a & (XEiJ.BUS_MOTHER_MASK & 0xffff803f)) == MC68901.MFP_UDR ? MC68901.mfpUdrQueueArray[MC68901.mfpUdrQueueRead] : mmdRbz (a);
      } catch (M68kException e) {
      }
      return 0;
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       switch (a & (XEiJ.BUS_MOTHER_MASK & 0xffff803f)) {  //$00E88040～$00E89FFF→$00E88000～$00E8803F
       case MC68901.MFP_GPIP_DATA:
         return MC68901.mfpGpipHsync | MC68901.mfpGpipRint | 0b00100000 | MC68901.mfpGpipVdisp | MC68901.mfpGpipOpmirq | MC68901.mfpGpipPower | MC68901.mfpGpipExpwon | MC68901.mfpGpipAlarm;  //GPIP5は常に1
       case MC68901.MFP_AER:
         return MC68901.mfpAer;
       case MC68901.MFP_DDR:
         return 0x00;
       case MC68901.MFP_IERA:
         return MC68901.mfpIer >>> 8;
       case MC68901.MFP_IERB:
         return MC68901.mfpIer & 0xff;
       case MC68901.MFP_IPRA:
         return ((MC68901.mfpInnerRequest[15] != MC68901.mfpInnerAcknowledged[15] ? 0b10000000 : 0) |
                 (MC68901.mfpInnerRequest[14] != MC68901.mfpInnerAcknowledged[14] ? 0b01000000 : 0) |
                 (MC68901.mfpInnerRequest[13] != MC68901.mfpInnerAcknowledged[13] ? 0b00100000 : 0) |
                 (MC68901.mfpInnerRequest[12] != MC68901.mfpInnerAcknowledged[12] ? 0b00010000 : 0) |
                 (MC68901.mfpInnerRequest[11] != MC68901.mfpInnerAcknowledged[11] ? 0b00001000 : 0) |
                 (MC68901.mfpInnerRequest[10] != MC68901.mfpInnerAcknowledged[10] ? 0b00000100 : 0) |
                 (MC68901.mfpInnerRequest[ 9] != MC68901.mfpInnerAcknowledged[ 9] ? 0b00000010 : 0) |
                 (MC68901.mfpInnerRequest[ 8] != MC68901.mfpInnerAcknowledged[ 8] ? 0b00000001 : 0));
       case MC68901.MFP_IPRB:
         return ((MC68901.mfpInnerRequest[ 7] != MC68901.mfpInnerAcknowledged[ 7] ? 0b10000000 : 0) |
                 (MC68901.mfpInnerRequest[ 6] != MC68901.mfpInnerAcknowledged[ 6] ? 0b01000000 : 0) |
                 (MC68901.mfpInnerRequest[ 5] != MC68901.mfpInnerAcknowledged[ 5] ? 0b00100000 : 0) |
                 (MC68901.mfpInnerRequest[ 4] != MC68901.mfpInnerAcknowledged[ 4] ? 0b00010000 : 0) |
                 (MC68901.mfpInnerRequest[ 3] != MC68901.mfpInnerAcknowledged[ 3] ? 0b00001000 : 0) |
                 (MC68901.mfpInnerRequest[ 2] != MC68901.mfpInnerAcknowledged[ 2] ? 0b00000100 : 0) |
                 (MC68901.mfpInnerRequest[ 1] != MC68901.mfpInnerAcknowledged[ 1] ? 0b00000010 : 0) |
                 (MC68901.mfpInnerRequest[ 0] != MC68901.mfpInnerAcknowledged[ 0] ? 0b00000001 : 0));
       case MC68901.MFP_ISRA:
         return ((MC68901.mfpInnerInService[15] ? 0b10000000 : 0) |
                 (MC68901.mfpInnerInService[14] ? 0b01000000 : 0) |
                 (MC68901.mfpInnerInService[13] ? 0b00100000 : 0) |
                 (MC68901.mfpInnerInService[12] ? 0b00010000 : 0) |
                 (MC68901.mfpInnerInService[11] ? 0b00001000 : 0) |
                 (MC68901.mfpInnerInService[10] ? 0b00000100 : 0) |
                 (MC68901.mfpInnerInService[ 9] ? 0b00000010 : 0) |
                 (MC68901.mfpInnerInService[ 8] ? 0b00000001 : 0));
       case MC68901.MFP_ISRB:
         return ((MC68901.mfpInnerInService[ 7] ? 0b10000000 : 0) |
                 (MC68901.mfpInnerInService[ 6] ? 0b01000000 : 0) |
                 (MC68901.mfpInnerInService[ 5] ? 0b00100000 : 0) |
                 (MC68901.mfpInnerInService[ 4] ? 0b00010000 : 0) |
                 (MC68901.mfpInnerInService[ 3] ? 0b00001000 : 0) |
                 (MC68901.mfpInnerInService[ 2] ? 0b00000100 : 0) |
                 (MC68901.mfpInnerInService[ 1] ? 0b00000010 : 0) |
                 (MC68901.mfpInnerInService[ 0] ? 0b00000001 : 0));
       case MC68901.MFP_IMRA:
         return MC68901.mfpImr >>> 8;
       case MC68901.MFP_IMRB:
         return MC68901.mfpImr & 0xff;
       case MC68901.MFP_VECTOR:
         return MC68901.mfpVectorHigh;
       case MC68901.MFP_TACR:
         return (MC68901.mfpTaEventcount ? 0x08 : 0) | MC68901.mfpTaPrescale;
       case MC68901.MFP_TBCR:
         return (MC68901.mfpTbEventcount ? 0x08 : 0) | MC68901.mfpTbPrescale;
       case MC68901.MFP_TCDCR:
         return MC68901.mfpTcPrescale << 4 | MC68901.mfpTdPrescale;
       case MC68901.MFP_TADR:
         if (MC68901.mfpTaEventcount || MC68901.mfpTaPrescale == 0) {
           return MC68901.mfpTaCurrent & 0xff;
         }
         return MC68901.mfpTaInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTaStart) / MC68901.mfpTaDelta) % MC68901.mfpTaInitial) & 0xff;
       case MC68901.MFP_TBDR:
         if (MC68901.mfpTbEventcount || MC68901.mfpTbPrescale == 0) {
           return MC68901.mfpTbCurrent & 0xff;
         }
         return MC68901.mfpTbInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTbStart) / MC68901.mfpTbDelta) % MC68901.mfpTbInitial) & 0xff;
       case MC68901.MFP_TCDR:
         if (MC68901.mfpTcPrescale == 0) {
           return MC68901.mfpTcCurrent & 0xff;
         }
         return MC68901.mfpTcInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTcStart) / MC68901.mfpTcDelta) % MC68901.mfpTcInitial) & 0xff;
       case MC68901.MFP_TDDR:
         if (MC68901.mfpTdPrescale == 0) {
           return MC68901.mfpTdCurrent & 0xff;
         }
         return MC68901.mfpTdInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTdStart) / MC68901.mfpTdDelta) % MC68901.mfpTdInitial) & 0xff;
       case MC68901.MFP_SYNC_CHAR:
         return 0;
       case MC68901.MFP_UCR:
         return 0;
       case MC68901.MFP_RSR:
         if (MC68901.MFP_KBD_ON) {
           return MC68901.mfpKbdReadPointer != MC68901.mfpKbdWritePointer ? 0x80 : 0;
         } else {
           return MC68901.mfpUdrQueueRead != MC68901.mfpUdrQueueWrite ? 0x80 : 0;
         }
       case MC68901.MFP_TSR:
         return 0x80;  //バッファエンプティ
       case MC68901.MFP_UDR:
         if (MC68901.MFP_KBD_ON) {
           return MC68901.mfpKbdReadData ();
         } else {
           if (MC68901.mfpUdrQueueRead != MC68901.mfpUdrQueueWrite) {  //キューが空でないとき
             MC68901.mfpUdrQueueRead = MC68901.mfpUdrQueueRead + 1 & MC68901.MFP_UDR_QUEUE_MASK;
             if (MC68901.mfpUdrQueueRead != MC68901.mfpUdrQueueWrite) {  //キューが空にならなければ再度割り込み要求を出す
               if ((MC68901.mfpIer & MC68901.MFP_INPUT_FULL_MASK) != 0) {
                 MC68901.mfpInnerRequest[MC68901.MFP_INPUT_FULL_LEVEL]++;
                 if ((MC68901.mfpImr & MC68901.MFP_INPUT_FULL_MASK) != 0) {
                   if (MC68901.MFP_DELAYED_INTERRUPT) {
                     XEiJ.mpuDIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
                   } else {
                     XEiJ.mpuIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
                   }
                 }
               }
             }
           }
           return MC68901.mfpUdrQueueArray[MC68901.mfpUdrQueueRead];  //最後に押されたまたは離されたキー
         }
       default:
         if ((a & 1) == 0) {  //偶数アドレスのバイトリード
           return super.mmdRbz (a);  //バスエラー
         }
         return 0xff;
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       return 0xff00 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       switch (a & (XEiJ.BUS_MOTHER_MASK & 0xffff803f)) {  //$00E88040～$00E89FFF→$00E88000～$00E8803F
       case MC68901.MFP_GPIP_DATA:
         return;
       case MC68901.MFP_AER:
         MC68901.mfpAer = d & 0xff;
         return;
       case MC68901.MFP_DDR:
         return;
       case MC68901.MFP_IERA:
         d = (char) (d << 8);
         int oldIera = MC68901.mfpIer;
         MC68901.mfpIer = d | (MC68901.mfpIer & 0xff);
         //MC68901.MFP_IERAのビットに0を書き込むとMC68901.MFP_IPRAの該当ビットも0になる
         if ((short) d >= 0) {  //(0b10000000_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[15] = MC68901.mfpInnerRequest[15];
         }
         if (d << 31 - 14 >= 0) {  //(0b01000000_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[14] = MC68901.mfpInnerRequest[14];
         }
         if (d << 31 - 13 >= 0) {  //(0b00100000_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[13] = MC68901.mfpInnerRequest[13];
         }
         if (d << 31 - 12 >= 0) {  //(0b00010000_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[12] = MC68901.mfpInnerRequest[12];
         } else if (oldIera << 31 - 12 >= 0 &&  //MC68901.MFP_INPUT_FULL_MASKが0→1
                    (MC68901.MFP_KBD_ON ?
                     MC68901.mfpKbdReadPointer != MC68901.mfpKbdWritePointer :  //バッファが空でないとき
                     MC68901.mfpUdrQueueRead != MC68901.mfpUdrQueueWrite)) {  //キューが空でないとき
           MC68901.mfpInnerRequest[MC68901.MFP_INPUT_FULL_LEVEL]++;
           if ((MC68901.mfpImr & MC68901.MFP_INPUT_FULL_MASK) != 0) {
             //MFPのキー入力割り込みを要求する
             if (MC68901.MFP_DELAYED_INTERRUPT) {
               XEiJ.mpuDIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
             } else {
               XEiJ.mpuIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
             }
           }
         }
         if (d << 31 - 11 >= 0) {  //(0b00001000_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[11] = MC68901.mfpInnerRequest[11];
         }
         if (d << 31 - 10 >= 0) {  //(0b00000100_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[10] = MC68901.mfpInnerRequest[10];
         }
         if (d << 31 - 9 >= 0) {  //(0b00000010_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 9] = MC68901.mfpInnerRequest[ 9];
         }
         if (d << 31 - 8 >= 0) {  //(0b00000001_00000000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 8] = MC68901.mfpInnerRequest[ 8];
         }
         return;
       case MC68901.MFP_IERB:
         MC68901.mfpIer = (MC68901.mfpIer & ~0xff) | (d & 0xff);
         //MC68901.MFP_IERBのビットに0を書き込むとMC68901.MFP_IPRBの該当ビットも0になる
         if ((byte) d >= 0) {  //(0b10000000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 7] = MC68901.mfpInnerRequest[ 7];
         }
         if (d << 31 - 6 >= 0) {  //(0b01000000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 6] = MC68901.mfpInnerRequest[ 6];
         }
         if (d << 31 - 5 >= 0) {  //(0b00100000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 5] = MC68901.mfpInnerRequest[ 5];
         }
         if (d << 31 - 4 >= 0) {  //(0b00010000 & d) == 0
           MC68901.mfpInnerAcknowledged[ 4] = MC68901.mfpInnerRequest[ 4];
         }
         if (XEiJ.TEST_BIT_3_SHIFT ? d << 31 - 3 >= 0 : (d & 8) == 0) {
           MC68901.mfpInnerAcknowledged[ 3] = MC68901.mfpInnerRequest[ 3];
         }
         if (XEiJ.TEST_BIT_2_SHIFT ? d << 31 - 2 >= 0 : (d & 4) == 0) {
           MC68901.mfpInnerAcknowledged[ 2] = MC68901.mfpInnerRequest[ 2];
         }
         if (XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 >= 0 : (d & 2) == 0) {
           MC68901.mfpInnerAcknowledged[ 1] = MC68901.mfpInnerRequest[ 1];
         }
         if (XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 >= 0 : (d & 1) == 0) {
           MC68901.mfpInnerAcknowledged[ 0] = MC68901.mfpInnerRequest[ 0];
         }
         return;
       case MC68901.MFP_IPRA:
         //MC68901.MFP_IPRAのビットは他のすべてのビットに1を書き込むことで直接クリアできる
         switch (d & 0xff) {
         case 0b01111111:
           MC68901.mfpInnerAcknowledged[15] = MC68901.mfpInnerRequest[15];
           break;
         case 0b10111111:
           MC68901.mfpInnerAcknowledged[14] = MC68901.mfpInnerRequest[14];
           break;
         case 0b11011111:
           MC68901.mfpInnerAcknowledged[13] = MC68901.mfpInnerRequest[13];
           break;
         case 0b11101111:
           MC68901.mfpInnerAcknowledged[12] = MC68901.mfpInnerRequest[12];
           break;
         case 0b11110111:
           MC68901.mfpInnerAcknowledged[11] = MC68901.mfpInnerRequest[11];
           break;
         case 0b11111011:
           MC68901.mfpInnerAcknowledged[10] = MC68901.mfpInnerRequest[10];
           break;
         case 0b11111101:
           MC68901.mfpInnerAcknowledged[ 9] = MC68901.mfpInnerRequest[ 9];
           break;
         case 0b11111110:
           MC68901.mfpInnerAcknowledged[ 8] = MC68901.mfpInnerRequest[ 8];
           break;
         }
         return;
       case MC68901.MFP_IPRB:
         //MC68901.MFP_IPRBのビットは他のすべてのビットに1を書き込むことで直接クリアできる
         switch (d & 0xff) {
         case 0b01111111:
           MC68901.mfpInnerAcknowledged[ 7] = MC68901.mfpInnerRequest[ 7];
           break;
         case 0b10111111:
           MC68901.mfpInnerAcknowledged[ 6] = MC68901.mfpInnerRequest[ 6];
           break;
         case 0b11011111:
           MC68901.mfpInnerAcknowledged[ 5] = MC68901.mfpInnerRequest[ 5];
           break;
         case 0b11101111:
           MC68901.mfpInnerAcknowledged[ 4] = MC68901.mfpInnerRequest[ 4];
           break;
         case 0b11110111:
           MC68901.mfpInnerAcknowledged[ 3] = MC68901.mfpInnerRequest[ 3];
           break;
         case 0b11111011:
           MC68901.mfpInnerAcknowledged[ 2] = MC68901.mfpInnerRequest[ 2];
           break;
         case 0b11111101:
           MC68901.mfpInnerAcknowledged[ 1] = MC68901.mfpInnerRequest[ 1];
           break;
         case 0b11111110:
           MC68901.mfpInnerAcknowledged[ 0] = MC68901.mfpInnerRequest[ 0];
           break;
         }
         return;
       case MC68901.MFP_ISRA:
         //MC68901.MFP_ISRAのビットは他のすべてのビットに1を書き込むことで直接クリアできる
         switch (d & 0xff) {
         case 0b01111111:
           MC68901.mfpInnerInService[15] = false;
           break;
         case 0b10111111:
           MC68901.mfpInnerInService[14] = false;
           break;
         case 0b11011111:
           MC68901.mfpInnerInService[13] = false;
           break;
         case 0b11101111:
           MC68901.mfpInnerInService[12] = false;
           break;
         case 0b11110111:
           MC68901.mfpInnerInService[11] = false;
           break;
         case 0b11111011:
           MC68901.mfpInnerInService[10] = false;
           break;
         case 0b11111101:
           MC68901.mfpInnerInService[ 9] = false;
           break;
         case 0b11111110:
           MC68901.mfpInnerInService[ 8] = false;
           break;
         }
         return;
       case MC68901.MFP_ISRB:
         //MC68901.MFP_ISRBのビットは他のすべてのビットに1を書き込むことで直接クリアできる
         switch (d & 0xff) {
         case 0b01111111:
           MC68901.mfpInnerInService[ 7] = false;
           break;
         case 0b10111111:
           MC68901.mfpInnerInService[ 6] = false;
           break;
         case 0b11011111:
           MC68901.mfpInnerInService[ 5] = false;
           break;
         case 0b11101111:
           MC68901.mfpInnerInService[ 4] = false;
           break;
         case 0b11110111:
           MC68901.mfpInnerInService[ 3] = false;
           break;
         case 0b11111011:
           MC68901.mfpInnerInService[ 2] = false;
           break;
         case 0b11111101:
           MC68901.mfpInnerInService[ 1] = false;
           break;
         case 0b11111110:
           MC68901.mfpInnerInService[ 0] = false;
           break;
         }
         return;
       case MC68901.MFP_IMRA:
         MC68901.mfpImr = (d & 0xff) << 8 | (MC68901.mfpImr & 0xff);
         //MC68901.MFP_IMRAのビットに1を書き込んだときMC68901.MFP_IPRAの該当ビットが1ならば割り込み発生
         if ((byte) d < 0 && MC68901.mfpInnerRequest[15] != MC68901.mfpInnerAcknowledged[15] ||  //(0b10000000 & d) != 0
             d << 31 - 6 < 0 && MC68901.mfpInnerRequest[14] != MC68901.mfpInnerAcknowledged[14] ||  //(0b01000000 & d) != 0
             d << 31 - 5 < 0 && MC68901.mfpInnerRequest[13] != MC68901.mfpInnerAcknowledged[13] ||  //(0b00100000 & d) != 0
             d << 31 - 4 < 0 && MC68901.mfpInnerRequest[12] != MC68901.mfpInnerAcknowledged[12] ||  //(0b00010000 & d) != 0
             (XEiJ.TEST_BIT_3_SHIFT ? d << 31 - 3 < 0 : (d & 8) != 0) && MC68901.mfpInnerRequest[11] != MC68901.mfpInnerAcknowledged[11] ||
             (XEiJ.TEST_BIT_2_SHIFT ? d << 31 - 2 < 0 : (d & 4) != 0) && MC68901.mfpInnerRequest[10] != MC68901.mfpInnerAcknowledged[10] ||
             (XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 < 0 : (d & 2) != 0) && MC68901.mfpInnerRequest[ 9] != MC68901.mfpInnerAcknowledged[ 9] ||
             (XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 < 0 : (d & 1) != 0) && MC68901.mfpInnerRequest[ 8] != MC68901.mfpInnerAcknowledged[ 8]) {
           if (MC68901.MFP_DELAYED_INTERRUPT) {
             XEiJ.mpuDIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
           } else {
             XEiJ.mpuIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
           }
         }
         return;
       case MC68901.MFP_IMRB:
         MC68901.mfpImr = (MC68901.mfpImr & ~0xff) | (d & 0xff);
         //MC68901.MFP_IMRBのビットに1を書き込んだときMC68901.MFP_IPRBの該当ビットが1ならば割り込み発生
         if ((byte) d < 0 && MC68901.mfpInnerRequest[ 7] != MC68901.mfpInnerAcknowledged[ 7] ||  //(0b10000000 & d) != 0
             d << 31 - 6 < 0 && MC68901.mfpInnerRequest[ 6] != MC68901.mfpInnerAcknowledged[ 6] ||  //(0b01000000 & d) != 0
             d << 31 - 5 < 0 && MC68901.mfpInnerRequest[ 5] != MC68901.mfpInnerAcknowledged[ 5] ||  //(0b00100000 & d) != 0
             d << 31 - 4 < 0 && MC68901.mfpInnerRequest[ 4] != MC68901.mfpInnerAcknowledged[ 4] ||  //(0b00010000 & d) != 0
             (XEiJ.TEST_BIT_3_SHIFT ? d << 31 - 3 < 0 : (d & 8) != 0) && MC68901.mfpInnerRequest[ 3] != MC68901.mfpInnerAcknowledged[ 3] ||
             (XEiJ.TEST_BIT_2_SHIFT ? d << 31 - 2 < 0 : (d & 4) != 0) && MC68901.mfpInnerRequest[ 2] != MC68901.mfpInnerAcknowledged[ 2] ||
             (XEiJ.TEST_BIT_1_SHIFT ? d << 31 - 1 < 0 : (d & 2) != 0) && MC68901.mfpInnerRequest[ 1] != MC68901.mfpInnerAcknowledged[ 1] ||
             (XEiJ.TEST_BIT_0_SHIFT ? d << 31 - 0 < 0 : (d & 1) != 0) && MC68901.mfpInnerRequest[ 0] != MC68901.mfpInnerAcknowledged[ 0]) {
           if (MC68901.MFP_DELAYED_INTERRUPT) {
             XEiJ.mpuDIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
           } else {
             XEiJ.mpuIRR |= XEiJ.MPU_MFP_INTERRUPT_MASK;
           }
         }
         return;
       case MC68901.MFP_VECTOR:
         MC68901.mfpVectorHigh = 0xf0 & d;  //ビット3は0(全チャンネル自動割込み終了モード)に固定
         return;
       case MC68901.MFP_TACR:
         {
           int prevPrescale = MC68901.mfpTaPrescale;
           MC68901.mfpTaEventcount = (d & 0x08) != 0;
           MC68901.mfpTaPrescale = d & 0x07;
           if (MC68901.mfpTaEventcount && MC68901.mfpTaPrescale != 0) {  //パルス幅計測モードはキャンセル
             MC68901.mfpTaEventcount = false;
             MC68901.mfpTaPrescale = 0;
           }
           if (MC68901.mfpTaEventcount) {  //イベントカウントモード
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTaCurrent = MC68901.mfpTaInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTaStart) / MC68901.mfpTaDelta) % MC68901.mfpTaInitial);
             }
             MC68901.mfpTaClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTaTicker);
           } else if (MC68901.mfpTaPrescale != 0) {  //ディレイモード
             //! ディレイモードで動作中にプリスケールを変更されるとカウンタが即座に初期値に戻ってしまう
             MC68901.mfpTaClock = (MC68901.mfpTaStart = XEiJ.mpuClockTime) + (MC68901.mfpTaDelta = MC68901.MFP_DELTA[MC68901.mfpTaPrescale]) * MC68901.mfpTaInitial;
             TickerQueue.tkqAdd (MC68901.mfpTaTicker, MC68901.mfpTaClock);
           } else {  //カウント停止
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTaCurrent = MC68901.mfpTaInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTaStart) / MC68901.mfpTaDelta) % MC68901.mfpTaInitial);
             }
             MC68901.mfpTaClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTaTicker);
           }
         }
         return;
       case MC68901.MFP_TBCR:
         {
           int prevPrescale = MC68901.mfpTbPrescale;
           MC68901.mfpTbEventcount = (d & 0x08) != 0;
           MC68901.mfpTbPrescale = d & 0x07;
           MC68901.mfpTbKeyboardOn = !MC68901.mfpTbEventcount && MC68901.mfpTbPrescale == 0x01 && MC68901.mfpTbInitial == 0x0d;
           if (MC68901.mfpTbEventcount && MC68901.mfpTbPrescale != 0) {  //パルス幅計測モードはキャンセル
             MC68901.mfpTbEventcount = false;
             MC68901.mfpTbPrescale = 0;
           }
           if (MC68901.mfpTbEventcount) {  //イベントカウントモード
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTbCurrent = MC68901.mfpTbInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTbStart) / MC68901.mfpTbDelta) % MC68901.mfpTbInitial);
             }
             MC68901.mfpTbClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTbTicker);
           } else if (MC68901.mfpTbPrescale != 0) {  //ディレイモード
             //! ディレイモードで動作中にプリスケールを変更されるとカウンタが即座に初期値に戻ってしまう
             MC68901.mfpTbClock = (MC68901.mfpTbStart = XEiJ.mpuClockTime) + (MC68901.mfpTbDelta = MC68901.MFP_DELTA[MC68901.mfpTbPrescale]) * MC68901.mfpTbInitial;
             TickerQueue.tkqAdd (MC68901.mfpTbTicker, MC68901.mfpTbClock);
           } else {  //カウント停止
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTbCurrent = MC68901.mfpTbInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTbStart) / MC68901.mfpTbDelta) % MC68901.mfpTbInitial);
             }
             MC68901.mfpTbClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTbTicker);
           }
         }
         return;
       case MC68901.MFP_TCDCR:
         {
           int prevPrescale = MC68901.mfpTcPrescale;
           MC68901.mfpTcPrescale = d >> 4 & 0x07;
           if (MC68901.mfpTcPrescale != 0) {  //ディレイモード
             //! ディレイモードで動作中にプリスケールを変更されるとカウンタが即座に初期値に戻ってしまう
             MC68901.mfpTcClock = (MC68901.mfpTcStart = XEiJ.mpuClockTime) + (MC68901.mfpTcDelta = MC68901.MFP_DELTA[MC68901.mfpTcPrescale]) * MC68901.mfpTcInitial;
             TickerQueue.tkqAdd (MC68901.mfpTcTicker, MC68901.mfpTcClock);
           } else {  //カウント停止
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTcCurrent = MC68901.mfpTcInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTcStart) / MC68901.mfpTcDelta) % MC68901.mfpTcInitial);
             }
             MC68901.mfpTcClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTcTicker);
           }
         }
         {
           int prevPrescale = MC68901.mfpTdPrescale;
           MC68901.mfpTdPrescale = d & 0x07;
           if (MC68901.mfpTdPrescale != 0) {  //ディレイモード
             //! ディレイモードで動作中にプリスケールを変更されるとカウンタが即座に初期値に戻ってしまう
             MC68901.mfpTdClock = (MC68901.mfpTdStart = XEiJ.mpuClockTime) + (MC68901.mfpTdDelta = MC68901.MFP_DELTA[MC68901.mfpTdPrescale]) * MC68901.mfpTdInitial;
             TickerQueue.tkqAdd (MC68901.mfpTdTicker, MC68901.mfpTdClock);
           } else {  //カウント停止
             if (prevPrescale != 0) {  //ディレイモードで動作中だったとき
               MC68901.mfpTdCurrent = MC68901.mfpTdInitial - (int) (Math.floor ((double) (XEiJ.mpuClockTime - MC68901.mfpTdStart) / MC68901.mfpTdDelta) % MC68901.mfpTdInitial);
             }
             MC68901.mfpTdClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (MC68901.mfpTdTicker);
           }
         }
         return;
       case MC68901.MFP_TADR:
         //! ディレイモードで動作中に初期値を変更するとオーバーフローするまでMC68901.MFP_TADRの値がずれてしまう
         d &= 0xff;
         MC68901.mfpTaInitial = d == 0 ? 256 : d;  //初期値
         if (!MC68901.mfpTaEventcount && MC68901.mfpTaPrescale == 0) {  //動作中でなければカウンタの値も変更される
           MC68901.mfpTaCurrent = MC68901.mfpTaInitial;
         }
         return;
       case MC68901.MFP_TBDR:
         //! ディレイモードで動作中に初期値を変更するとオーバーフローするまでMC68901.MFP_TADRの値がずれてしまう
         d &= 0xff;
         MC68901.mfpTbInitial = d == 0 ? 256 : d;  //初期値
         MC68901.mfpTbKeyboardOn = !MC68901.mfpTbEventcount && MC68901.mfpTbPrescale == 0x01 && MC68901.mfpTbInitial == 0x0d;
         if (!MC68901.mfpTbEventcount && MC68901.mfpTbPrescale == 0) {  //動作中でなければカウンタの値も変更される
           MC68901.mfpTbCurrent = MC68901.mfpTbInitial;
         }
         return;
       case MC68901.MFP_TCDR:
         //! ディレイモードで動作中に初期値を変更するとオーバーフローするまでMC68901.MFP_TADRの値がずれてしまう
         d &= 0xff;
         MC68901.mfpTcInitial = d == 0 ? 256 : d;  //初期値
         if (MC68901.mfpTcPrescale == 0) {  //動作中でなければカウンタの値も変更される
           MC68901.mfpTcCurrent = MC68901.mfpTcInitial;
         }
         return;
       case MC68901.MFP_TDDR:
         //! ディレイモードで動作中に初期値を変更するとオーバーフローするまでMC68901.MFP_TADRの値がずれてしまう
         d &= 0xff;
         MC68901.mfpTdInitial = d == 0 ? 256 : d;  //初期値
         if (MC68901.mfpTdPrescale == 0) {  //動作中でなければカウンタの値も変更される
           MC68901.mfpTdCurrent = MC68901.mfpTdInitial;
         }
         return;
       case MC68901.MFP_SYNC_CHAR:
         return;
       case MC68901.MFP_UCR:
         return;
       case MC68901.MFP_RSR:
         return;
       case MC68901.MFP_TSR:
         return;
       case MC68901.MFP_UDR:
         if (!MC68901.mfpTbKeyboardOn) {
           return;
         }
         if ((byte) d < 0) {  //LEDの状態
           Keyboard.kbdSetLedStatus (d);
         } else if ((d & 0xf8) == 0x40) {  //MSCTRL
         } else if ((d & 0xf8) == 0x48) {  //ロック
         } else if ((d & 0xfc) == 0x54) {  //LEDの明るさ
         } else if ((d & 0xfc) == 0x58) {  //テレビコントロール
         } else if ((d & 0xfc) == 0x5c) {  //OPT.2キーによるテレビコントロール
         } else if ((d & 0xf0) == 0x60) {  //リピート開始時間
           Keyboard.kbdSetRepeatDelay (0x0f & d);
         } else if ((d & 0xf0) == 0x70) {  //リピート間隔
           Keyboard.kbdSetRepeatInterval (0x0f & d);
         }
         return;
       default:
         if ((a & 1) == 0) {  //偶数アドレスのバイトライト
           super.mmdWb (a, d);  //バスエラー
         }
         return;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       mmdWb (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWw (a, d >> 16);
       mmdWw (a + 2, d);
     }
  },  //MMD_MFP

  //--------------------------------------------------------------------------------
  //MMD_RTC_FIRST RTC
  //
  //  $00E8A000～$00E8A01F(偶数)  バイトアクセスはバスエラー。ワードリードは$FFxx。ワードライトは$??xx
  //  $00E8A000～$00E8A01F(奇数)  バイトレジスタ
  //  $00E8A020～$00E8AFFF  $00E8A000～$00E8A01Fの繰り返し
  //
  MMD_RTC_FIRST {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "RTC" : "RTC";  //Real-Time Clock
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      return (byte) RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPbz (int a) {
      return RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPws (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPwz (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPls (int a) {
      return (RP5C15.rtcPeekByte (a + 1) << 16 |
              RP5C15.rtcPeekByte (a + 3));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       return (byte) RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       return RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRws (int a) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       return (RP5C15.rtcReadByte (a + 1) << 16 |
               RP5C15.rtcReadByte (a + 3));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       RP5C15.rtcWriteByte (a, d);
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       RP5C15.rtcWriteByte (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       XEiJ.busSuper (MemoryMappedDevice.MMD_RTC_NEXT, 0x00e8a000, 0x00e8c000);  //RTC RTC
       RP5C15.rtcUpdate ();
       RP5C15.rtcWriteByte (a + 1, d >> 16);
       RP5C15.rtcWriteByte (a + 3, d);
     }
  },  //MMD_RTC_FIRST

  //--------------------------------------------------------------------------------
  //MMD_RTC_NEXT RTC
  MMD_RTC_NEXT {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "RTC" : "RTC";  //Real-Time Clock
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      return (byte) RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPbz (int a) {
      return RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPws (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPwz (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPls (int a) {
      return (RP5C15.rtcPeekByte (a + 1) << 16 |
              RP5C15.rtcPeekByte (a + 3));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       return (byte) RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       return RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRws (int a) throws M68kException {
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return (RP5C15.rtcReadByte (a + 1) << 16 |
               RP5C15.rtcReadByte (a + 3));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       RP5C15.rtcWriteByte (a, d);
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       RP5C15.rtcWriteByte (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       RP5C15.rtcWriteByte (a + 1, d >> 16);
       RP5C15.rtcWriteByte (a + 3, d);
     }
  },  //MMD_RTC_NEXT

  //--------------------------------------------------------------------------------
  //MMD_RTC_TEST RTC
  MMD_RTC_TEST {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "RTC テスト" : "RTC test";  //Real-Time Clock
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      return (byte) RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPbz (int a) {
      return RP5C15.rtcPeekByte (a);
    }
    @Override protected int mmdPws (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPwz (int a) {
      return RP5C15.rtcPeekByte (a + 1);
    }
    @Override protected int mmdPls (int a) {
      return (RP5C15.rtcPeekByte (a + 1) << 16 |
              RP5C15.rtcPeekByte (a + 3));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       RP5C15.rtcTestUpdate ();
       return (byte) RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       RP5C15.rtcTestUpdate ();
       return RP5C15.rtcReadByte (a);
     }
    @Override protected int mmdRws (int a) throws M68kException {
       RP5C15.rtcTestUpdate ();
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       RP5C15.rtcTestUpdate ();
       return RP5C15.rtcReadByte (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       RP5C15.rtcTestUpdate ();
       return (RP5C15.rtcReadByte (a + 1) << 16 |
               RP5C15.rtcReadByte (a + 3));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       RP5C15.rtcTestUpdate ();
       RP5C15.rtcWriteByte (a, d);
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       RP5C15.rtcTestUpdate ();
       RP5C15.rtcWriteByte (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       RP5C15.rtcTestUpdate ();
       RP5C15.rtcWriteByte (a + 1, d >> 16);
       RP5C15.rtcWriteByte (a + 3, d);
     }
  },  //MMD_RTC_TEST

  //--------------------------------------------------------------------------------
  //MMD_PRN プリンタポート
  //
  //  $00E8C000～$00E8C003(偶数)  バイトリードは$FF。バイトライトは$??。ワードリードは$FFxx。ワードライトは$??xx
  //  $00E8C000～$00E8C003(奇数)  バイトリードは$FF。バイトライトは$??
  //  $00E8C004～$00E8DFFF  $00E8C000～$00E8C003の繰り返し(?)
  //
  MMD_PRN {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "プリンタポート" : "Printer Port";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a == PrinterPort.PRN_DATA ? PrinterPort.prnReadData () :
               a == PrinterPort.PRN_STROBE ? PrinterPort.prnReadStrobe () :
               0);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a == PrinterPort.PRN_DATA) {
         PrinterPort.prnWriteData (d);
       } else if (a == PrinterPort.PRN_STROBE) {
         PrinterPort.prnWriteStrobe (d);
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_PRN

  //--------------------------------------------------------------------------------
  //MMD_SYS システムポート
  //
  //  $00E8E000～$00E8E00F(偶数)  バイトリードは$FF、バイトライトは$??、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E8E000～$00E8E00F(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E8E010～$00E8FFFF  $00E8E000～$00E8E00Fの繰り返し
  //
  MMD_SYS {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "システムポート" : "System Port";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       switch (a & 15) {
       case 0x01:
         return 0b11110000 | VideoController.vcnTargetContrastPort;
       case 0x03:
         return 0b11111000;
       case 0x07:
         return 0b11110100 | (Keyboard.kbdOn ? 8 : 0) | CRTC.crtHRLPort << 1;
       case 0x0b:
         return (XEiJ.mdlCurrentBody == 0 ?
                 XEiJ.mpuCoreType == 0 ?
                 XEiJ.mpuClockMHz <= 10.0 ? 0xff : 0xfe :
                 XEiJ.mpuClockMHz <= 20.0 ? 0xff : 0xfe :
                 0xdc);
       }
       return 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       d &= 0xff;
       switch (a & 15) {
       case 0x01:
         VideoController.vcnTargetContrastPort = d & 15;
         {
           int curr = VideoController.vcnTargetContrastMask == 0 ? VideoController.vcnTargetContrastPort : VideoController.vcnTargetContrastTest;
           if (VideoController.vcnTargetContrastCurr != curr) {
             VideoController.vcnTargetContrastCurr = curr;
             VideoController.vcnTargetScaledContrast = VideoController.VCN_CONTRAST_SCALE * VideoController.vcnTargetContrastCurr;
             CRTC.crtContrastClock = XEiJ.mpuClockTime;
             CRTC.crtFrameTaskClock = Math.min (CRTC.crtContrastClock, CRTC.crtCaptureClock);
           }
         }
         return;
       case 0x07:
         {
           CRTC.crtHRLPort = d >> 1 & 1;
           int curr = CRTC.crtHRLMask == 0 ? CRTC.crtHRLPort : CRTC.crtHRLTest;
           if (CRTC.crtHRLCurr != curr) {
             CRTC.crtHRLCurr = curr;
             CRTC.crtRestart ();
           }
           if ((d & 1 << 2) != 0) {
             XEiJ.sysResetNMI ();  //NMIリセット
           }
         }
         return;
       case 0x09:
         if (XEiJ.mdlCurrentBody != 0) {
           //X68030のとき
           //  d=ROMウェイト設定値<<4|RAMウェイト設定値
           //  ROMウェイト
           //    ROMウェイト設定値+2
           //  RAMウェイト
           //    RAMウェイト設定値が0のとき
           //      スタティックカラムが有効なとき0
           //      スタティックカラムが無効なとき4
           //    RAMウェイト設定値が0でないとき
           //      RAMウェイト設定値+2
           //  ROMとRAMにはキャッシュが効く
           //XEiJ.mpuRomWait30 = (d >> 4 & 15) + 2;
           //XEiJ.mpuRamWait30 = (d & 15) == 0 ? 0 : (d & 15) + 2;
         }
         return;
       case 0x0d:
         XEiJ.smrWriteEnableOn = d == 0x31;
         return;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_SYS

  //--------------------------------------------------------------------------------
  //MMD_OPM FM音源
  //
  //  $00E90000  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFFF、ワードライトは$??xx
  //  $00E90001  バイトリードは$FF、バイトライトは$xx
  //  $00E90002  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E90003  バイトリードは$xx、バイトライトは$xx
  //  $00E90004～$00E91FFF  $00E90000～$00E90003の繰り返し
  //
  MMD_OPM {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "FM 音源" : "FM Sound Generator";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & 0x03) == 3 ? (XEiJ.mpuClockTime < YM2151.opmBusyClock ? 0x80 : 0x00) | YM2151.opmISTB | YM2151.opmISTA : 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       a &= 3;
       if (a == 3) {  //データレジスタ
         //ビジーフラグをセットする
         //  データレジスタに書き込んでからビジーフラグがセットされるまでのタイムラグを再現することもできなくはないが、
         //  実機をクロックアップすると動かなくなるソフトが実機と同じように動かなくなるだけのような気がするのでやめておく
         YM2151.opmBusyClock = XEiJ.mpuClockTime + XEiJ.TMR_FREQ / YM2151.OPM_OSC_FREQ * 68;
         //除算の負荷を軽減するためYM2151.opmPointerを進める必要があるかどうかを先に確認する
         if (SoundSource.sndBlockClock - SoundSource.SND_BLOCK_TIME + YM2151.OPM_SAMPLE_TIME / SoundSource.SND_CHANNELS * YM2151.opmPointer < XEiJ.mpuClockTime) {
           //現在のブロックの残り時間が1サンプルの時間の倍数になるように切り上げる
           YM2151.opmUpdate (SoundSource.SND_CHANNELS * (YM2151.OPM_BLOCK_SAMPLES - Math.max (0, (int) ((double) (SoundSource.sndBlockClock - XEiJ.mpuClockTime) / (double) YM2151.OPM_SAMPLE_TIME))));
         }
         YM2151.opmSetData (YM2151.opmAddress, d);
       } else if (a == 1) {  //アドレスレジスタ
         //アドレスレジスタへの書き込みではビジーフラグをセットしない
         //! 未対応。X68030のときアドレスレジスタへの書き込みに限って約1.2μsのウェイトを追加する
         //  http://retropc.net/x68000/software/hardware/060turbo/060opmp/060opmp.htm
         YM2151.opmAddress = d & 0xff;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_OPM

  //--------------------------------------------------------------------------------
  //MMD_PCM ADPCM音源
  //
  //  $00E92000～$00E92003(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E92000～$00E92003(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E92004～$00E93FFF  $00E92000～$00E92003の繰り返し
  //
  MMD_PCM {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "ADPCM 音源" : "ADPCM Sound Generator";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & 0x03) == 1 ? (ADPCM.pcmActive ? 0b10000000 : 0) | 0x40 : 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       a &= 3;
       if (a == 1) {  //コマンド
         if ((d & 0b00000001) != 0) {  //動作終了
           if (ADPCM.pcmActive) {  //動作中
             ADPCM.pcmClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (SoundSource.sndPcmTicker);
             ADPCM.pcmActive = false;
             ADPCM.pcmEncodedData = -1;
             ADPCM.pcmDecoderPointer = 0;
             HD63450.dmaRisePCL (3);
           }
         } else if ((d & 0b00000010) != 0) {  //動作開始
           if (!ADPCM.pcmActive) {  //停止中
             //現在のブロックの残り時間が1サンプルの時間の倍数になるように切り上げる
             int remainingSamples = Math.max (0, (int) ((double) (SoundSource.sndBlockClock - XEiJ.mpuClockTime) / (double) ADPCM.PCM_SAMPLE_TIME));  //現在のブロックの残りサンプル数
             ADPCM.pcmClock = SoundSource.sndBlockClock - ADPCM.PCM_SAMPLE_TIME * (long) remainingSamples;  //書き込み開始時刻
             TickerQueue.tkqAdd (SoundSource.sndPcmTicker, ADPCM.pcmClock);
             ADPCM.pcmActive = true;
             int newPointer = SoundSource.SND_CHANNELS * (ADPCM.PCM_BLOCK_SAMPLES - remainingSamples);  //書き込み開始位置
             if (ADPCM.pcmPointer < newPointer) {
               ADPCM.pcmFillBuffer (newPointer);
             } else {
               ADPCM.pcmPointer = newPointer;  //少し戻る場合がある
             }
             //DMAに最初のデータを要求する
             HD63450.dmaFallPCL (3);
           }
           //} else if ((d & 0b00000100) != 0) {  //録音開始
           //! 非対応
         }
       } else if (a == 3) {  //データ
         if (ADPCM.pcmActive) {
           ADPCM.pcmEncodedData = d & 0xff;
           HD63450.dmaRisePCL (3);
         }
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_PCM

  //--------------------------------------------------------------------------------
  //MMD_FDC FDコントローラ
  //
  //  $00E94000～$00E94007(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E94000～$00E94007(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E94008～$00E95FFF  $00E94000～$00E94007の繰り返し
  //
  MMD_FDC {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "FD コントローラ" : "FD Controller";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      return (byte) mmdPbz (a);
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK & ~1;  //bit0を無視する
      switch (a) {
      case FDC.FDC_STATUS_PORT & ~1:  //0x00e94001
        return FDC.fdcPeekStatus ();
      case FDC.FDC_DATA_PORT & ~1:  //0x00e94003
        return FDC.fdcPeekData ();
      case FDC.FDC_DRIVE_STATUS & ~1:  //0x00e94005
        return FDC.fdcPeekDriveStatus ();
      }
      return 0xff;
    }
    @Override protected int mmdPws (int a) {
      return (short) (mmdPbz (a) << 8 | mmdPbz (a + 1));
    }
    @Override protected int mmdPwz (int a) {
      return mmdPbz (a) << 8 | mmdPbz (a + 1);
    }
    @Override protected int mmdPls (int a) {
      return mmdPwz (a) << 16 | mmdPwz (a + 2);
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK & ~1;  //bit0を無視する
       switch (a) {
       case FDC.FDC_STATUS_PORT & ~1:  //0x00e94001
         return FDC.fdcReadStatus ();
       case FDC.FDC_DATA_PORT & ~1:  //0x00e94003
         return FDC.fdcReadData ();
       case FDC.FDC_DRIVE_STATUS & ~1:  //0x00e94005
         return FDC.fdcReadDriveStatus ();
       }
       return 0xff;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK & ~1;  //bit0を無視する
       switch (a) {
       case FDC.FDC_STATUS_PORT & ~1:  //0x00e94001
         FDC.fdcWriteCommand (d);
         break;
       case FDC.FDC_DATA_PORT & ~1:  //0x00e94003
         FDC.fdcWriteData (d);
         break;
       case FDC.FDC_DRIVE_STATUS & ~1:  //0x00e94005
         FDC.fdcWriteDriveControl (d);
         break;
       case FDC.FDC_DRIVE_SELECT & ~1:  //0x00e94007
         FDC.fdcWriteDriveSelect (d);
         break;
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_FDC

  //--------------------------------------------------------------------------------
  //MMD_HDC SASI HDコントローラ
  //
  //  $00E96000～$00E9603F(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E96000～$00E9603F(奇数)  バイトリードは$xx、バイトライトは$xx
  //                              X68030の$00E96000～$00E9601Fは$00,$00,$FF,$FFの繰り返し。$00E9603Fは$FF
  //  $00E96040～$00E97FFF  $00E96000～$00E9603Fの繰り返し
  //
  MMD_HDC {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "内蔵 SASI/SCSI ポート" : "Internal SASI/SCSI Port";  //SCSIのIはInterfaceなのだからSCSIインタフェイスはおかしい
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if ((a & 0x01) == 0) {  //偶数番地
        return (byte) 0xff;
      } else {  //奇数番地
        if (SPC.spcSCSIINOn) {  //SCSI内蔵機
          if ((a & 0x20) == 0) {  //SASIポート
            return (byte) ((a & 0x02) == 0 ? 0x00 : 0xff);
          } else {  //SCSIポート
            return (byte) SPC.spcSCSIINChip.spiPeek (a);
          }
        } else {  //SASI内蔵機
          switch (a & 0x3f) {
          case HDC.HDC_DATA_PORT & 0x3f:  //0x00e96001
            return (byte) HDC.hdcPeekData ();
          case HDC.HDC_STATUS_PORT & 0x3f:  //0x00e96003
            return (byte) HDC.hdcPeekStatus ();
          case HDC.HDC_RESET_PORT & 0x3f:  //0x00e96005
            return (byte) 0xff;
          case HDC.HDC_SELECTION_PORT & 0x3f:  //0x00e96007
            return (byte) 0xff;
          default:
            return (byte) 0xff;
          }
        }
      }
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if ((a & 0x01) == 0) {  //偶数番地
        return 0xff;
      } else {  //奇数番地
        if (SPC.spcSCSIINOn) {  //SCSI内蔵機
          if ((a & 0x20) == 0) {  //SASIポート
            return ((a & 0x02) == 0 ? 0x00 : 0xff);
          } else {  //SCSIポート
            return SPC.spcSCSIINChip.spiPeek (a);
          }
        } else {  //SASI内蔵機
          switch (a & 0x3f) {
          case HDC.HDC_DATA_PORT & 0x3f:  //0x00e96001
            return HDC.hdcPeekData ();
          case HDC.HDC_STATUS_PORT & 0x3f:  //0x00e96003
            return HDC.hdcPeekStatus ();
          case HDC.HDC_RESET_PORT & 0x3f:  //0x00e96005
            return 0xff;
          case HDC.HDC_SELECTION_PORT & 0x3f:  //0x00e96007
            return 0xff;
          default:
            return 0xff;
          }
        }
      }
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return (short) (0xff00 | mmdPbz (a + 1));
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return 0xff00 | mmdPbz (a + 1);
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return 0xff00ff00 | mmdPbz (a + 1) << 16 | mmdPbz (a + 3);
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if ((a & 0x01) == 0) {  //偶数番地
         return super.mmdRbs (a);  //バスエラー
       } else {  //奇数番地
         if (SPC.spcSCSIINOn) {  //SCSI内蔵機
           if ((a & 0x20) == 0) {  //SASIポート
             return (byte) ((a & 0x02) == 0 ? 0x00 : 0xff);
           } else {  //SCSIポート
             return (byte) SPC.spcSCSIINChip.spiRead (a);
           }
         } else {  //SASI内蔵機
           switch (a & 0x3f) {
           case HDC.HDC_DATA_PORT & 0x3f:  //0x00e96001
             return (byte) HDC.hdcReadData ();
           case HDC.HDC_STATUS_PORT & 0x3f:  //0x00e96003
             return (byte) HDC.hdcReadStatus ();
           case HDC.HDC_RESET_PORT & 0x3f:  //0x00e96005
             return (byte) 0xff;
           case HDC.HDC_SELECTION_PORT & 0x3f:  //0x00e96007
             return (byte) 0xff;
           default:
             return (byte) 0xff;
           }
         }
       }
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if ((a & 0x01) == 0) {  //偶数番地
         return super.mmdRbz (a);  //バスエラー
       } else {  //奇数番地
         if (SPC.spcSCSIINOn) {  //SCSI内蔵機
           if ((a & 0x20) == 0) {  //SASIポート
             return ((a & 0x02) == 0 ? 0x00 : 0xff);
           } else {  //SCSIポート
             return SPC.spcSCSIINChip.spiRead (a);
           }
         } else {  //SASI内蔵機
           switch (a & 0x3f) {
           case HDC.HDC_DATA_PORT & 0x3f:  //0x00e96001
             return HDC.hdcReadData ();
           case HDC.HDC_STATUS_PORT & 0x3f:  //0x00e96003
             return HDC.hdcReadStatus ();
           case HDC.HDC_RESET_PORT & 0x3f:  //0x00e96005
             return 0xff;
           case HDC.HDC_SELECTION_PORT & 0x3f:  //0x00e96007
             return 0xff;
           default:
             return 0xff;
           }
         }
       }
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (short) (0xff00 | mmdRbz (a + 1));
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return 0xff00 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return 0xff00ff00 | mmdRbz (a + 1) << 16 | mmdRbz (a + 3);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if ((a & 0x01) == 0) {  //偶数番地
         super.mmdWb (a, d);  //バスエラー
       } else {  //奇数番地
         if (SPC.spcSCSIINOn) {  //SCSI内蔵機
           if ((a & 0x20) == 0) {  //SASIポート
             return;
           } else {  //SCSIポート
             SPC.spcSCSIINChip.spiWrite (a, d);
             return;
           }
         } else {  //SASI内蔵機
           switch (a & 0x3f) {
           case HDC.HDC_DATA_PORT & 0x3f:  //0x00e96001
             HDC.hdcWriteData (d);
             return;
           case HDC.HDC_STATUS_PORT & 0x3f:  //0x00e96003
             HDC.hdcWriteCommand (d);
             return;
           case HDC.HDC_RESET_PORT & 0x3f:  //0x00e96005
             HDC.hdcWriteReset (d);
             return;
           case HDC.HDC_SELECTION_PORT & 0x3f:  //0x00e96007
             HDC.hdcWriteSelect (d);
             return;
           default:
             return;
           }
         }
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a + 1, d >> 16);
       mmdWb (a + 3, d);
     }
  },  //MMD_HDC

  //--------------------------------------------------------------------------------
  //MMD_SCC SCC
  //
  //  $00E98000～$00E98007(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E98000～$00E98007(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E98008～$00E99FFF  $00E98000～$00E98007の繰り返し
  //
  MMD_SCC {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "SCC" : "SCC";  //Serial Communication Controller
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      try {
        //! 未対応
        return mmdRbs (a);
      } catch (M68kException e) {
      }
      return -1;
    }
    @Override protected int mmdPbz (int a) {
      try {
        //! 未対応
        return mmdRbz (a);
      } catch (M68kException e) {
      }
      return 0xff;
    }
    @Override protected int mmdPws (int a) {
      return 0xffffff00 | mmdPbz (a + 1);
    }
    @Override protected int mmdPwz (int a) {
      return 0xff00 | mmdPbz (a + 1);
    }
    @Override protected int mmdPls (int a) {
      return 0xff00ff00 | mmdPbz (a + 1) << 16 | mmdPbz (a + 3);
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       int d = 0;
       switch (a & 7) {
         //------------------------------------------------
       case Z8530.SCC_0_COMMAND & 7:  //ポートBコマンド読み出し
         switch (Z8530.scc0RegisterNumber) {
         case 0:  //RR0
           //  0x80  ブレークまたはアボート
           //  0x40  送信アンダーラン
           //  0x20  CTS(0=送信禁止,1=送信許可)
           //  0x10  SYNC
           //  0x08  DCD
           //  0x04  送信バッファ空
           //  0x02  ボーレートカウント0
           //  0x01  受信バッファフル
           d = Z8530.scc0InputCounter < 3 ? 0x25 : 0x24;
           break;
         case 2:  //RR2
           //修飾割り込みベクタ
           //  ポートBのRR2はWR2に設定したベクタを割り込み要求で加工して返す
           d = (Z8530.scc1aReceiveRequest != 0 ? Z8530.scc1aReceiveVector :  //1A受信バッファフル(RS-232C受信)
                Z8530.scc1aSendRequest != 0 ? Z8530.scc1aSendVector :  //1A送信バッファ空(RS-232C送信)
                Z8530.scc0bReceiveRequest != 0 ? Z8530.scc0bReceiveVector :  //0B受信バッファフル(マウス受信)
                Z8530.sccInterruptVector);
           break;
         case 3:  //RR3
           //ポートBのRR3は常に0
           //  ポートBの割り込みペンディングを見るときはポートAのRR3を参照する
           //d = 0;
           break;
         case 12:  //RR12
           d = Z8530.scc0BaudRateGen & 0xff;
           break;
         case 13:  //RR13
           d = Z8530.scc0BaudRateGen >> 8 & 0xff;
           break;
         default:
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.println ("unimplemented register");
           }
         }
         Z8530.scc0RegisterNumber = 0;
         break;
         //------------------------------------------------
       case Z8530.SCC_0_DATA & 7:  //ポートBデータ読み出し(マウス受信)
         if (Z8530.scc0InputCounter == 0) {  //1バイト目
           d = XEiJ.musExtraData;
           XEiJ.musExtraData = XEiJ.musData;
           if (!XEiJ.musOnScreen) {  //ホストのマウスカーソルがスクリーン上にない
             //XEiJ.musShow ();
             if (XEiJ.musCursorNumber != 1 && XEiJ.musCursorAvailable) {
               XEiJ.musCursorNumber = 1;
               XEiJ.pnlPanel.setCursor (XEiJ.musCursorArray[1]);  //ホストのマウスカーソルを表示する
             }
             Z8530.scc0Data1 = Z8530.scc0Data2 = 0;
           } else if (XEiJ.musSeamlessOn) {  //シームレス
             int on, dx, dy, coeff = 256;
             if (XEiJ.mpuCoreType < 6) {  //MMUなし
               if (Z8530.SCC_FSX_MOUSE &&
                   Z8530.sccFSXMouseHook != 0 &&  //FSX.Xが常駐している
                   MainMemory.mmrRls (0x0938) == Z8530.sccFSXMouseHook) {  //マウス受信データ処理ルーチンがFSX.Xを指している。SX-Windowが動作中
                 on = MainMemory.mmrRws (Z8530.sccFSXMouseWork + 0x2e) == 0 ? 1 : 0;  //SX-Windowのマウスカーソルの表示状態。Obscureのときは表示されているとみなす
                 int xy = MainMemory.mmrRls (Z8530.sccFSXMouseWork + 0x0a);
                 dx = (xy >> 16) - CRTC.crtR10TxXPort;  //SX-Windowのマウスカーソルの見かけのX座標
                 dy = (short) xy - CRTC.crtR11TxYPort;  //SX-Windowのマウスカーソルの見かけのY座標
                 coeff = MainMemory.mmrRwz (Z8530.sccFSXMouseWork + 0x04);  //ポインタの移動量。係数*256
               } else {  //SX-Windowが動作中ではない
                 on = MainMemory.mmrRbs (0x0aa2);  //IOCSのマウスカーソルの表示状態
                 int xy = MainMemory.mmrRls (0x0ace);
                 dx = xy >> 16;  //IOCSのマウスカーソルのX座標
                 dy = (short) xy;  //IOCSのマウスカーソルのY座標
               }
             } else {  //MMUあり
               if (Z8530.SCC_FSX_MOUSE &&
                   Z8530.sccFSXMouseHook != 0 &&  //FSX.Xが常駐している
                   MC68060.mmuPeekLongData (0x0938, 1) == Z8530.sccFSXMouseHook) {  //マウス受信データ処理ルーチンがFSX.Xを指している。SX-Windowが動作中
                 on = MC68060.mmuPeekWordSignData (Z8530.sccFSXMouseWork + 0x2e, 1) == 0 ? 1 : 0;  //SX-Windowのマウスカーソルの表示状態。Obscureのときは表示されているとみなす
                 int xy = MC68060.mmuPeekLongData (Z8530.sccFSXMouseWork + 0x0a, 1);
                 dx = (xy >> 16) - CRTC.crtR10TxXPort;  //SX-Windowのマウスカーソルの見かけのX座標
                 dy = (short) xy - CRTC.crtR11TxYPort;  //SX-Windowのマウスカーソルの見かけのY座標
                 coeff = MC68060.mmuPeekWordZeroData (Z8530.sccFSXMouseWork + 0x04, 1);  //ポインタの移動量。係数*256
               } else {  //SX-Windowが動作中ではない
                 on = MC68060.mmuPeekByteSignData (0x0aa2, 1);  //IOCSのマウスカーソルの表示状態
                 int xy = MC68060.mmuPeekLongData (0x0ace, 1);
                 dx = xy >> 16;  //IOCSのマウスカーソルのX座標
                 dy = (short) xy;  //IOCSのマウスカーソルのY座標
               }
             }
             dx = XEiJ.musScreenX - dx;  //X方向の移動量
             dy = XEiJ.musScreenY - dy;  //Y方向の移動量
             if (XEiJ.musEdgeAccelerationOn) {  //縁部加速を行う
               final int range = 10;  //加速領域の幅
               final int speed = 10;  //移動速度
               if (XEiJ.musScreenX < range) {
                 dx = -speed;  //左へ
               } else if (XEiJ.pnlScreenWidth - range <= XEiJ.musScreenX) {
                 dx = speed;  //右へ
               }
               if (XEiJ.musScreenY < range) {
                 dy = -speed;  //上へ
               } else if (XEiJ.pnlScreenHeight - range <= XEiJ.musScreenY) {
                 dy = speed;  //下へ
               }
             }
             if (on != 0) {  //X68000のマウスカーソルが表示されている
               //XEiJ.musHide ();
               if (XEiJ.musCursorNumber != 0 && XEiJ.musCursorAvailable) {
                 XEiJ.musCursorNumber = 0;
                 XEiJ.pnlPanel.setCursor (XEiJ.musCursorArray[0]);  //ホストのマウスカーソルを消す
               }
             } else {  //X68000のマウスカーソルが表示されていない
               //XEiJ.musShow ();
               if (XEiJ.musCursorNumber != 1 && XEiJ.musCursorAvailable) {
                 XEiJ.musCursorNumber = 1;
                 XEiJ.pnlPanel.setCursor (XEiJ.musCursorArray[1]);  //ホストのマウスカーソルを表示する
               }
             }
             if (coeff != 256 && coeff != 0) {
               //SX-Windowのポインタの移動量の補正
               dx = (dx << 8) / coeff;
               dy = (dy << 8) / coeff;
             }
             //  XEiJ.MUS_DEACCELERATION_TABLEの値が127を越えることはないのでシームレスでオーバーフローフラグがセットされることはない
             //  rbzで返すので負数のときのゼロ拡張を忘れないこと
             Z8530.scc0Data1 = (dx == 0 ? 0 : dx >= 0 ?
                          XEiJ.MUS_DEACCELERATION_TABLE[Math.min (1024, dx)] :
                          -XEiJ.MUS_DEACCELERATION_TABLE[Math.min (1024, -dx)] & 0xff);
             Z8530.scc0Data2 = (dy == 0 ? 0 : dy >= 0 ?
                          XEiJ.MUS_DEACCELERATION_TABLE[Math.min (1024, dy)] :
                          -XEiJ.MUS_DEACCELERATION_TABLE[Math.min (1024, -dy)] & 0xff);
           } else if (!XEiJ.frmIsActive) {  //エクスクルーシブだがフォーカスがない
             //XEiJ.musShow ();
             if (XEiJ.musCursorNumber != 1 && XEiJ.musCursorAvailable) {
               XEiJ.musCursorNumber = 1;
               XEiJ.pnlPanel.setCursor (XEiJ.musCursorArray[1]);  //ホストのマウスカーソルを表示する
             }
             Z8530.scc0Data1 = Z8530.scc0Data2 = 0;
             XEiJ.musExclusiveStart = true;  //フォーカスを得たときエクスクルーシブに切り替えた直後のように振る舞う
           } else {  //エクスクルーシブ
             //XEiJ.musHide ();
             if (XEiJ.musCursorNumber != 0 && XEiJ.musCursorAvailable) {
               XEiJ.musCursorNumber = 0;
               XEiJ.pnlPanel.setCursor (XEiJ.musCursorArray[0]);  //ホストのマウスカーソルを消す
             }
             int ox = XEiJ.pnlScreenX + (XEiJ.pnlZoomWidth >> 1);  //画面の中央
             int oy = XEiJ.pnlScreenY + (XEiJ.pnlZoomHeight >> 1);
             XEiJ.rbtRobot.mouseMove (XEiJ.pnlGlobalX + ox, XEiJ.pnlGlobalY + oy);  //マウスカーソルを画面の中央に戻す
             int dx = XEiJ.musPanelX - ox;
             int dy = XEiJ.musPanelY - oy;
             if (XEiJ.musExclusiveStart) {  //エクスクルーシブに切り替えた直後
               //エクスクルーシブに切り替えた直後の1回だけ相対位置を無視する
               //  エクスクルーシブに切り替える直前にマウスカーソルが画面の中央から離れていると切り替えた瞬間に画面の端に飛んでしまう
               dx = 0;
               dy = 0;
               XEiJ.musExclusiveStart = false;
             }
             //  上下左右のレスポンスが非対称になると気持ち悪いので冗長に書く
             //  rbzで返すので負数のときのゼロ拡張を忘れないこと
             if (dx != 0) {
               if (dx >= 0) {
                 //dx = dx * Z8530.scc0RatioX + 32768 >> 16;
                 dx = dx * Z8530.scc0RatioX >> 16;
                 if (127 < dx) {
                   d |= 0x10;
                   dx = 127;
                 }
               } else {
                 //dx = -(-dx * Z8530.scc0RatioX + 32768 >> 16);
                 dx = -(-dx * Z8530.scc0RatioX >> 16);
                 if (dx < -128) {
                   d |= 0x20;
                   dx = -128;
                 }
                 dx &= 0xff;
               }
             }
             if (dy != 0) {
               if (dy >= 0) {
                 //dy = dy * Z8530.scc0RatioY + 32768 >> 16;
                 dy = dy * Z8530.scc0RatioY >> 16;
                 if (127 < dy) {
                   d |= 0x40;
                   dy = 127;
                 }
               } else {
                 //dy = -(-dy * Z8530.scc0RatioY + 32768 >> 16);
                 dy = -(-dy * Z8530.scc0RatioY >> 16);
                 if (dy < -128) {
                   d |= 0x80;
                   dy = -128;
                 }
                 dy &= 0xff;
               }
             }
             Z8530.scc0Data1 = dx;
             Z8530.scc0Data2 = dy;
           }  //if シームレス else エクスクルーシブ
           Z8530.scc0InputCounter = 1;
           //d = d;
         } else if (Z8530.scc0InputCounter == 1) {  //2バイト目
           Z8530.scc0InputCounter = 2;
           d = Z8530.scc0Data1;
         } else if (Z8530.scc0InputCounter == 2) {  //3バイト目
           Z8530.scc0InputCounter = 3;
           d = Z8530.scc0Data2;
         }
         break;
         //------------------------------------------------
       case Z8530.SCC_1_COMMAND & 7:  //ポートAコマンド読み出し
         switch (Z8530.scc1RegisterNumber) {
         case 0:  //RR0
           //  0x80  ブレークまたはアボート
           //  0x40  送信アンダーラン
           //  0x20  CTS(0=送信禁止,1=送信許可)
           //  0x10  SYNC
           //  0x08  DCD
           //  0x04  送信バッファ空
           //  0x02  ボーレートカウント0
           //  0x01  受信バッファフル
           d = Z8530.scc1InputRead != Z8530.scc1InputWrite && Z8530.scc1InputClock <= XEiJ.mpuClockTime ? 0x25 : 0x24;
           break;
         case 2:  //RR2
           //非修飾割り込みベクタ
           //  ポートAのRR2はWR2に設定したベクタをそのまま返す
           d = Z8530.sccInterruptVector;
           break;
         case 3:  //RR3
           //割り込みペンディング
           //  RR3リクエストからインサービスまでの間セットされている
           //  許可されていない割り込みのビットはセットされない
           d = Z8530.scc1aReceiveRR3 | Z8530.scc1aSendRR3 | Z8530.scc0bReceiveRR3;
           break;
         case 12:  //RR12
           d = Z8530.scc1BaudRateGen & 0xff;
           break;
         case 13:  //RR13
           d = Z8530.scc1BaudRateGen >> 8 & 0xff;
           break;
         default:
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.println ("unimplemented register");
           }
         }
         Z8530.scc1RegisterNumber = 0;
         break;
         //------------------------------------------------
       case Z8530.SCC_1_DATA & 7:  //ポートAデータ読み出し(RS-232C受信)
         if (Z8530.scc1InputRead != Z8530.scc1InputWrite && Z8530.scc1InputClock <= XEiJ.mpuClockTime) {  //受信バッファが空ではなく、受信予定時刻になっているとき
           d = Z8530.scc1InputBuffer[Z8530.scc1InputRead];  //ゼロ拡張済み
           Z8530.scc1InputRead = Z8530.scc1InputRead + 1 & Z8530.SCC_1_INPUT_MASK;
           if (Z8530.scc1InputRead != Z8530.scc1InputWrite) {  //受信バッファが空にならなかったとき
             Z8530.scc1InputClock += Z8530.scc1Interval;  //受信バッファの先頭のデータの受信予定時刻
             TickerQueue.tkqAdd (Z8530.sccTicker, Z8530.scc1InputClock);  //割り込みを発生させる
           } else {  //受信バッファが空になったとき
             Z8530.scc1InputClock = XEiJ.FAR_FUTURE;
             TickerQueue.tkqRemove (Z8530.sccTicker);
           }
         }
         break;
         //------------------------------------------------
       default:
         return super.mmdRbz (a);  //バスエラー
       }
       if (Z8530.SCC_DEBUG_TRACE && Z8530.sccTraceOn) {
         System.out.printf ("%08x Z8530.sccRead(0x%08x)=0x%02x\n", XEiJ.regPC0, a, d);
       }
       return d;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       return 0xffffff00 | mmdRbz (a + 1);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       return 0xff00 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       return 0xff00ff00 | mmdRbz (a + 1) << 16 | mmdRbz (a + 3);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       d &= 0xff;
       if (Z8530.SCC_DEBUG_TRACE && Z8530.sccTraceOn) {
         System.out.printf ("%08x Z8530.sccWrite(0x%08x,0x%02x)\n", XEiJ.regPC0, a, d);
       }
       switch (a & 7) {
         //------------------------------------------------
       case Z8530.SCC_0_COMMAND & 7:  //ポートBコマンド書き込み
         switch (Z8530.scc0RegisterNumber) {
         case 0:  //WR0
           if ((d & 0xf0) == 0) {  //レジスタ選択
             Z8530.scc0RegisterNumber = d;
           } else if (d == 0x38) {  //IUSリセット。割り込み処理が終了し、次の割り込みを受け付ける
             if (Z8530.scc0bReceiveRR3 != 0) {
               Z8530.scc0bReceiveRR3 = 0;
               if (Z8530.scc0InputCounter < 3) {  //3バイト受信するまで割り込み要求を続ける
                 if (Z8530.scc0bReceiveMask != 0) {
                   Z8530.scc0bReceiveRR3 = Z8530.SCC_0B_RECEIVE_MASK;
                   Z8530.scc0bReceiveRequest = Z8530.SCC_0B_RECEIVE_MASK;
                   XEiJ.mpuIRR |= XEiJ.MPU_SCC_INTERRUPT_MASK;
                 }
               }
             }
           } else if (d == 0x10) {  //ステータス割り込みリセット
           } else if (d == 0x30) {  //エラーリセット
           } else if (d == 0x80) {  //送信CRCジェネレータリセット
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented command");
             }
           }
           return;
         case 1:  //WR1
           Z8530.scc0bReceiveMask = (d & 0x18) != 0 ? Z8530.SCC_0B_RECEIVE_MASK : 0;
           if ((d & 0xec) != 0x00) {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented interrupt mode");
             }
           }
           break;
         case 2:  //WR2
           Z8530.sccInterruptVector = d;  //割り込みベクタ
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.printf ("Z8530.sccInterruptVector=0x%02x\n", Z8530.sccInterruptVector);
           }
           Z8530.sccUpdateVector ();
           break;
         case 3:  //WR3
           if (d == 0xc0) {  //受信禁止
           } else if (d == 0xc1) {  //受信許可
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented receiver configuration");
             }
           }
           break;
         case 4:  //WR4
           break;
         case 5:  //WR5
           //  0x80  DTR
           //  0x60  ビット長(0x00=5bit,0x20=7bit,0x40=6bit,0x60=8bit)
           //  0x10  ブレーク
           //  0x08  送信許可
           //  0x04  CRC-16
           //  0x02  RTS
           //  0x01  送信CRC
           {
             int rts = d >> 1 & 1;
             if ((~Z8530.scc0Rts & rts) != 0) {  //RTS=0→1。MSCTRL=H→Lとなってマウスに送信要求が出される
               Z8530.scc0InputCounter = 0;
               //マウスデータ受信開始
               if (Z8530.scc0bReceiveMask != 0) {
                 Z8530.scc0bReceiveRR3 = Z8530.SCC_0B_RECEIVE_MASK;
                 Z8530.scc0bReceiveRequest = Z8530.SCC_0B_RECEIVE_MASK;
                 XEiJ.mpuIRR |= XEiJ.MPU_SCC_INTERRUPT_MASK;
               }
             }
             Z8530.scc0Rts = rts;
             if ((d & 0x75) == 0x60) {
             } else {
               if (Z8530.SCC_DEBUG_TRACE) {
                 System.out.println ("unimplemented sender configuration");
               }
             }
           }
           break;
         case 6:  //WR6
           break;
         case 7:  //WR7
           break;
         case 9:  //WR9
           if ((d & 0xc0) == 0x40) {  //ポートBリセット
             Z8530.scc0Rts = 0;
           } else if ((d & 0xc0) == 0x80) {  //ポートAリセット
           } else if ((d & 0xc0) == 0xc0) {  //ハードウェアリセット
             Z8530.scc0Rts = 0;
           }
           Z8530.sccVectorInclude = d & 0x11;
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.printf ("Z8530.sccVectorInclude=0x%02x\n", Z8530.sccVectorInclude);
           }
           Z8530.sccUpdateVector ();
           if ((d & 0x26) != 0x00) {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented interrupt configuration");
             }
           }
           break;
         case 10:  //WR10
           if (d == 0x00) {
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented SDLC configuration");
             }
           }
           break;
         case 11:  //WR11
           if (d == 0x50) {  //TRxCは入力
           } else if (d == 0x56) {  //TRxCからボーレートジェネレータを出力
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented clock control");
             }
           }
           break;
         case 12:  //WR12
           Z8530.scc0BaudRateGen = (Z8530.scc0BaudRateGen & ~0xff) | d;
           break;
         case 13:  //WR13
           Z8530.scc0BaudRateGen = d << 8 | (Z8530.scc0BaudRateGen & 0xff);
           break;
         case 14:  //WR14
           if (d == 0x02) {  //ボーレートジェネレータ停止
           } else if (d == 0x03) {  //ボーレートジェネレータ動作
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented DPLL configuration");
             }
           }
           break;
         case 15:  //WR15
           if (d == 0x00) {
           } else if (d == 0x80) {
           } else if (d == 0x88) {
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented status interrupt configuration");
             }
           }
           break;
         default:
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.println ("unimplemented register");
           }
         }
         Z8530.scc0RegisterNumber = 0;
         return;
         //------------------------------------------------
       case Z8530.SCC_0_DATA & 7:  //ポートBデータ書き込み(マウス送信)
         return;
         //------------------------------------------------
       case Z8530.SCC_1_COMMAND & 7:  //ポートAコマンド書き込み
         switch (Z8530.scc1RegisterNumber) {
         case 0:  //WR0
           if ((d & 0xf0) == 0) {  //レジスタ選択
             Z8530.scc1RegisterNumber = d;
           } else if (d == 0x38) {  //IUSリセット。割り込み処理が終了し、次の割り込みを受け付ける
             if (Z8530.scc1aReceiveRR3 != 0) {
               Z8530.scc1aReceiveRR3 = 0;
               if (Z8530.scc1InputRead != Z8530.scc1InputWrite) {  //バッファが空になるまで割り込み要求を続ける
                 TickerQueue.tkqAdd (Z8530.sccTicker, Z8530.scc1InputClock);  //受信バッファの先頭のデータの受信予定時刻
               }
             }
           } else if (d == 0x10) {  //ステータス割り込みリセット
           } else if (d == 0x30) {  //エラーリセット
           } else if (d == 0x80) {  //送信CRCジェネレータリセット
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented command");
             }
           }
           return;
         case 1:  //WR1
           Z8530.scc1aReceiveMask = (d & 0x18) != 0 ? Z8530.SCC_1A_RECEIVE_MASK : 0;
           Z8530.scc1aSendMask = (d & 0x02) != 0 ? Z8530.SCC_1A_SEND_MASK : 0;
           if ((d & 0xec) != 0x00) {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented interrupt mode");
             }
           }
           break;
         case 2:  //WR2
           Z8530.sccInterruptVector = d;  //割り込みベクタ
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.printf ("Z8530.sccInterruptVector=0x%02x\n", Z8530.sccInterruptVector);
           }
           Z8530.sccUpdateVector ();
           break;
         case 3:  //WR3
           if (d == 0xc0) {  //受信禁止
           } else if (d == 0xc1) {  //受信許可
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented receiver configuration");
             }
           }
           break;
         case 4:  //WR4
           Z8530.scc1ClockModeShift = d >> 6 == 0 ? 0 : (d >> 6) + 3;  //0=2^0,1=2^4,2=2^5,3=2^6
           Z8530.scc1Interval = XEiJ.TMR_FREQ / ((Z8530.SCC_FREQ / 2 >> Z8530.scc1ClockModeShift) / (Z8530.scc1BaudRateGen + 2));
           break;
         case 5:  //WR5
           //  0x80  DTR
           //  0x60  ビット長(0x00=5bit,0x20=7bit,0x40=6bit,0x60=8bit)
           //  0x10  ブレーク
           //  0x08  送信許可
           //  0x04  CRC-16
           //  0x02  RTS
           //  0x01  送信CRC
           {
             if ((d & 0x75) == 0x60) {
             } else {
               if (Z8530.SCC_DEBUG_TRACE) {
                 System.out.println ("unimplemented sender configuration");
               }
             }
           }
           break;
         case 6:  //WR6
           break;
         case 7:  //WR7
           break;
         case 9:  //WR9
           if ((d & 0xc0) == 0x40) {  //ポートBリセット
             Z8530.scc0Rts = 0;
           } else if ((d & 0xc0) == 0x80) {  //ポートAリセット
           } else if ((d & 0xc0) == 0xc0) {  //ハードウェアリセット
             Z8530.scc0Rts = 0;
           }
           Z8530.sccVectorInclude = d & 0x11;
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.printf ("Z8530.sccVectorInclude=0x%02x\n", Z8530.sccVectorInclude);
           }
           Z8530.sccUpdateVector ();
           if ((d & 0x2e) != 0x08) {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented interrupt configuration");
             }
           }
           break;
         case 10:  //WR10
           if (d == 0x00) {
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented SDLC configuration");
             }
           }
           break;
         case 11:  //WR11
           if (d == 0x50) {  //TRxCは入力
           } else if (d == 0x56) {  //TRxCからボーレートジェネレータを出力
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented clock control");
             }
           }
           break;
         case 12:  //WR12
           Z8530.scc1BaudRateGen = (Z8530.scc1BaudRateGen & ~0xff) | d;
           Z8530.scc1Interval = XEiJ.TMR_FREQ / ((Z8530.SCC_FREQ / 2 >> Z8530.scc1ClockModeShift) / (Z8530.scc1BaudRateGen + 2));
           break;
         case 13:  //WR13
           Z8530.scc1BaudRateGen = d << 8 | (Z8530.scc1BaudRateGen & 0xff);
           Z8530.scc1Interval = XEiJ.TMR_FREQ / ((Z8530.SCC_FREQ / 2 >> Z8530.scc1ClockModeShift) / (Z8530.scc1BaudRateGen + 2));
           break;
         case 14:  //WR14
           if (d == 0x02) {  //ボーレートジェネレータ停止
           } else if (d == 0x03) {  //ボーレートジェネレータ動作
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented DPLL configuration");
             }
           }
           break;
         case 15:  //WR15
           if (d == 0x00) {
           } else if (d == 0x80) {
           } else if (d == 0x88) {
           } else {
             if (Z8530.SCC_DEBUG_TRACE) {
               System.out.println ("unimplemented status interrupt configuration");
             }
           }
           break;
         default:
           if (Z8530.SCC_DEBUG_TRACE) {
             System.out.println ("unimplemented register");
           }
         }
         Z8530.scc1RegisterNumber = 0;
         return;
         //------------------------------------------------
       case Z8530.SCC_1_DATA & 7:  //ポートAデータ書き込み(RS-232C送信)
         RS232CTerminal.trmPrintSJIS (d);
         return;
         //------------------------------------------------
       default:
         super.mmdWb (a, d);  //バスエラー
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       mmdWb (a + 1, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       mmdWb (a + 1, d >> 16);
       mmdWb (a + 3, d);
     }
  },  //MMD_SCC

  //--------------------------------------------------------------------------------
  //MMD_PPI PPI
  //
  //  $00E9A000～$00E9A007(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E9A000～$00E9A007(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E9A008～$00E9BFFF  $00E9A000～$00E9A007の繰り返し
  //
  MMD_PPI {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "PPI" : "PPI";  //Programmable Peripheral Interface
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       return PPI.ppiRbz (a);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       PPI.ppiWb (a, d);
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_PPI

  //--------------------------------------------------------------------------------
  //MMD_IOI I/O割り込み
  //
  //  $00E9C000～$00E9C00F(偶数)  バイトリードはバスエラー、バイトライトはバスエラー、ワードリードは$FFxx、ワードライトは$??xx
  //  $00E9C000～$00E9C00F(奇数)  バイトリードは$xx、バイトライトは$xx
  //  $00E9C010～$00E9BFFF  $00E9C000～$00E9C00Fの繰り返し
  //
  MMD_IOI {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "I/O 割り込み" : "I/O Interrupt";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a == IOInterrupt.IOI_STATUS) {
         return IOInterrupt.ioiReadStatus ();
       } else if (a == IOInterrupt.IOI_VECTOR) {  //ライトオンリー。リードすると$FFが返る
         //return IOInterrupt.ioiReadVector ();
         return 0xff;
       } else {
         return super.mmdRbz (a);  //バスエラー
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a == IOInterrupt.IOI_STATUS) {
         IOInterrupt.ioiWriteEnable (d);
       } else if (a == IOInterrupt.IOI_VECTOR) {
         IOInterrupt.ioiWriteVector (d);
       } else {
         super.mmdWb (a, d);  //バスエラー
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_IOI

  //--------------------------------------------------------------------------------
  //MMD_XB1 拡張ボード領域1
  MMD_XB1 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張ボード領域 1" : "Expansion Board Area 1";
    }
    //ピーク
    @Override protected int mmdPbz (int a) {
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
        return XEiJ.fpuCoproboard1.cirPeekByteZero (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
        return XEiJ.fpuCoproboard2.cirPeekByteZero (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
        return MainMemory.mmrM8[a] & 0xff;
      }
      return 0xff;
    }
    @Override protected int mmdPwz (int a) {
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
        return XEiJ.fpuCoproboard1.cirPeekWordZero (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
        return XEiJ.fpuCoproboard2.cirPeekWordZero (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
      return 0xffff;
    }
    @Override protected int mmdPls (int a) {
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
        return XEiJ.fpuCoproboard1.cirPeekLong (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
        return XEiJ.fpuCoproboard2.cirPeekLong (a);
      }
      if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
      return -1;
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         return XEiJ.fpuCoproboard1.cirReadByteZero (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         return XEiJ.fpuCoproboard2.cirReadByteZero (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
         return MainMemory.mmrM8[a] & 0xff;
       }
       return super.mmdRbz (a);  //バスエラー
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         return XEiJ.fpuCoproboard1.cirReadWordZero (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         return XEiJ.fpuCoproboard2.cirReadWordZero (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
       return super.mmdRwz (a);  //バスエラー
     }
    @Override protected int mmdRls (int a) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         return XEiJ.fpuCoproboard1.cirReadLong (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         return XEiJ.fpuCoproboard2.cirReadLong (a);
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -HFS.HFS_ROM_SIZE)) == HFS.HFS_ADDRESS) {  //ホストファイルシステムインタフェイス
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
       return super.mmdRls (a);  //バスエラー
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         XEiJ.fpuCoproboard1.cirWriteByte (a, d);
         return;
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         XEiJ.fpuCoproboard2.cirWriteByte (a, d);
         return;
       }
       super.mmdWb (a, d);  //バスエラー
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         XEiJ.fpuCoproboard1.cirWriteWord (a, d);
         return;
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         XEiJ.fpuCoproboard2.cirWriteWord (a, d);
         return;
       }
       super.mmdWw (a, d);  //バスエラー
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e000) {  //数値演算プロセッサボード1
         XEiJ.fpuCoproboard1.cirWriteLong (a, d);
         return;
       }
       if ((a & (XEiJ.BUS_MOTHER_MASK & -0x20)) == 0x00e9e080) {  //数値演算プロセッサボード2
         XEiJ.fpuCoproboard2.cirWriteLong (a, d);
         return;
       }
       super.mmdWl (a, d);  //バスエラー
     }
  },  //MMD_XB1

  //--------------------------------------------------------------------------------
  //MMD_EXS 拡張SCSI
  //  必要なときだけ接続される
  //  拡張SCSIのROMのサイズは8KBなのでリードのときのバスエラーのチェックは不要
  //  ライトのときはROMには書き込めないのでSPCのレジスタでなければバスエラー
  MMD_EXS {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張 SCSI ポート" : "Expansion SCSI Port";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
       return (a & -32) == SPC.SPC_BASE_EX ? (byte) SPC.spcSCSIEXChip.spiPeek (a) : MainMemory.mmrM8[a];
     }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
       return (a & -32) == SPC.SPC_BASE_EX ? SPC.spcSCSIEXChip.spiPeek (a) : MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
       return mmdPbs (a) << 8 | mmdPbz (a + 1);
     }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
       return mmdPbz (a) << 8 | mmdPbz (a + 1);
     }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
       return mmdPbs (a) << 24 | mmdPbz (a + 1) << 16 | mmdPbz (a + 2) << 8 | mmdPbz (a + 3);
     }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & -32) == SPC.SPC_BASE_EX ? (byte) SPC.spcSCSIEXChip.spiRead (a) : MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & -32) == SPC.SPC_BASE_EX ? SPC.spcSCSIEXChip.spiRead (a) : MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbs (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbz (a) << 8 | mmdRbz (a + 1);
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRbs (a) << 24 | mmdRbz (a + 1) << 16 | mmdRbz (a + 2) << 8 | mmdRbz (a + 3);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if ((a & -32) == SPC.SPC_BASE_EX) {
         SPC.spcSCSIEXChip.spiWrite (a, d);
         return;
       }
       super.mmdWb (a, d);  //バスエラー
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWb (a    , d >> 8);
       mmdWb (a + 1, d     );
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_EXS

  //--------------------------------------------------------------------------------
  //MMD_XB2 拡張ボード領域2
  //
  //  $00EAFF80  リードはバスエラー、バイトライトはバスエラー、ワードライトは$00200000～$003FFFFFがあるとき$??xx、ないときバスエラー
  //  $00EAFF81  リードはバスエラー、ライトは$00200000～$003FFFFFがあるとき$xx、ないときバスエラー
  //  $00EAFF82  リードはバスエラー、バイトライトはバスエラー、ワードライトは$00400000～$005FFFFFがあるとき$??xx、ないときバスエラー
  //  $00EAFF83  リードはバスエラー、ライトは$00400000～$005FFFFFがあるとき$xx、ないときバスエラー
  //  $00EAFF84  リードはバスエラー、バイトライトはバスエラー、ワードライトは$00600000～$007FFFFFがあるとき$??xx、ないときバスエラー
  //  $00EAFF85  リードはバスエラー、ライトは$00600000～$007FFFFFがあるとき$xx、ないときバスエラー
  //  $00EAFF86  リードはバスエラー、バイトライトはバスエラー、ワードライトは$00800000～$009FFFFFがあるとき$??xx、ないときバスエラー
  //  $00EAFF87  リードはバスエラー、ライトは$00800000～$009FFFFFがあるとき$xx、ないときバスエラー
  //  $00EAFF88  リードはバスエラー、バイトライトはバスエラー、ワードライトは$00A00000～$00BFFFFFがあるとき$??xx、ないときバスエラー
  //  $00EAFF89  リードはバスエラー、ライトは$00A00000～$00BFFFFFがあるとき$xx、ないときバスエラー
  //  $00EAFF8A～$00EAFF8F  バスエラー
  //  $00EAFF90～$00EAFFFF  $00EAFF80～$00EAFF8Fの繰り返し
  //
  //  拡張
  //  $00EAFF7F  バンクメモリのページ番号
  //
  MMD_XB2 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張ボード領域 2" : "Expansion Board Area 2";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (0x00eaff81 <= a && a <= 0x00eaff89 && (a & 1) != 0) {  //スーパーバイザエリア設定ポート
         return MainMemory.mmrM8[a] & 0xff;  //読み出せるようにしておく(本来はライトオンリー)
       }
       if (XEiJ.bnkOn && a == 0x00eaff7f) {  //バンクメモリのページ番号
         return XEiJ.bnkPageStart >> 17;
       }
       return super.mmdRbz (a);  //バスエラー
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (0x00eaff81 <= a && a <= 0x00eaff89 && (a & 1) != 0) {  //スーパーバイザエリア設定ポート
         MainMemory.mmrM8[a] = (byte) d;  //読み出せるようにしておく(本来はライトオンリー)
         a = (a & 14) + 2 << 20;  //1,3,5,7,9→2,4,6,8,a
         for (int m = 1; m <= 128; m <<= 1) {
           if ((d & m) == 0) {  //ユーザエリア
             XEiJ.busUser ( MemoryMappedDevice.MMD_MMR, a, a + 0x00040000);
           } else {  //スーパーバイザエリア
             XEiJ.busSuper (MemoryMappedDevice.MMD_MMR, a, a + 0x00040000);
           }
           a += 0x00040000;
         }
         return;
       }
       if (XEiJ.bnkOn && a == 0x00eaff7f) {  //バンクメモリのページ番号
         XEiJ.bnkPageStart = (d & 0xff) << 17;
         return;
       }
       super.mmdWb (a, d);  //バスエラー
     }
  },  //MMD_XB2

  //--------------------------------------------------------------------------------
  //MMD_SPR スプライト画面
  //
  //  CRTCのR20の下位5ビットが%1??1?でないとき
  //    $00EB0000-$00EB03FF  スプライトスクロールレジスタ
  //    $00EB0400-$00EB07FF  $FF
  //    $00EB0800-$00EB0811  設定
  //    $00EB0812-$00EB7FFF  $FF
  //    $00EB8000-$00EBFFFF  スプライトPCG・テキストエリア
  //
  //  CRTCのR20の下位5ビットが%1??1?のとき
  //    $00EB0000-$00EB03FF  バスエラー
  //    $00EB0400-$00EB07FF  バスエラー
  //    $00EB0800-$00EB0811  設定
  //    $00EB0812-$00EB7FFF  $FF
  //    $00EB8000-$00EBFFFF  バスエラー
  //
  //  メモ
  //    スプライトPCG・テキストエリアにバイトサイズで書き込むとデータが破壊される
  //
  MMD_SPR {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "スプライト画面" : "Sprite Screen";
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & 1) == 0 ? (byte) (mmdRwz (a) >> 8) : (byte) mmdRwz (a - 1);
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (a & 1) == 0 ? mmdRwz (a) >> 8 : mmdRwz (a - 1) & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return (short) mmdRwz (a);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a < 0x00eb0400) {  //スプライトスクロールレジスタ
         if (!SpriteScreen.sprAccessible) {
           return super.mmdRwz (a);  //バスエラー
         }
         int n = (a >> 3) & 0x7f;
         switch (a & 0x06) {
         case 0:  //x座標
           return SpriteScreen.sprX[n];
         case 2:  //y座標
           return SpriteScreen.sprY[n];
         case 4:  //パターン番号、パレットブロック、水平反転、垂直反転
           return (SpriteScreen.sprV[n] ? 0x8000 : 0) | (SpriteScreen.sprH[n] ? 0x4000 : 0) | SpriteScreen.sprColPort[n] << 4 | SpriteScreen.sprNum[n];
         case 6:  //プライオリティ
           return SpriteScreen.sprPrw[n];
         }
       } else if (a < 0x00eb0800) {  //$FF
         if (!SpriteScreen.sprAccessible) {
           return super.mmdRwz (a);  //バスエラー
         }
         return 0xffff;
       } else if (a < 0x00eb8000) {  //各種レジスタ
         switch (a) {
         case SpriteScreen.SPR_REG0_BG0_X:
           return SpriteScreen.sprReg0Bg0XPort;
         case SpriteScreen.SPR_REG1_BG0_Y:
           return SpriteScreen.sprReg1Bg0YPort;
         case SpriteScreen.SPR_REG2_BG1_X:
           return SpriteScreen.sprReg2Bg1XPort;
         case SpriteScreen.SPR_REG3_BG1_Y:
           return SpriteScreen.sprReg3Bg1YPort;
         case SpriteScreen.SPR_REG4_BG_CTRL:
           return SpriteScreen.sprReg4BgCtrlPort;
         case SpriteScreen.SPR_REG5_H_FRONT_END:
           return SpriteScreen.sprReg5HFrontEndPort;
         case SpriteScreen.SPR_REG6_H_BACK_END:
           return SpriteScreen.sprReg6HBackEndPort;
         case SpriteScreen.SPR_REG7_V_BACK_END:
           return SpriteScreen.sprReg7VBackEndPort;
         case SpriteScreen.SPR_REG8_RESO:
           return SpriteScreen.sprReg8ResoPort;
         }
         return 0xffff;
       } else {  //PCGエリアとテキストエリア
         if (!SpriteScreen.sprAccessible) {
           return super.mmdRwz (a);  //バスエラー
         }
         int t = a >> 2 & 0x1fff;
         return (a & 0x02) == 0 ? SpriteScreen.sprPatPort[t] >>> 16 : (char) SpriteScreen.sprPatPort[t];
       }
       return 0;
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       return mmdRwz (a) << 16 | mmdRwz (a + 2);
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a < 0x00eb0800 || 0x00eb8000 <= a) {
         mmdWw (a & ~1, d << 8 | (d & 0xff));  //上位バイトと下位バイトに同じ値を書き込む
       } else {
         if ((a & 1) == 0) {
           mmdWw (a, d << 8 | (mmdRwz (a) & 0xff));
         } else {
           mmdWw (a - 1, (mmdRwz (a - 1) & ~0xff) | (d & 0xff));
         }
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (a < 0x00eb0400) {  //スプライトスクロールレジスタ
         if (!SpriteScreen.sprAccessible) {
           super.mmdWw (a, d);  //バスエラー
         }
         int n = a >> 3 & 0x7f;  //スプライト番号
         switch (a & 0x06) {
         case 0:  //x座標
           SpriteScreen.sprX[n] = (short) (d & 1023);
/*
           if (SpriteScreen.sprPrw[n] != 0 &&
               VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
               SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
             int y = SpriteScreen.sprY[n];
             if (true) {
               if (y < 16) {
                 for (int i = 0; i < y; i++) {
                   CRTC.crtRasterStamp[i] = 0;
                 }
               } else {
                 CRTC.crtRasterStamp[y - 16] = 0;
                 CRTC.crtRasterStamp[y - 15] = 0;
                 CRTC.crtRasterStamp[y - 14] = 0;
                 CRTC.crtRasterStamp[y - 13] = 0;
                 CRTC.crtRasterStamp[y - 12] = 0;
                 CRTC.crtRasterStamp[y - 11] = 0;
                 CRTC.crtRasterStamp[y - 10] = 0;
                 CRTC.crtRasterStamp[y -  9] = 0;
                 CRTC.crtRasterStamp[y -  8] = 0;
                 CRTC.crtRasterStamp[y -  7] = 0;
                 CRTC.crtRasterStamp[y -  6] = 0;
                 CRTC.crtRasterStamp[y -  5] = 0;
                 CRTC.crtRasterStamp[y -  4] = 0;
                 CRTC.crtRasterStamp[y -  3] = 0;
                 CRTC.crtRasterStamp[y -  2] = 0;
                 CRTC.crtRasterStamp[y -  1] = 0;
               }
             } else {
               Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
             }
           }
*/
           break;
         case 2:  //y座標
           d &= 1023;
           if (SpriteScreen.sprY[n] != d) {
             int y = SpriteScreen.sprY[n];
             SpriteScreen.sprY[n] = (short) d;
             if (SpriteScreen.sprPrw[n] != 0) {
               if (SpriteScreen.SPR_RRMAP) {
                 int mask = ~(0x80000000 >>> n);  //intのシフトカウントは5bitでマスクされる
                 int i = y << 2 | n >> 5;
                 SpriteScreen.sprRRmap[i            ] &= mask;
                 SpriteScreen.sprRRmap[i + ( 1 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 2 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 3 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 4 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 5 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 6 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 7 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 8 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 9 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (10 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (11 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (12 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (13 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (14 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (15 << 2)] &= mask;
                 mask = ~mask;
                 i = d << 2 | n >> 5;
                 SpriteScreen.sprRRmap[i            ] |= mask;
                 SpriteScreen.sprRRmap[i + ( 1 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 2 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 3 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 4 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 5 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 6 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 7 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 8 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 9 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (10 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (11 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (12 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (13 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (14 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (15 << 2)] |= mask;
               } else {  //!SpriteScreen.SPR_RRMAP
                 int[] map = SpriteScreen.sprRmap[n >> 5];
                 int mask = ~(1 << n);  //intのシフトカウントは5bitでマスクされる
                 map[y     ] &= mask;
                 map[y +  1] &= mask;
                 map[y +  2] &= mask;
                 map[y +  3] &= mask;
                 map[y +  4] &= mask;
                 map[y +  5] &= mask;
                 map[y +  6] &= mask;
                 map[y +  7] &= mask;
                 map[y +  8] &= mask;
                 map[y +  9] &= mask;
                 map[y + 10] &= mask;
                 map[y + 11] &= mask;
                 map[y + 12] &= mask;
                 map[y + 13] &= mask;
                 map[y + 14] &= mask;
                 map[y + 15] &= mask;
                 mask = ~mask;
                 map[d     ] |= mask;
                 map[d +  1] |= mask;
                 map[d +  2] |= mask;
                 map[d +  3] |= mask;
                 map[d +  4] |= mask;
                 map[d +  5] |= mask;
                 map[d +  6] |= mask;
                 map[d +  7] |= mask;
                 map[d +  8] |= mask;
                 map[d +  9] |= mask;
                 map[d + 10] |= mask;
                 map[d + 11] |= mask;
                 map[d + 12] |= mask;
                 map[d + 13] |= mask;
                 map[d + 14] |= mask;
                 map[d + 15] |= mask;
               }  //if SpriteScreen.SPR_RRMAP/!SpriteScreen.SPR_RRMAP
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 if (true) {
                   if (y < 16) {
                     for (int i = 0; i < y; i++) {
                       CRTC.crtRasterStamp[i] = 0;
                     }
                   } else {
                     CRTC.crtRasterStamp[y - 16] = 0;
                     CRTC.crtRasterStamp[y - 15] = 0;
                     CRTC.crtRasterStamp[y - 14] = 0;
                     CRTC.crtRasterStamp[y - 13] = 0;
                     CRTC.crtRasterStamp[y - 12] = 0;
                     CRTC.crtRasterStamp[y - 11] = 0;
                     CRTC.crtRasterStamp[y - 10] = 0;
                     CRTC.crtRasterStamp[y -  9] = 0;
                     CRTC.crtRasterStamp[y -  8] = 0;
                     CRTC.crtRasterStamp[y -  7] = 0;
                     CRTC.crtRasterStamp[y -  6] = 0;
                     CRTC.crtRasterStamp[y -  5] = 0;
                     CRTC.crtRasterStamp[y -  4] = 0;
                     CRTC.crtRasterStamp[y -  3] = 0;
                     CRTC.crtRasterStamp[y -  2] = 0;
                     CRTC.crtRasterStamp[y -  1] = 0;
                   }
                 } else {
                   Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
                 }
                 if (true) {
                   if (d < 16) {
                     for (int i = 0; i < d; i++) {
                       CRTC.crtRasterStamp[i] = 0;
                     }
                   } else {
                     CRTC.crtRasterStamp[d - 16] = 0;
                     CRTC.crtRasterStamp[d - 15] = 0;
                     CRTC.crtRasterStamp[d - 14] = 0;
                     CRTC.crtRasterStamp[d - 13] = 0;
                     CRTC.crtRasterStamp[d - 12] = 0;
                     CRTC.crtRasterStamp[d - 11] = 0;
                     CRTC.crtRasterStamp[d - 10] = 0;
                     CRTC.crtRasterStamp[d -  9] = 0;
                     CRTC.crtRasterStamp[d -  8] = 0;
                     CRTC.crtRasterStamp[d -  7] = 0;
                     CRTC.crtRasterStamp[d -  6] = 0;
                     CRTC.crtRasterStamp[d -  5] = 0;
                     CRTC.crtRasterStamp[d -  4] = 0;
                     CRTC.crtRasterStamp[d -  3] = 0;
                     CRTC.crtRasterStamp[d -  2] = 0;
                     CRTC.crtRasterStamp[d -  1] = 0;
                   }
                 } else {
                   Arrays.fill (CRTC.crtRasterStamp, d < 16 ? 0 : d - 16, d, 0);
                 }
               }
*/
             }
           }
           break;
         case 4:  //パターン番号、パレットブロック、水平反転、垂直反転
           {
             int num = SpriteScreen.sprNum[n];
             SpriteScreen.sprNum[n] = (short) (d & 0xff);
             SpriteScreen.sprColPort[n] = (short) (d >> 4 & 0x0f << 4);
             SpriteScreen.sprH[n] = (short) (d << 1) < 0;
             SpriteScreen.sprV[n] = (short) d < 0;
             if (SpriteScreen.sprPrw[n] != 0) {
               if (SpriteScreen.SPR_RRMAP) {
                 int mask = 0x80000000 >>> n;  //intのシフトカウントは5bitでマスクされる
                 SpriteScreen.sprPPmap[num << 2 | n >> 5] -= mask;
                 SpriteScreen.sprPPmap[(d & 0xff) << 2 | n >> 5] += mask;
               } else {
                 int[] map = SpriteScreen.sprPmap[n >> 5];
                 int mask = 1 << n;  //intのシフトカウントは5bitでマスクされる
                 map[num] -= mask;
                 map[d & 0xff] += mask;
               }
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 int y = SpriteScreen.sprY[n];
                 if (true) {
                   if (y < 16) {
                     for (int i = 0; i < y; i++) {
                       CRTC.crtRasterStamp[i] = 0;
                     }
                   } else {
                     CRTC.crtRasterStamp[y - 16] = 0;
                     CRTC.crtRasterStamp[y - 15] = 0;
                     CRTC.crtRasterStamp[y - 14] = 0;
                     CRTC.crtRasterStamp[y - 13] = 0;
                     CRTC.crtRasterStamp[y - 12] = 0;
                     CRTC.crtRasterStamp[y - 11] = 0;
                     CRTC.crtRasterStamp[y - 10] = 0;
                     CRTC.crtRasterStamp[y -  9] = 0;
                     CRTC.crtRasterStamp[y -  8] = 0;
                     CRTC.crtRasterStamp[y -  7] = 0;
                     CRTC.crtRasterStamp[y -  6] = 0;
                     CRTC.crtRasterStamp[y -  5] = 0;
                     CRTC.crtRasterStamp[y -  4] = 0;
                     CRTC.crtRasterStamp[y -  3] = 0;
                     CRTC.crtRasterStamp[y -  2] = 0;
                     CRTC.crtRasterStamp[y -  1] = 0;
                   }
                 } else {
                   Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
                 }
               }
*/
             }
           }
           break;
         case 6:  //プライオリティ
           d &= 0x0003;
           int prw = SpriteScreen.sprPrw[n];
           SpriteScreen.sprPrw[n] = (byte) d;
           if (prw != d) {
             if (prw == 0) {  //出現
               int y = SpriteScreen.sprY[n];
               if (SpriteScreen.SPR_RRMAP) {
                 int mask = 0x80000000 >>> n;  //intのシフトカウントは5bitでマスクされる
                 int i = y << 2 | n >> 5;
                 SpriteScreen.sprRRmap[i            ] |= mask;
                 SpriteScreen.sprRRmap[i + ( 1 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 2 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 3 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 4 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 5 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 6 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 7 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 8 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + ( 9 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (10 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (11 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (12 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (13 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (14 << 2)] |= mask;
                 SpriteScreen.sprRRmap[i + (15 << 2)] |= mask;
                 SpriteScreen.sprPPmap[SpriteScreen.sprNum[n] << 2 | n >> 5] |= mask;
               } else {  //!SpriteScreen.SPR_RRMAP
                 int map[] = SpriteScreen.sprRmap[n >> 5];
                 int mask = 1 << n;  //intのシフトカウントは5bitでマスクされる
                 map[y     ] |= mask;
                 map[y +  1] |= mask;
                 map[y +  2] |= mask;
                 map[y +  3] |= mask;
                 map[y +  4] |= mask;
                 map[y +  5] |= mask;
                 map[y +  6] |= mask;
                 map[y +  7] |= mask;
                 map[y +  8] |= mask;
                 map[y +  9] |= mask;
                 map[y + 10] |= mask;
                 map[y + 11] |= mask;
                 map[y + 12] |= mask;
                 map[y + 13] |= mask;
                 map[y + 14] |= mask;
                 map[y + 15] |= mask;
                 SpriteScreen.sprPmap[n >> 5][SpriteScreen.sprNum[n]] |= mask;
               }  //if SpriteScreen.SPR_RRMAP/!SpriteScreen.SPR_RRMAP
             } else if (d == 0) {  //消滅
               int y = SpriteScreen.sprY[n];
               if (SpriteScreen.SPR_RRMAP) {
                 int mask = ~(0x80000000 >>> n);  //intのシフトカウントは5bitでマスクされる
                 int i = y << 2 | n >> 5;
                 SpriteScreen.sprRRmap[i            ] &= mask;
                 SpriteScreen.sprRRmap[i + ( 1 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 2 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 3 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 4 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 5 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 6 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 7 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 8 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + ( 9 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (10 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (11 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (12 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (13 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (14 << 2)] &= mask;
                 SpriteScreen.sprRRmap[i + (15 << 2)] &= mask;
                 SpriteScreen.sprPPmap[SpriteScreen.sprNum[n] << 2 | n >> 5] &= mask;
               } else {  //!SpriteScreen.SPR_RRMAP
                 int map[] = SpriteScreen.sprRmap[n >> 5];
                 int mask = ~(1 << n);  //intのシフトカウントは5bitでマスクされる
                 map[y     ] &= mask;
                 map[y +  1] &= mask;
                 map[y +  2] &= mask;
                 map[y +  3] &= mask;
                 map[y +  4] &= mask;
                 map[y +  5] &= mask;
                 map[y +  6] &= mask;
                 map[y +  7] &= mask;
                 map[y +  8] &= mask;
                 map[y +  9] &= mask;
                 map[y + 10] &= mask;
                 map[y + 11] &= mask;
                 map[y + 12] &= mask;
                 map[y + 13] &= mask;
                 map[y + 14] &= mask;
                 map[y + 15] &= mask;
                 SpriteScreen.sprPmap[n >> 5][SpriteScreen.sprNum[n]] &= mask;
               }  //if SpriteScreen.SPR_RRMAP/!SpriteScreen.SPR_RRMAP
             }
/*
             if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                 SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
               int y = SpriteScreen.sprY[n];
               if (true) {
                 if (y < 16) {
                   for (int i = 0; i < y; i++) {
                     CRTC.crtRasterStamp[i] = 0;
                   }
                 } else {
                   CRTC.crtRasterStamp[y - 16] = 0;
                   CRTC.crtRasterStamp[y - 15] = 0;
                   CRTC.crtRasterStamp[y - 14] = 0;
                   CRTC.crtRasterStamp[y - 13] = 0;
                   CRTC.crtRasterStamp[y - 12] = 0;
                   CRTC.crtRasterStamp[y - 11] = 0;
                   CRTC.crtRasterStamp[y - 10] = 0;
                   CRTC.crtRasterStamp[y -  9] = 0;
                   CRTC.crtRasterStamp[y -  8] = 0;
                   CRTC.crtRasterStamp[y -  7] = 0;
                   CRTC.crtRasterStamp[y -  6] = 0;
                   CRTC.crtRasterStamp[y -  5] = 0;
                   CRTC.crtRasterStamp[y -  4] = 0;
                   CRTC.crtRasterStamp[y -  3] = 0;
                   CRTC.crtRasterStamp[y -  2] = 0;
                   CRTC.crtRasterStamp[y -  1] = 0;
                 }
               } else {
                 Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
               }
             }
*/
           }
           break;
         }  //スプライトスクロールレジスタのswitch
         return;
       } else if (a < 0x00eb0800) {  //$FF
         if (!SpriteScreen.sprAccessible) {
           super.mmdWw (a, d);  //バスエラー
         }
       } else if (a < 0x00eb8000) {  //各種レジスタ
         switch (a) {
         case SpriteScreen.SPR_REG0_BG0_X:
           SpriteScreen.sprReg0Bg0XPort = d & 1023;
           {
             int curr = SpriteScreen.sprReg0Bg0XMask == 0 ? SpriteScreen.sprReg0Bg0XPort : SpriteScreen.sprReg0Bg0XTest;
             if (SpriteScreen.sprReg0Bg0XCurr != curr) {
               SpriteScreen.sprReg0Bg0XCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   (SpriteScreen.sprReg4BgCtrlCurr & (1 << 9 | 1 << 0)) == (1 << 9 | 1 << 0)) {  //BG0が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG1_BG0_Y:
           SpriteScreen.sprReg1Bg0YPort = d & 1023;
           {
             int curr = SpriteScreen.sprReg1Bg0YMask == 0 ? SpriteScreen.sprReg1Bg0YPort : SpriteScreen.sprReg1Bg0YTest;
             if (SpriteScreen.sprReg1Bg0YCurr != curr) {
               SpriteScreen.sprReg1Bg0YCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   (SpriteScreen.sprReg4BgCtrlCurr & (1 << 9 | 1 << 0)) == (1 << 9 | 1 << 0)) {  //BG0が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG2_BG1_X:
           SpriteScreen.sprReg2Bg1XPort = d & 1023;
           {
             int curr = SpriteScreen.sprReg2Bg1XMask == 0 ? SpriteScreen.sprReg2Bg1XPort : SpriteScreen.sprReg2Bg1XTest;
             if (SpriteScreen.sprReg2Bg1XCurr != curr) {
               SpriteScreen.sprReg2Bg1XCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   (SpriteScreen.sprReg4BgCtrlCurr & (1 << 9 | 1 << 3)) == (1 << 9 | 1 << 3)) {  //BG1が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG3_BG1_Y:
           SpriteScreen.sprReg3Bg1YPort = d & 1023;
           {
             int curr = SpriteScreen.sprReg3Bg1YMask == 0 ? SpriteScreen.sprReg3Bg1YPort : SpriteScreen.sprReg3Bg1YTest;
             if (SpriteScreen.sprReg3Bg1YCurr != curr) {
               SpriteScreen.sprReg3Bg1YCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   (SpriteScreen.sprReg4BgCtrlCurr & (1 << 9 | 1 << 3)) == (1 << 9 | 1 << 3)) {  //BG1が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG4_BG_CTRL:
           SpriteScreen.sprReg4BgCtrlPort = d & 2047;
           {
             int curr = SpriteScreen.sprReg4BgCtrlPort & ~SpriteScreen.sprReg4BgCtrlMask | SpriteScreen.sprReg4BgCtrlTest & SpriteScreen.sprReg4BgCtrlMask;
             if (SpriteScreen.sprReg4BgCtrlCurr != curr) {
               SpriteScreen.sprReg4BgCtrlCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0) {  //スプライト画面が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG5_H_FRONT_END:
           SpriteScreen.sprReg5HFrontEndPort = d & 0xff;
           {
             int curr = SpriteScreen.sprReg5HFrontEndMask == 0 ? SpriteScreen.sprReg5HFrontEndPort : SpriteScreen.sprReg5HFrontEndTest;
             if (SpriteScreen.sprReg5HFrontEndCurr != curr) {
               SpriteScreen.sprReg5HFrontEndCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG6_H_BACK_END:
           SpriteScreen.sprReg6HBackEndPort = d & 63;
           {
             int curr = SpriteScreen.sprReg6HBackEndMask == 0 ? SpriteScreen.sprReg6HBackEndPort : SpriteScreen.sprReg6HBackEndTest;
             if (SpriteScreen.sprReg6HBackEndCurr != curr) {
               SpriteScreen.sprReg6HBackEndCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG7_V_BACK_END:
           SpriteScreen.sprReg7VBackEndPort = d & 0xff;
           {
             int curr = SpriteScreen.sprReg7VBackEndMask == 0 ? SpriteScreen.sprReg7VBackEndPort : SpriteScreen.sprReg7VBackEndTest;
             if (SpriteScreen.sprReg7VBackEndCurr != curr) {
               SpriteScreen.sprReg7VBackEndCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         case SpriteScreen.SPR_REG8_RESO:
           SpriteScreen.sprReg8ResoPort = d & 31;
           {
             int curr = SpriteScreen.sprReg8ResoPort & ~SpriteScreen.sprReg8ResoMask | SpriteScreen.sprReg8ResoTest & SpriteScreen.sprReg8ResoMask;
             if (SpriteScreen.sprReg8ResoCurr != curr) {
               SpriteScreen.sprReg8ResoCurr = curr;
/*
               if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
                   SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
                 CRTC.crtAllStamp += 2;
               }
*/
             }
           }
           return;
         }
         return;
       } else {  //PCGエリアとテキストエリア
         if (!SpriteScreen.sprAccessible) {
           super.mmdWw (a, d);  //バスエラー
         }
         if (0x00ebe000 <= a) {  //テキストエリア1
           int n = a >> 1 & 0x0fff;
           SpriteScreen.sprT1Num[n] = (short) ((d & 0xff) << 3);
           SpriteScreen.sprT1ColPort[n] = (short) ((d & 0x0f00) >> 4);
           SpriteScreen.sprT1H[n] = (short) (d << 1) < 0;
           SpriteScreen.sprT1V[n] = (short) d < 0;
/*
           if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
               SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
             if ((SpriteScreen.sprReg4BgCtrlCurr & 0b111) == 0b011) {  //BG0が表示されていてBG0にTEXT1が割り当てられている
               if ((SpriteScreen.sprReg8ResoCurr & 0x0003) == 0) {  //水平256ドット→8x8
                 int y = (n >> 6 << 3) - SpriteScreen.sprReg1Bg0YCurr;
                 CRTC.crtRasterStamp[y     & 511] = 0;
                 CRTC.crtRasterStamp[y + 1 & 511] = 0;
                 CRTC.crtRasterStamp[y + 2 & 511] = 0;
                 CRTC.crtRasterStamp[y + 3 & 511] = 0;
                 CRTC.crtRasterStamp[y + 4 & 511] = 0;
                 CRTC.crtRasterStamp[y + 5 & 511] = 0;
                 CRTC.crtRasterStamp[y + 6 & 511] = 0;
                 CRTC.crtRasterStamp[y + 7 & 511] = 0;
               } else {  //水平512ドット→16x16
                 int y = (n >> 6 << 4) - SpriteScreen.sprReg1Bg0YCurr;
                 CRTC.crtRasterStamp[y      & 1023] = 0;
                 CRTC.crtRasterStamp[y +  1 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  2 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  3 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  4 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  5 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  6 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  7 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  8 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  9 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 10 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 11 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 12 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 13 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 14 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 15 & 1023] = 0;
               }
             }
             if ((SpriteScreen.sprReg4BgCtrlCurr & 0b111_000) == 0b011_000) {  //BG1が表示されていてBG1にTEXT1が割り当てられている
               int y = (n >> 6 << 3) - SpriteScreen.sprReg3Bg1YCurr;
               CRTC.crtRasterStamp[y     & 511] = 0;
               CRTC.crtRasterStamp[y + 1 & 511] = 0;
               CRTC.crtRasterStamp[y + 2 & 511] = 0;
               CRTC.crtRasterStamp[y + 3 & 511] = 0;
               CRTC.crtRasterStamp[y + 4 & 511] = 0;
               CRTC.crtRasterStamp[y + 5 & 511] = 0;
               CRTC.crtRasterStamp[y + 6 & 511] = 0;
               CRTC.crtRasterStamp[y + 7 & 511] = 0;
             }
           }
*/
         } else if (0x00ebc000 <= a) {  //テキストエリア0
           int n = a >> 1 & 0x0fff;
           SpriteScreen.sprT0Num[n] = (short) ((d & 0xff) << 3);
           SpriteScreen.sprT0ColPort[n] = (short) ((d & 0x0f00) >> 4);
           SpriteScreen.sprT0H[n] = (short) (d << 1) < 0;
           SpriteScreen.sprT0V[n] = (short) d < 0;
/*
           if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
               SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
             if ((SpriteScreen.sprReg4BgCtrlCurr & 0b111) == 0b001) {  //BG0が表示されていてBG0にTEXT0が割り当てられている
               if ((SpriteScreen.sprReg8ResoCurr & 0x0003) == 0) {  //水平256ドット→8x8
                 int y = (n >> 6 << 3) - SpriteScreen.sprReg1Bg0YCurr;
                 CRTC.crtRasterStamp[y     & 511] = 0;
                 CRTC.crtRasterStamp[y + 1 & 511] = 0;
                 CRTC.crtRasterStamp[y + 2 & 511] = 0;
                 CRTC.crtRasterStamp[y + 3 & 511] = 0;
                 CRTC.crtRasterStamp[y + 4 & 511] = 0;
                 CRTC.crtRasterStamp[y + 5 & 511] = 0;
                 CRTC.crtRasterStamp[y + 6 & 511] = 0;
                 CRTC.crtRasterStamp[y + 7 & 511] = 0;
               } else {  //水平512ドット→16x16
                 int y = (n >> 6 << 4) - SpriteScreen.sprReg1Bg0YCurr;
                 CRTC.crtRasterStamp[y      & 1023] = 0;
                 CRTC.crtRasterStamp[y +  1 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  2 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  3 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  4 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  5 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  6 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  7 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  8 & 1023] = 0;
                 CRTC.crtRasterStamp[y +  9 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 10 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 11 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 12 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 13 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 14 & 1023] = 0;
                 CRTC.crtRasterStamp[y + 15 & 1023] = 0;
               }
             }
             if ((SpriteScreen.sprReg4BgCtrlCurr & 0b111_000) == 0b001_000) {  //BG1が表示されていてBG1にTEXT0が割り当てられている
               int y = (n >> 6 << 3) - SpriteScreen.sprReg3Bg1YCurr;
               CRTC.crtRasterStamp[y     & 511] = 0;
               CRTC.crtRasterStamp[y + 1 & 511] = 0;
               CRTC.crtRasterStamp[y + 2 & 511] = 0;
               CRTC.crtRasterStamp[y + 3 & 511] = 0;
               CRTC.crtRasterStamp[y + 4 & 511] = 0;
               CRTC.crtRasterStamp[y + 5 & 511] = 0;
               CRTC.crtRasterStamp[y + 6 & 511] = 0;
               CRTC.crtRasterStamp[y + 7 & 511] = 0;
             }
           }
*/
         }
         int t = a >> 2 & 0x1fff;
         SpriteScreen.sprPatPort[t] = (a & 0x02) == 0 ? d << 16 | (char) SpriteScreen.sprPatPort[t] : (SpriteScreen.sprPatPort[t] & ~0xffff) | (char) d;
/*
         if (VideoController.vcnReg3Curr << 31 - 6 < 0 &&  //スプライト画面が表示されている
             SpriteScreen.sprReg4BgCtrlCurr << 31 - 9 < 0) {  //スプライト画面が表示されている
           int num = a >> 7 & 0xff;  //num=パターン番号
           if (SpriteScreen.SPR_RRMAP) {
             for (int i = num << 2, nn = 0; nn <= 128 - 32; nn += 32) {
               for (int map = SpriteScreen.sprPPmap[i++], n = nn; map != 0; map <<= 1, n++) {
                 if (map >= 0) {  //このスプライトには割り当てられていない
                   continue;
                 }
                 int y = SpriteScreen.sprY[n];
                 if (true) {
                   if (y < 16) {
                     for (int k = 0; k < y; k++) {
                       CRTC.crtRasterStamp[k] = 0;
                     }
                   } else {
                     CRTC.crtRasterStamp[y - 16] = 0;
                     CRTC.crtRasterStamp[y - 15] = 0;
                     CRTC.crtRasterStamp[y - 14] = 0;
                     CRTC.crtRasterStamp[y - 13] = 0;
                     CRTC.crtRasterStamp[y - 12] = 0;
                     CRTC.crtRasterStamp[y - 11] = 0;
                     CRTC.crtRasterStamp[y - 10] = 0;
                     CRTC.crtRasterStamp[y -  9] = 0;
                     CRTC.crtRasterStamp[y -  8] = 0;
                     CRTC.crtRasterStamp[y -  7] = 0;
                     CRTC.crtRasterStamp[y -  6] = 0;
                     CRTC.crtRasterStamp[y -  5] = 0;
                     CRTC.crtRasterStamp[y -  4] = 0;
                     CRTC.crtRasterStamp[y -  3] = 0;
                     CRTC.crtRasterStamp[y -  2] = 0;
                     CRTC.crtRasterStamp[y -  1] = 0;
                   }
                 } else {
                   Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
                 }
               }  //for map,n
             }  //for i,nn
           } else {  //!SpriteScreen.SPR_RRMAP
             for (int i = 96; i >= 0; i -= 32) {
               int map = SpriteScreen.sprPmap[i >> 5][num];
               if (map == 0) {
                 continue;
               }
               for (int n = i + 31; n >= i; n--) {  //n=スプライト番号
                 if (map < 0) {
                   int y = SpriteScreen.sprY[n];
                   if (true) {
                     if (y < 16) {
                       for (int k = 0; k < y; k++) {
                         CRTC.crtRasterStamp[k] = 0;
                       }
                     } else {
                       CRTC.crtRasterStamp[y - 16] = 0;
                       CRTC.crtRasterStamp[y - 15] = 0;
                       CRTC.crtRasterStamp[y - 14] = 0;
                       CRTC.crtRasterStamp[y - 13] = 0;
                       CRTC.crtRasterStamp[y - 12] = 0;
                       CRTC.crtRasterStamp[y - 11] = 0;
                       CRTC.crtRasterStamp[y - 10] = 0;
                       CRTC.crtRasterStamp[y -  9] = 0;
                       CRTC.crtRasterStamp[y -  8] = 0;
                       CRTC.crtRasterStamp[y -  7] = 0;
                       CRTC.crtRasterStamp[y -  6] = 0;
                       CRTC.crtRasterStamp[y -  5] = 0;
                       CRTC.crtRasterStamp[y -  4] = 0;
                       CRTC.crtRasterStamp[y -  3] = 0;
                       CRTC.crtRasterStamp[y -  2] = 0;
                       CRTC.crtRasterStamp[y -  1] = 0;
                     }
                   } else {
                     Arrays.fill (CRTC.crtRasterStamp, y < 16 ? 0 : y - 16, y, 0);
                   }
                 }  //if map<0
                 map <<= 1;
               }  //for n
             }  //for i
           }  //if SpriteScreen.SPR_RRMAP/!SpriteScreen.SPR_RRMAP
         }  //(VideoController.vcnReg3Curr&0x40)!=0
*/
       }  //PCGエリアとテキストエリア
     }  //mmdWw
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       mmdWw (a    , d >> 16);
       mmdWw (a + 2, d      );
     }
  },  //MMD_SPR

  //--------------------------------------------------------------------------------
  //MMD_XB3 拡張ボード領域3
  //  $00EC0000～$00ECFFFF  64KB
  //  0x00ec0000  ユーザI/Oエリア
  //  0x00ec0000  Awesome
  //  0x00ec0000  Xellent30
  //  0x00ecc000  Mercury
  //  0x00ece000  Neptune
  //  0x00ecf000  Venus-X/030
  MMD_XB3 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張ボード領域 3" : "Expansion Board Area 3";
    }
    //リード
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         return (a & 1) == 0 ? 0 : XEiJ.xt3PortRead ();
       }
       return super.mmdRbz (a);  //バスエラー
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         return XEiJ.xt3PortRead ();
       }
       return super.mmdRwz (a);  //バスエラー
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         return XEiJ.xt3PortRead () << 16 | XEiJ.xt3PortRead ();
       }
       return super.mmdRls (a);  //バスエラー
     }
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         if ((a & 1) == 0) {
         } else {
           XEiJ.xt3PortWrite (d);
         }
         return;
       }
       super.mmdWb (a, d);  //バスエラー
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         XEiJ.xt3PortWrite (d);
         return;
       }
       super.mmdWw (a, d);  //バスエラー
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       if (XEiJ.mdlModel == XEiJ.MDL_XELLENT30 &&
           XEiJ.XT3_PORT_ADDRESS <= a && a < XEiJ.XT3_PORT_ADDRESS + 0x4000) {  //Xellent30設定ポート
         XEiJ.xt3PortWrite (d >> 16);
         XEiJ.xt3PortWrite (d);
         return;
       }
       super.mmdWl (a, d);  //バスエラー
     }
  },  //MMD_XB3

  //--------------------------------------------------------------------------------
  //MMD_XTM Xellent30のSRAM
  MMD_XTM {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "Xellent30 の SRAM" : "SRAM on Xellent30";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a -= XEiJ.xt3MemoryAddress;
      return XEiJ.xt3MemoryArray[a];
    }
    @Override protected int mmdPbz (int a) {
      a -= XEiJ.xt3MemoryAddress;
      return XEiJ.xt3MemoryArray[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a -= XEiJ.xt3MemoryAddress;
      return XEiJ.xt3MemoryArray[a] << 8 | (XEiJ.xt3MemoryArray[a + 1] & 0xff);
    }
    @Override protected int mmdPwz (int a) {
      a -= XEiJ.xt3MemoryAddress;
      return (char) (XEiJ.xt3MemoryArray[a] << 8 | (XEiJ.xt3MemoryArray[a + 1] & 0xff));
    }
    @Override protected int mmdPls (int a) {
      a -= XEiJ.xt3MemoryAddress;
      return XEiJ.xt3MemoryArray[a] << 24 | (XEiJ.xt3MemoryArray[a + 1] & 0xff) << 16 | (char) (XEiJ.xt3MemoryArray[a + 2] << 8 | (XEiJ.xt3MemoryArray[a + 3] & 0xff));
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       return XEiJ.xt3MemoryArray[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       return XEiJ.xt3MemoryArray[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       return XEiJ.xt3MemoryArray[a] << 8 | (XEiJ.xt3MemoryArray[a + 1] & 0xff);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       return (char) (XEiJ.xt3MemoryArray[a] << 8 | (XEiJ.xt3MemoryArray[a + 1] & 0xff));
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       return XEiJ.xt3MemoryArray[a] << 24 | (XEiJ.xt3MemoryArray[a + 1] & 0xff) << 16 | (char) (XEiJ.xt3MemoryArray[a + 2] << 8 | (XEiJ.xt3MemoryArray[a + 3] & 0xff));
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       XEiJ.xt3MemoryArray[a    ] = (byte)  d;
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       XEiJ.xt3MemoryArray[a    ] = (byte) (d >> 8);
       XEiJ.xt3MemoryArray[a + 1] = (byte)  d;
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       a -= XEiJ.xt3MemoryAddress;
       XEiJ.xt3MemoryArray[a    ] = (byte) (d >> 24);
       XEiJ.xt3MemoryArray[a + 1] = (byte) (d >> 16);
       XEiJ.xt3MemoryArray[a + 2] = (byte) (d >> 8);
       XEiJ.xt3MemoryArray[a + 3] = (byte)  d;
     }
  },  //MMD_XTM

  //--------------------------------------------------------------------------------
  //MMD_SMR SRAM
  MMD_SMR {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "SRAM" : "SRAM";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       return (short) mmdRwz (a);
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * XEiJ.busCycleUnit;
       int d;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         d = MainMemory.mmrBuffer.getChar (a);
       } else {
         d = (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
       return d;
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * 2 * XEiJ.busCycleUnit;
       int d;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         d = MainMemory.mmrBuffer.getInt (a);
       } else {
         d = MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
       return d;
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if (XEiJ.smrWriteEnableOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * XEiJ.busCycleUnit;
         MainMemory.mmrM8[a] = (byte) d;
         if (a == 0x00ed002b || a == 0x00ed0059) {  //キーボードの配列または字体が変化した
           Keyboard.kbdRepaint ();  //キーボードが表示されているときkbdImageを作り直して再描画する
         } else if (a == 0x00ed005a) {  //HDMAX
           if (0x00fe0000 <= XEiJ.regPC0 && XEiJ.regPC0 <= 0x00ffffff) {  //IPLROMがSRAMのHDMAXを書き換えようとしている 
             XEiJ.smrOverride ();  //SRAMの設定を上書きする
           }
         }
       } else {
         System.out.printf ("%08X writing $%02X to $%08X in a write-protected state\n", XEiJ.regPC0, d & 0xff, a);
         if (XEiJ.smrSRAMBusErrorOn) {
           super.mmdWb (a, d);  //バスエラー
         }
       }
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       if (XEiJ.smrWriteEnableOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         XEiJ.mpuClockTime += XEiJ.MPU_SRAM_WAIT * XEiJ.busCycleUnit;
         MainMemory.mmrM8[a    ] = (byte) (d >> 8);
         MainMemory.mmrM8[a + 1] = (byte)  d;
         if (a == (0x00ed002b & -2) || a == (0x00ed0059 & -2)) {  //キーボードの配列または字体が変化した
           Keyboard.kbdRepaint ();  //キーボードが表示されているときkbdImageを作り直して再描画する
         }
       } else {
         System.out.printf ("%08X writing $%04X to $%08X in a write-protected state\n", XEiJ.regPC0, d & 0xffff, a);
         if (XEiJ.smrSRAMBusErrorOn) {
           super.mmdWw (a, d);  //バスエラー
         }
       }
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       if (XEiJ.smrWriteEnableOn) {
         mmdWw (a    , d >> 16);
         mmdWw (a + 2, d      );
       } else {
         System.out.printf ("%08X writing $%08X to $%08X in a write-protected state\n", XEiJ.regPC0, d, a);
         if (XEiJ.smrSRAMBusErrorOn) {
           super.mmdWl (a, d);  //バスエラー
         }
       }
     }
  },  //MMD_SMR

  //--------------------------------------------------------------------------------
  //MMD_XB4 拡張ボード領域4
  //  $00EE0000～$00EFFFFF  バンクメモリ
  MMD_XB4 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "拡張ボード領域 4" : "Expansion Board Area 4";
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         return XEiJ.bnkMemory[a];
       }
       return super.mmdRbs (a);  //バスエラー
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         return XEiJ.bnkMemory[a] & 0xff;
       }
       return super.mmdRbz (a);  //バスエラー
     }
    @Override protected int mmdRws (int a) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         return ((XEiJ.bnkMemory[a    ]       ) << 8 |
                 (XEiJ.bnkMemory[a + 1] & 0xff));
       }
       return super.mmdRws (a);  //バスエラー
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         return ((XEiJ.bnkMemory[a    ] & 0xff) << 8 |
                 (XEiJ.bnkMemory[a + 1] & 0xff));
       }
       return super.mmdRwz (a);  //バスエラー
     }
    @Override protected int mmdRls (int a) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         return ((XEiJ.bnkMemory[a    ]       ) << 24 |
                 (XEiJ.bnkMemory[a + 1] & 0xff) << 16 |
                 (XEiJ.bnkMemory[a + 2] & 0xff) <<  8 |
                 (XEiJ.bnkMemory[a + 3] & 0xff));
       }
       return super.mmdRls (a);  //バスエラー
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         XEiJ.bnkMemory[a] = (byte) d;
         return;
       }
       super.mmdWb (a, d);  //バスエラー
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         XEiJ.bnkMemory[a    ] = (byte) (d >> 8);
         XEiJ.bnkMemory[a + 1] = (byte)  d;
         return;
       }
       super.mmdWw (a, d);  //バスエラー
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       if (XEiJ.bnkOn) {
         a &= XEiJ.BUS_MOTHER_MASK;
         a = XEiJ.bnkPageStart + (a - 0x00ee0000);
         XEiJ.bnkMemory[a    ] = (byte) (d >> 24);
         XEiJ.bnkMemory[a + 1] = (byte) (d >> 16);
         XEiJ.bnkMemory[a + 2] = (byte) (d >>  8);
         XEiJ.bnkMemory[a + 3] = (byte)  d;
         return;
       }
       super.mmdWl (a, d);  //バスエラー
     }
  },  //MMD_XB4

  //--------------------------------------------------------------------------------
  //MMD_CG1 CGROM1
  MMD_CG1 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "CGROM" : "CGROM";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getShort (a);
       } else {
         return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getChar (a);
       } else {
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit * 2;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getInt (a);
       } else {
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
     }
  },  //MMD_CG1

  //--------------------------------------------------------------------------------
  //MMD_CG2 CGROM2
  MMD_CG2 {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "CGROM" : "CGROM";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getShort (a);
       } else {
         return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getChar (a);
       } else {
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit * 2;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getInt (a);
       } else {
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
     }
  },  //MMD_CG2

  //--------------------------------------------------------------------------------
  //MMD_ROM ROM
  MMD_ROM {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "ROM" : "ROM";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a];
    }
    @Override protected int mmdPbz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      return MainMemory.mmrM8[a] & 0xff;
    }
    @Override protected int mmdPws (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getShort (a);
      } else {
        return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
      }
    }
    @Override protected int mmdPwz (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getChar (a);
      } else {
        return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
      }
    }
    @Override protected int mmdPls (int a) {
      a &= XEiJ.BUS_MOTHER_MASK;
      if (MainMemory.MMR_USE_BYTE_BUFFER) {
        return MainMemory.mmrBuffer.getInt (a);
      } else {
        return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
      }
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a];
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       return MainMemory.mmrM8[a] & 0xff;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getShort (a);
       } else {
         return MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff);
       }
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getChar (a);
       } else {
         return (char) (MainMemory.mmrM8[a] << 8 | (MainMemory.mmrM8[a + 1] & 0xff));
       }
     }
    @Override protected int mmdRls (int a) throws M68kException {
       a &= XEiJ.BUS_MOTHER_MASK;
       XEiJ.mpuClockTime += XEiJ.mpuRomWait * XEiJ.busCycleUnit * 2;
       if (MainMemory.MMR_USE_BYTE_BUFFER) {
         return MainMemory.mmrBuffer.getInt (a);
       } else {
         return MainMemory.mmrM8[a] << 24 | (MainMemory.mmrM8[a + 1] & 0xff) << 16 | (char) (MainMemory.mmrM8[a + 2] << 8 | (MainMemory.mmrM8[a + 3] & 0xff));
       }
     }
  },  //MMD_ROM

  //--------------------------------------------------------------------------------
  //MMD_IBP 命令ブレークポイント
  MMD_IBP {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "命令ブレークポイント" : "Instruction Break Point";
    }
    @Override protected int mmdRwz (int a) throws M68kException {
       if (InstructionBreakPoint.IBP_ON) {
         InstructionBreakPoint.InstructionBreakRecord[] hashTable = (XEiJ.regSRS != 0 ?
                                                                     InstructionBreakPoint.ibpSuperHashTable :
                                                                     InstructionBreakPoint.ibpUserHashTable);
         for (InstructionBreakPoint.InstructionBreakRecord r = hashTable[a >> 1 & InstructionBreakPoint.IBP_HASH_MASK];
              r != null;
              r = r.ibrNext) {  //同じ物理ハッシュコードを持つ命令ブレークポイントについて
           if (r.ibrPhysicalAddress == a) {  //命令ブレークポイントが設定されているとき
             if (r.ibrValue == r.ibrTarget) {  //現在値が目標値と一致しているとき
               if (r.ibrThreshold < 0) {  //インスタントのとき
                 InstructionBreakPoint.ibpRemove (r.ibrLogicalAddress, XEiJ.regSRS);  //取り除く
                 XEiJ.mpuContinue = true;  //ステップ実行を継続する
               } else if (r.ibrTarget < r.ibrThreshold) {  //インスタント化しているとき
                 r.ibrTarget = r.ibrThreshold;  //目標値を閾値に戻す
                 XEiJ.mpuContinue = true;  //ステップ実行を継続する
               } else {  //インスタントでなくインスタント化していないとき
                 if (r.ibrScriptElement != null &&  //スクリプトが指定されていて
                     r.ibrScriptElement.exlEval (ExpressionEvaluator.EVM_EXPRESSION).exlFloatValue.iszero ()) {  //条件が成立していないとき
                   break;  //続行する
                 }
                 r.ibrTarget++;  //目標値を増やす
                 XEiJ.mpuContinue = false;  //ステップ実行を継続しない
               }
               M68kException.m6eNumber = -1;  //命令ブレークポイントによる停止。XEiJ.mpuContinueを設定すること
               throw M68kException.m6eSignal;  //停止する
             } else {  //現在値が目標値と一致していないとき
               r.ibrValue++;  //現在値を増やす
               break;  //続行する
             }
           }
         }
       }
       if (DataBreakPoint.DBP_ON) {
         return DataBreakPoint.dbpMemoryMap[a >>> XEiJ.BUS_PAGE_BITS].mmdRwz (a);
       } else {
         return XEiJ.busMemoryMap[a >>> XEiJ.BUS_PAGE_BITS].mmdRwz (a);
       }
     }
  },  //MMD_IBP

  //--------------------------------------------------------------------------------
  //MMD_DBP データブレークポイント
  MMD_DBP {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "データブレークポイント" : "Data Break Point";
    }
    //ピーク
    @Override protected byte mmdPbs (int a) {
      return (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdPbs (a);
    }
    @Override protected int mmdPbz (int a) {
      return (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdPbz (a);
    }
    @Override protected int mmdPws (int a) {
      return (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdPws (a);
    }
    @Override protected int mmdPwz (int a) {
      return (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdPwz (a);
    }
    @Override protected int mmdPls (int a) {
      return (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdPls (a);
    }
    //リード
    @Override protected byte mmdRbs (int a) throws M68kException {
       int d = (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdRbs (a);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_BYTE, a, d);
       return (byte) d;
     }
    @Override protected int mmdRbz (int a) throws M68kException {
       int d = (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdRbz (a);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_BYTE, a, d);
       return d;
     }
    @Override protected int mmdRws (int a) throws M68kException {
       int d = (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdRws (a);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_WORD, a, d);
       return d;
     }
    @Override protected int mmdRwz (int a) throws M68kException {
       int d = (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdRwz (a);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_WORD, a, d);
       return d;
     }
    @Override protected int mmdRls (int a) throws M68kException {
       int d = (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdRls (a);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_LONG, a, d);
       return d;
     }
    //ライト
    @Override protected void mmdWb (int a, int d) throws M68kException {
       (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdWb (a, d);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_BYTE, a, d);
     }
    @Override protected void mmdWw (int a, int d) throws M68kException {
       (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdWw (a, d);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_WORD, a, d);
     }
    @Override protected void mmdWl (int a, int d) throws M68kException {
       (XEiJ.regSRS != 0 ? XEiJ.busSuperMap : XEiJ.busUserMap)[a >>> XEiJ.BUS_PAGE_BITS].mmdWl (a, d);
       DataBreakPoint.dbpBreak (DataBreakPoint.DBP_LONG, a, d);
     }
  },  //MMD_DBP

  //--------------------------------------------------------------------------------
  //MMD_NUL ヌルデバイス
  MMD_NUL {
    @Override public String toString () {
      return Multilingual.mlnJapanese ? "ヌルデバイス" : "Null Device";
    }
  };  //MMD_NUL

  //--------------------------------------------------------------------------------
  //ピークのデフォルト
  //  エラーや副作用なしでリードする
  //  バスエラーのときは-1をキャストした値を返す
  //  リードがデバイスの状態を変化させる可能性がある場合は個別に処理すること
  protected byte mmdPbs (int a) {
    try {
      return (byte) mmdRbz (a);
    } catch (M68kException e) {
    }
    return -1;
  }
  protected int mmdPbz (int a) {
    try {
      return mmdRbz (a);
    } catch (M68kException e) {
    }
    return 0xff;
  }
  protected int mmdPws (int a) {
    try {
      return (short) mmdRwz (a);
    } catch (M68kException e) {
    }
    return -1;
  }
  protected int mmdPwz (int a) {
    try {
      return mmdRwz (a);
    } catch (M68kException e) {
    }
    return 0xffff;
  }
  protected int mmdPls (int a) {
    try {
      return mmdRls (a);
    } catch (M68kException e) {
    }
    return -1;
  }
  //リードのデフォルト
  //  バイトとワードの符号拡張はゼロ拡張を呼び出す
  //  符号なしとロングはバスエラー
  protected byte mmdRbs (int a) throws M68kException {
    return (byte) mmdRbz (a);
  }
  protected int mmdRbz (int a) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_READ;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_READ;
    M68kException.m6eSize = XEiJ.MPU_SS_BYTE;
    throw M68kException.m6eSignal;
  }
  protected int mmdRws (int a) throws M68kException {
    return (short) mmdRwz (a);
  }
  protected int mmdRwz (int a) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_READ;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_READ;
    M68kException.m6eSize = XEiJ.MPU_SS_WORD;
    throw M68kException.m6eSignal;
  }
  protected int mmdRls (int a) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_READ;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_READ;
    M68kException.m6eSize = XEiJ.MPU_SS_LONG;
    throw M68kException.m6eSignal;
  }
  //ポークのデフォルト
  //  エラーや副作用なしでライトする
  //  ライトがデバイスの状態を変化させる可能性がある場合は個別に処理すること
  protected void mmdVb (int a, int d) {
    try {
      mmdWb (a, d);
    } catch (M68kException e) {
    }
  }
  protected void mmdVw (int a, int d) {
    try {
      mmdWw (a, d);
    } catch (M68kException e) {
    }
  }
  protected void mmdVl (int a, int d) {
    try {
      mmdWl (a, d);
    } catch (M68kException e) {
    }
  }
  //ライトのデフォルト
  //  すべてバスエラー
  protected void mmdWb (int a, int d) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_WRITE;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_WRITE;
    M68kException.m6eSize = XEiJ.MPU_SS_BYTE;
    throw M68kException.m6eSignal;
  }
  protected void mmdWw (int a, int d) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_WRITE;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_WRITE;
    M68kException.m6eSize = XEiJ.MPU_SS_WORD;
    throw M68kException.m6eSignal;
  }
  protected void mmdWl (int a, int d) throws M68kException {
    M68kException.m6eFSLW |= M68kException.M6E_FSLW_BUS_ERROR_ON_WRITE;  //サイズは設定済み。分割されている場合があることに注意
    M68kException.m6eNumber = M68kException.M6E_ACCESS_FAULT;
    M68kException.m6eAddress = a;
    M68kException.m6eDirection = XEiJ.MPU_WR_WRITE;
    M68kException.m6eSize = XEiJ.MPU_SS_LONG;
    throw M68kException.m6eSignal;
  }
}  //enum MemoryMappedDevice



